/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ ou Rg[ ]
 */

#if XM7_VER == 1 && defined(BUBBLE)

#include <string.h>
#include <stdlib.h>
#include "xm7.h"
#include "device.h"
#include "bubble.h"
#include "fdc.h"
#include "opn.h"

#ifdef XM7DASH
/*
 *	O[o [N
 */

/* WX^ */
BYTE bmc_datareg[2];					/* f[^WX^ */
BYTE bmc_command[2];					/* R}hWX^ */
BYTE bmc_status[2];						/* Xe[^XWX^ */
BYTE bmc_errorreg[2];					/* G[Xe[^XWX^ */
WORD bmc_pagereg[2];					/* y[WWX^ */
WORD bmc_countreg[2];					/* y[WJEgWX^ */

/* O[N */
WORD bmc_totalcnt[2];					/* g[^JE^ */
WORD bmc_nowcnt[2];						/* JgJE^ */
BYTE bmc_unit[2];						/* jbg */
BYTE bmc_ready[2][BMC_UNITS_128];		/* fB */
BOOL bmc_teject[2][BMC_UNITS_128];		/* ꎞCWFNg */
BOOL bmc_writep[2][BMC_UNITS_128];		/* ݋֎~ */

char bmc_fname[2][BMC_UNITS_128][256+1];/* t@C */
char bmc_name[2][BMC_UNITS_128][BMC_MEDIAS][17];
BOOL bmc_fwritep[2][BMC_UNITS_128];		/* CgveNg */
BYTE bmc_header[2][BMC_UNITS_128][0x20];/* B77t@Cwb_ */
BYTE bmc_medias[2][BMC_UNITS_128];		/* fBA */
BYTE bmc_media[2][BMC_UNITS_128];		/* fBAZNg */
BYTE bmc_access[2][BMC_UNITS_128];		/* ANZXLED */

BOOL bmc_enable[2];						/* LEtO */
BOOL bmc_use[2];						/* gptO */

/*
 *	X^eBbN [N
 */
static BYTE bmc_buffer0[BMC_PSIZE_32];	/* 32bytef[^obt@ */
static BYTE bmc_buffer1[BMC_PSIZE_128];	/* 128bytef[^obt@ */
static BYTE * const bmc_buffer[2] = {&bmc_buffer0[0], &bmc_buffer1[0]};
static BYTE *bmc_dataptr[2];			/* f[^|C^ */
static DWORD bmc_offset[2];				/* ItZbg */
static DWORD bmc_foffset[2][BMC_UNITS_128][BMC_MEDIAS];
static DWORD bmc_fsize[2][BMC_UNITS_128];
#ifdef FDDSND
static BOOL bmc_wait;					/* EFCg[hstO */
#endif

/*
 *	Œle[u
 */
static const WORD bmc_page_size[2] = {	BMC_PSIZE_32,
										BMC_PSIZE_128};
static const WORD bmc_max_addr[2] = {	BMC_MAXADDR_32,
										BMC_MAXADDR_128};

/*
 *	ou Rg[
 *	
 */
BOOL FASTCALL bmc_init(void)
{
	/* t@C֌WZbg */
	memset(bmc_totalcnt, 0, sizeof(bmc_totalcnt));
	memset(bmc_nowcnt, 0, sizeof(bmc_nowcnt));

	memset(bmc_ready, BMC_TYPE_NOTREADY, sizeof(bmc_ready));
	memset(bmc_teject, FALSE, sizeof(bmc_teject));
	memset(bmc_writep, FALSE, sizeof(bmc_writep));
	memset(bmc_fname, 0, sizeof(bmc_fname));
	memset(bmc_fwritep, FALSE, sizeof(bmc_fwritep));
	memset(bmc_medias, 0, sizeof(bmc_medias));

	/* t@CItZbgSăNA */
	memset(bmc_foffset, 0, sizeof(bmc_foffset));

	/* ftHg32KBE128KBL */
	bmc_enable[BMC_32] = FALSE;
	bmc_enable[BMC_128] = TRUE;

	/* EFCg}[htO */
#ifdef FDDSND
	bmc_wait = FALSE;
#endif

	return TRUE;
}

/*
 *	ou Rg[
 *	N[Abv
 */
void FASTCALL bmc_cleanup(void)
{
	/* t@C֌WZbg */
	memset(bmc_ready, 0, sizeof(bmc_ready));
}

/*
 *	ou Rg[
 *	Zbg
 */
void FASTCALL bmc_reset(void)
{
	/* WX^Zbg */
	memset(bmc_datareg, 0, sizeof(bmc_datareg));
	memset(bmc_command, 0xff, sizeof(bmc_command));
	memset(bmc_status, 0, sizeof(bmc_status));
	memset(bmc_errorreg, 0, sizeof(bmc_errorreg));
	memset(bmc_pagereg, 0, sizeof(bmc_pagereg));
	memset(bmc_countreg, 0, sizeof(bmc_countreg));

	memset(bmc_access, 0, sizeof(bmc_access));
	memset(bmc_unit, 0, sizeof(bmc_unit));
	memset(bmc_dataptr, 0, sizeof(bmc_dataptr));
	memset(bmc_offset, 0, sizeof(bmc_offset));

	memset(bmc_use, FALSE, sizeof(bmc_use));
}

/*-[ t@CǗ ]---------------------------------------------------------*/

/*
 *	y[WݏI
 */
static BOOL FASTCALL bmc_write_page(int type)
{
	DWORD offset;
	int handle;

	/* assert */
	ASSERT(bmc_ready[type][bmc_unit[type]] != BMC_TYPE_NOTREADY);
	ASSERT(bmc_dataptr[type]);
	ASSERT(bmc_totalcnt[type] > 0);

	/* ItZbgZo */
	offset = (DWORD)((bmc_pagereg[type] & bmc_max_addr[type]) * bmc_page_size[type]);
	if (bmc_ready[type][bmc_unit[type]] == BMC_TYPE_B77) {
		if (bmc_fsize[type][bmc_unit[type]] < offset + bmc_page_size[type] + 0x0020) {
			return FALSE;
		}
		offset += *(DWORD *)(&bmc_header[type][bmc_unit[type]][0x0014]);
	}

	/*  */
	handle = file_open(bmc_fname[type][bmc_unit[type]], OPEN_RW);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_write(handle, bmc_dataptr[type], bmc_totalcnt[type])) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	return TRUE;
}

/*
 *	y[Wǂݍ
 */
static BOOL FASTCALL bmc_readbuf(int type)
{
	int handle;
	DWORD offset;

	/* y[WAhX`FbN */
	if (type == BMC_32) {
		if (bmc_unit[type] >= BMC_UNITS_32) {
			memset(bmc_buffer[type], 0, bmc_page_size[type]);
			return FALSE;
		}
	}
	else {
		if (bmc_unit[type] >= BMC_UNITS_128) {
			memset(bmc_buffer[type], 0, bmc_page_size[type]);
			return FALSE;
		}
	}

	/* fB`FbN */
	if (bmc_ready[type][bmc_unit[type]] == BMC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* ItZbgZo */
	offset = (DWORD)(bmc_pagereg[type] & bmc_max_addr[type]);
	offset *= bmc_page_size[type];
	if (bmc_ready[type][bmc_unit[type]] == BMC_TYPE_B77) {
		if (bmc_fsize[type][bmc_unit[type]] < offset + bmc_page_size[type] + 0x0020) {
			return FALSE;
		}
		offset += *(DWORD *)(&bmc_header[type][bmc_unit[type]][0x0014]);
	}
	bmc_offset[type] = offset;

	/* ǂݍ */
	memset(bmc_buffer[type], 0, bmc_page_size[type]);
	handle = file_open(bmc_fname[type][bmc_unit[type]], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	file_read(handle, bmc_buffer[type], bmc_page_size[type]);
	file_close(handle);
	return TRUE;
}

/*
 *	B77t@C wb_ǂݍ
 */
static BOOL FASTCALL bmc_readhead(int type, int unit, int index)
{
	int i;
	DWORD offset;
	DWORD temp;
	int handle;

	/* assert */
	ASSERT((type == BMC_32) || (type == BMC_128));
	ASSERT((unit >= 0) && (unit < BMC_UNITS_128));
	ASSERT((index >= 0) && (index < BMC_MEDIAS));
	ASSERT(bmc_ready[type][unit] == BMC_TYPE_B77);

	/* ItZbg */
	offset = bmc_foffset[type][unit][index];

	/* V[NAǂݍ */
	handle = file_open(bmc_fname[type][unit], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_read(handle, bmc_header[type][unit], 0x20)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	/* JZbgTCY */
	temp = 0;
	temp |= bmc_header[type][unit][0x001c + 3];
	temp *= 256;
	temp |= bmc_header[type][unit][0x001c + 2];
	temp *= 256;
	temp |= bmc_header[type][unit][0x001c + 1];
	temp *= 256;
	temp |= bmc_header[type][unit][0x001c + 0];
	bmc_fsize[type][unit] = temp;

	/* ^Cv`FbN */
	if (type == BMC_32) {
		if (bmc_header[type][unit][0x001b] != 0x80) {
			/* 32KBłȂ */
			return FALSE;
		}
#if 0
		if (temp != 32800) {
			/* 32KBłȂ */
			return FALSE;
		}
#endif
	}
	if (type == BMC_128) {
		if (bmc_header[type][unit][0x001b] != 0x90) {
			/* 128KBłȂ */
			return FALSE;
		}
#if 0
		if (temp != 131104) {
			/* 128KBłȂ */
			return FALSE;
		}
#endif
	}

	/* CgveNgݒ */
	if (bmc_fwritep[type][unit]) {
		bmc_writep[type][unit] = TRUE;
	}
	else {
		if (bmc_header[type][unit][0x001a] & 0x10) {
			bmc_writep[type][unit] = TRUE;
		}
		else {
			bmc_writep[type][unit] = FALSE;
		}
	}

	/* ItZbg */
	*(DWORD *)(&bmc_header[type][unit][0x0014]) = offset + 0x0020;

	return TRUE;
}

/*
 *	݂̃fBÃCgveNg؂ւ
 */
BOOL FASTCALL bmc_setwritep(int type, int unit, BOOL writep)
{
	/* assert */
	ASSERT((type == BMC_32) || (type == BMC_128));
	ASSERT((unit >= 0) && (unit < BMC_UNITS_128));
	ASSERT((writep == TRUE) || (writep == FALSE));

	/* fBłȂ΂ȂȂ */
	if (bmc_ready[type][unit] == BMC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* t@CݕsȂ_ */
	if (bmc_fwritep[type][unit]) {
		return FALSE;
	}

	/* ǂݍ݁AݒA */
	if (bmc_ready[type][unit] == BMC_TYPE_B77) {
		BYTE header[0x2b0];
		DWORD offset;
		int handle;

		offset = bmc_foffset[type][unit][bmc_media[type][unit]];
		handle = file_open(bmc_fname[type][unit], OPEN_RW);
		if (handle == -1) {
			return FALSE;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_read(handle, header, 0x20)) {
			file_close(handle);
			return FALSE;
		}
		if (writep) {
			header[0x001a] |= 0x10;
		}
		else {
			header[0x001a] &= ~0x10;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_write(handle, header, 0x20)) {
			file_close(handle);
			return FALSE;
		}

		file_close(handle);
	}

	/*  */
	bmc_writep[type][unit] = writep;
	return TRUE;
}

/*
 *	fBAԍݒ
 */
BOOL FASTCALL bmc_setmedia(int type, int unit, int index)
{
	/* assert */
	ASSERT((type == BMC_32) || (type == BMC_128));
	ASSERT((unit >= 0) && (unit < BMC_UNITS_128));
	ASSERT((index >= 0) && (index < BMC_MEDIAS));

	/* fBԂ */
	if (bmc_ready[type][unit] == BMC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* 32KBt@C̏ꍇAindex = 0 */
	if ((bmc_ready[type][unit] == BMC_TYPE_32) &&
		((index != 0) || (type != BMC_32))) {
		return FALSE;
	}
	/* 128KBt@C̏ꍇAindex = 0 */
	if ((bmc_ready[type][unit] == BMC_TYPE_128) &&
		((index != 0) || (type != BMC_128))) {
		return FALSE;
	}

	/* index > 0 ȂAbmc_foffset𒲂ׂ>0Kv */
	if (index > 0) {
		if (bmc_foffset[type][unit][index] == 0) {
			return FALSE;
		}
	}

	/* B77t@C̏ꍇAwb_ǂݍ */
	if (bmc_ready[type][unit] == BMC_TYPE_B77) {
		/* CgveNg͓Őݒ */
		if (!bmc_readhead(type, unit, index)) {
			return FALSE;
		}
	}
	else {
		/* 32KB/128KBt@CȂAt@Cɏ] */
		bmc_writep[type][unit] = bmc_fwritep[type][unit];
	}

	/* fBAꂽꍇAꎞCWFNg */
	if (bmc_media[type][unit] != (BYTE)index) {
		bmc_media[type][unit] = (BYTE)index;
		bmc_teject[type][unit] = FALSE;
	}

	return TRUE;
}

/*
 *	B77t@ĆAfBAі̎擾
 */
static int FASTCALL bmc_chkb77(int type, int unit)
{
	int i;
	int handle;
	int count;
	DWORD offset;
	DWORD len;
	BYTE buf[0x20];

	/*  */
	for (i=0; i<BMC_MEDIAS; i++) {
		bmc_foffset[type][unit][i] = 0;
		bmc_name[type][unit][i][0] = '\0';
	}
	count = 0;
	offset = 0;

	/* t@CI[v */
	handle = file_open(bmc_fname[type][unit], OPEN_R);
	if (handle == -1) {
		return count;
	}

	/* fBA[v */
	while (count < BMC_MEDIAS) {
		/* V[N */
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return count;
		}

		/* ǂݍ */
		if (!file_read(handle, buf, 0x0020)) {
			file_close(handle);
			return count;
		}

		/* ^Cv`FbNB32KB,128KBɂꂼΉ */
		if (((type == BMC_32) && (buf[0x001b] != 0x80)) ||
			((type == BMC_128) && (buf[0x001b] != 0x90))) {
			file_close(handle);
			return count;
		}

		/* ok,t@CAItZbgi[ */
		buf[17] = '\0';
		memcpy(bmc_name[type][unit][count], buf, 17);
		bmc_foffset[type][unit][count] = offset;

		/* next */
		len = 0;
		len |= buf[0x1f];
		len *= 256;
		len |= buf[0x1e];
		len *= 256;
		len |= buf[0x1d];
		len *= 256;
		len |= buf[0x1c];
		offset += len;
		count++;
	}

	/* ő僁fBAɒB */
	file_close(handle);
	return count;
}

/*
 *	t@Cݒ
 */
int FASTCALL bmc_setfile(int type, int unit, char *fname)
{
	BOOL writep;
	int handle;
	DWORD fsize;
	int count;

	ASSERT((type == BMC_32) || (type == BMC_128));
	ASSERT((unit >= 0) && (unit < BMC_UNITS_128));

	/* mbgfBɂꍇ */
	if (fname == NULL) {
		bmc_ready[type][unit] = BMC_TYPE_NOTREADY;
		bmc_fname[type][unit][0] = '\0';
		return 1;
	}

	/* t@CI[vAt@CTCY𒲂ׂ */
	if (strlen(fname) >= sizeof(bmc_fname[type][unit])) {
		bmc_ready[type][unit] = BMC_TYPE_NOTREADY;
		bmc_fname[type][unit][0] = '\0';
		return 1;
	}
	writep = FALSE;
	handle = file_open(fname, OPEN_RW);
	if (handle == -1) {
		handle = file_open(fname, OPEN_R);
		if (handle == -1) {
			bmc_ready[type][unit] = BMC_TYPE_NOTREADY;
			return 0;
		}
		writep = TRUE;
	}
	strcpy(bmc_fname[type][unit], fname);
	fsize = file_getsize(handle);
	file_close(handle);

	/*
	 * 32KBt@C
	 */
	if (type == BMC_32 && fsize == 32768) {
		/* ^CvAݑݒ */
		bmc_ready[type][unit] = BMC_TYPE_32;
		bmc_fwritep[type][unit] = writep;

		/* fBAݒ */
		if (!bmc_setmedia(type, unit, 0)) {
			bmc_ready[type][unit] = BMC_TYPE_NOTREADY;
			bmc_fname[type][unit][0] = '\0';
			return 0;
		}

		/* BꎞCWFNg */
		bmc_teject[type][unit] = FALSE;
		bmc_medias[type][unit] = 1;
		return 1;
	}

	/*
	 * 128KBt@C
	 */
	if (type == BMC_128 && fsize == 131072) {
		/* ^CvAݑݒ */
		bmc_ready[type][unit] = BMC_TYPE_128;
		bmc_fwritep[type][unit] = writep;

		/* fBAݒ */
		if (!bmc_setmedia(type, unit, 0)) {
			bmc_ready[type][unit] = BMC_TYPE_NOTREADY;
			bmc_fname[type][unit][0] = '\0';
			return 0;
		}

		/* BꎞCWFNg */
		bmc_teject[type][unit] = FALSE;
		bmc_medias[type][unit] = 1;
		return 1;
	}

	/*
	 * B77t@C
	 */
	bmc_ready[type][unit] = BMC_TYPE_B77;
	bmc_fwritep[type][unit] = writep;

	/* t@C */
	count = bmc_chkb77(type, unit);
	if (count != 0){
		/* fBAݒ */
		if (bmc_setmedia(type, unit, 0)) {
			/* BꎞCWFNg */
			bmc_teject[type][unit] = FALSE;
			bmc_medias[type][unit] = (BYTE)count;
			return count;
		}
	}

	bmc_ready[type][unit] = BMC_TYPE_NOTREADY;
	bmc_fname[type][unit][0] = '\0';
	return 0;
}

/*-[ BMCR}h ]----------------------------------------------------------*/

/*
 *	Xe[^X쐬
 */
static void FASTCALL bmc_make_stat(int type)
{
	/* Lȃjbg */
	if (bmc_ready[type][bmc_unit[type]] == BMC_TYPE_NOTREADY) {
		bmc_status[type] |= (BYTE)BMC_ST_NOTREADY;
		bmc_status[type] |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg[type] |= (BYTE)BMC_ES_EJECT;
	}
	else {
		bmc_status[type] &= (BYTE)(~BMC_ST_NOTREADY);
	}

	/* ꎞCWFNg */
	if (bmc_teject[type][bmc_unit[type]]) {
		bmc_status[type] |= (BYTE)BMC_ST_NOTREADY;
		bmc_status[type] |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg[type] |= (BYTE)BMC_ES_EJECT;
	}

	/* CgveNg */
	if (bmc_writep[type][bmc_unit[type]]) {
		bmc_status[type] |= (BYTE)BMC_ST_WRITEP;
	}
	else {
		bmc_status[type] &= (BYTE)(~BMC_ST_WRITEP);
	}
}

/*
 *	ou Rg[
 */
static void FASTCALL bmc_initialize(int type)
{
	/*  */
	bmc_datareg[type] = (BYTE)0;
	bmc_status[type] = (BYTE)0;
	bmc_errorreg[type] = (BYTE)0;
	bmc_pagereg[type] = 0;
	bmc_countreg[type] = 0;
	bmc_totalcnt[type] = 0;
	bmc_nowcnt[type] = 0;

	/* Xe[^Xݒ肷 */
	bmc_status[type] = (BYTE)BMC_ST_CME;

	/* ANZX(READY) */
	bmc_access[type][bmc_unit[type]] = (BYTE)BMC_ACCESS_READY;

	bmc_unit[type] = 0;
	bmc_dataptr[type] = NULL;
}

/*
 *	READ/WRITE Tu
 */
static BOOL FASTCALL bmc_rw_sub(int type)
{
	bmc_status[type] = (BYTE)0;
	bmc_errorreg[type] = (BYTE)0;

	/* y[WAhX`FbN */
	if (type == BMC_32) {
		if ((bmc_unit[type] >= BMC_UNITS_32) ||
			(bmc_pagereg[type] >= (bmc_max_addr[type] + 1) * 2)) {
			bmc_status[type] |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
			bmc_errorreg[type] |= (BYTE)BMC_ES_PAGEOVER;
			return FALSE;
		}
	}
	else {
		if ((bmc_unit[type] >= BMC_UNITS_128) ||
			((bmc_pagereg[type] & 0x3fff) >= (bmc_max_addr[type] + 1))) {
			bmc_status[type] |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
			bmc_errorreg[type] |= (BYTE)BMC_ES_PAGEOVER;
			return FALSE;
		}
	}

	/* NOT READY`FbN */
	if ((bmc_ready[type][bmc_unit[type]] == BMC_TYPE_NOTREADY) ||
		bmc_teject[type][bmc_unit[type]]) {
		bmc_make_stat(type);

		/* Xe[^Xݒ肷 */
		bmc_status[type] &= (BYTE)(~BMC_ST_BUSY);
		bmc_status[type] |= (BYTE)BMC_ST_CME;
		return FALSE;
	}

	return TRUE;
}

/*
 *	y[Wǂݍ݊Jn
 */
static void FASTCALL bmc_read_data(int type)
{
	/* {`FbN */
	if (!bmc_rw_sub(type)) {
		return;
	}

	/* JE^ݒ */
	bmc_totalcnt[type] = bmc_page_size[type];
	bmc_nowcnt[type] = 0;

	/* Xe[^Xݒ */
	bmc_status[type] |= (BYTE)BMC_ST_BUSY;
	bmc_status[type] &= (BYTE)(~BMC_ST_CME);

	/* f[^|C^ݒ */
	bmc_dataptr[type] = &bmc_buffer[type][0];
	bmc_offset[type] = 0;

	/* ANZX(READ) */
	bmc_access[type][bmc_unit[type]] = (BYTE)BMC_ACCESS_READ;

	/* f[^obt@ǂݍ */
	if (!bmc_readbuf(type)) {
		bmc_status[type] |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg[type] |= (BYTE)BMC_ES_NOMAKER;
		return;
	}

	/* ŏ̃f[^ݒ */
	bmc_datareg[type] = bmc_dataptr[type][0];

#ifdef FDDSND
	if (!bmc_wait) {
		bmc_status[type] |= (BYTE)BMC_ST_RDA;
	}
#else
	bmc_status[type] |= (BYTE)BMC_ST_RDA;
#endif
}

/*
 *	y[Wݍ݊Jn
 */
static void FASTCALL bmc_write_data(int type)
{
	/* {`FbN */
	if (!bmc_rw_sub(type)) {
		return;
	}

	/* JE^ݒ */
	bmc_totalcnt[type] = bmc_page_size[type];
	bmc_nowcnt[type] = 0;

	/* Xe[^Xݒ */
	bmc_status[type] |= (BYTE)BMC_ST_BUSY;
	bmc_status[type] &= (BYTE)(~BMC_ST_CME);

	/* WRITE PROTECT`FbN */
	if (bmc_writep[type][bmc_unit[type]] != 0) {
		bmc_make_stat(type);
		bmc_status[type] &= (BYTE)(~BMC_ST_BUSY);
		bmc_status[type] |= (BYTE)BMC_ST_CME;
		return;
	}

	/* f[^|C^ݒ */
	bmc_dataptr[type] = &bmc_buffer[type][0];
	bmc_offset[type] = 0;

	/* ANZX(WRITE) */
	bmc_access[type][bmc_unit[type]] = (BYTE)BMC_ACCESS_WRITE;

#ifdef FDDSND
	if (!bmc_wait) {
		bmc_status[type] |= (BYTE)BMC_ST_TDRA;
	}
#else
	bmc_status[type] |= (BYTE)BMC_ST_TDRA;
#endif
}

/*
 *	R}h
 */
static void FASTCALL bmc_process_cmd(int type)
{
	/* f[^]sĂ΁A~߂ */
	bmc_dataptr[type] = NULL;

#ifdef FDDSND
	/* EFCgtOݒ */
	bmc_wait = fdc_waitmode;
#endif

	/*  */
	switch (bmc_command[type]) {
		/* bubble read */
		case 0x01:
			bmc_read_data(type);
			break;
		/* bubble write */
		case 0x02:
			bmc_write_data(type);
			break;
		/* initialize */
		case 0x04:
		case 0x0f:
			bmc_initialize(type);
			break;
		/* ȊO */
		default:
			/* Xe[^Xݒ肷 */
			bmc_status[type] = (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
			bmc_errorreg[type] = (BYTE)BMC_ES_UNDEF;
			ASSERT(FALSE);
			break;
	}
}

/*
 *	ou Rg[
 *	PoCgǂݏo
 */
BOOL FASTCALL bmc_readb(WORD addr, BYTE *dat)
{
	/* 32KB */
	int type = BMC_32;

	if (((addr & 0xfff8) != 0xfd10) && ((addr & 0xfff8) != 0xfdd0)) {
		return FALSE;
	}

	/* 128KB */
	if ((addr & 0xfff8) == 0xfdd0) {
		type = BMC_128;
	}

	/* LE`FbN */
	if (!bmc_enable[type]) {
		return FALSE;
	}

	/* 32KBFM-8[h */
	if ((fm_subtype != FMSUB_FM8) && (type == BMC_32)) {
		/* FM-7/77[hłOPNɂ͗LɂĂ݂ */
		if (opn_enable) {
			return FALSE;
		}
	}

	bmc_use[type] = TRUE;

	switch (addr) {
		/* f[^WX^(BDATA) */
		case 0xfd10:
		case 0xfdd0:
			*dat = bmc_datareg[type];
			/* JE^ */
			if (bmc_dataptr[type]) {
				bmc_nowcnt[type]++;
				if (bmc_nowcnt[type] == bmc_totalcnt[type]) {
					bmc_status[type] &= (BYTE)(~BMC_ST_BUSY);
					bmc_status[type] &= (BYTE)(~BMC_ST_RDA);
					bmc_status[type] |= (BYTE)BMC_ST_CME;

					bmc_countreg[type]--;
					if ((bmc_countreg[type] > 0) &&
						((bmc_pagereg[type] & bmc_max_addr[type]) < bmc_max_addr[type])) {
						/* }`y[W */
						bmc_status[type] |= (BYTE)BMC_ST_BUSY;
						bmc_status[type] &= (BYTE)(~BMC_ST_CME);
						bmc_pagereg[type]++;
						bmc_read_data(type);
						bmc_nowcnt[type] = 0;
						return TRUE;
					}
					/* VOy[W or }`y[WI */
					bmc_dataptr[type] = NULL;

					/* ANZX(READY) */
					bmc_access[type][bmc_unit[type]] = (BYTE)BMC_ACCESS_READY;
				}
				else {
					bmc_datareg[type] = bmc_dataptr[type][bmc_nowcnt[type]];
#ifdef FDDSND
					if (bmc_wait) {
						bmc_status[type] &= (BYTE)(~BMC_ST_RDA);
					}
#endif
				}
			}
			return TRUE;

		/* Xe[^XWX^(BSTAT) */
		case 0xfd12:
		case 0xfdd2:
			bmc_make_stat(type);
			*dat = bmc_status[type];
			/* BUSY */
			if ((bmc_status[type] & BMC_ST_BUSY) &&
				(bmc_dataptr[type] == NULL)) {
				/* BUSYtO𗎂Ƃ */
				bmc_status[type] &= (BYTE)(~BMC_ST_BUSY);
				bmc_access[type][bmc_unit[type]] = (BYTE)BMC_ACCESS_READY;
			}
			/* CAN READ */
			if ((bmc_command[type] == 0x01) &&
				(bmc_dataptr[type]) &&
				!(bmc_status[type] & BMC_ST_RDA)) {
				/* CAN READtO𗧂Ă */
				bmc_status[type] |= (BYTE)BMC_ST_RDA;
			}
			/* CAN WRITE */
			if ((bmc_command[type] == 0x02) &&
				(bmc_dataptr[type]) &&
				!(bmc_status[type] & BMC_ST_TDRA)) {
				/* CAN WRITEtO𗧂Ă */
				bmc_status[type] |= (BYTE)BMC_ST_TDRA;
			}
			return TRUE;

		/* G[Xe[^XWX^(BERRST) */
		case 0xfd13:
		case 0xfdd3:
			*dat = bmc_errorreg[type];
			return TRUE;
	}

	return FALSE;
}

/*
 *	ou Rg[
 *	PoCg
 */
BOOL FASTCALL bmc_writeb(WORD addr, BYTE dat)
{
	/* 32KB */
	int type = BMC_32;

	if (((addr & 0xfff8) != 0xfd10) && ((addr & 0xfff8) != 0xfdd0)) {
		return FALSE;
	}

	/* 128KB */
	if ((addr & 0xfff8) == 0xfdd0) {
		type = BMC_128;
	}

	/* LE`FbN */
	if (!bmc_enable[type]) {
		return FALSE;
	}

	/* 32KBFM-8[h */
	if ((fm_subtype != FMSUB_FM8) && (type == BMC_32)) {
		/* FM-7/77[hłOPNɂ͗LɂĂ݂ */
		if (opn_enable) {
			return FALSE;
		}
	}

	bmc_use[type] = TRUE;

	switch (addr) {
		/* f[^WX^(BDATA) */
		case 0xfd10:
		case 0xfdd0:
			bmc_datareg[type] = dat;
			/* JE^ */
			if (bmc_dataptr[type]) {
				bmc_dataptr[type][bmc_nowcnt[type]] = bmc_datareg[type];
				bmc_nowcnt[type]++;
				if (bmc_nowcnt[type] == bmc_totalcnt[type]) {
					bmc_status[type] &= (BYTE)(~BMC_ST_BUSY);
					bmc_status[type] &= (BYTE)(~BMC_ST_TDRA);
					bmc_status[type] |= (BYTE)BMC_ST_CME;

					/* Cgy[W */
					if (!bmc_write_page(type)) {
						bmc_status[type] |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
						bmc_errorreg[type] |= (BYTE)BMC_ES_NOMAKER;
					}

					bmc_countreg[type]--;
					if ((bmc_countreg[type] > 0) &&
						((bmc_pagereg[type] & bmc_max_addr[type]) < bmc_max_addr[type])) {
						/* }`y[W */
						bmc_status[type] |= (BYTE)BMC_ST_BUSY;
#ifdef FDDSND
						if (!bmc_wait) {
							bmc_status[type] |= (BYTE)BMC_ST_TDRA;
						}
#else
						bmc_status[type] |= (BYTE)BMC_ST_TDRA;
#endif
						bmc_status[type] &= (BYTE)(~BMC_ST_CME);
						bmc_pagereg[type]++;
						bmc_nowcnt[type] = 0;
						return TRUE;
					}
					/* VOy[W or }`y[WI */
					bmc_dataptr[type] = NULL;

					/* ANZX(READY) */
					bmc_access[type][bmc_unit[type]] = (BYTE)BMC_ACCESS_READY;
				}
				else {
#ifdef FDDSND
					if (bmc_wait) {
						bmc_status[type] &= (BYTE)(~BMC_ST_TDRA);
					}
#endif
				}
			}
			return TRUE;

		/* R}hWX^(BCMD) */
		case 0xfd11:
		case 0xfdd1:
			bmc_command[type] = (BYTE)(dat & 0x0f);
			bmc_process_cmd(type);
			return TRUE;

		/* y[WAhXWX^H(BPGADH) */
		case 0xfd14:
		case 0xfdd4:
			bmc_pagereg[type] &= 0x00ff;
			bmc_pagereg[type] |= (WORD)(dat << 8);

			/* y[W烆jbgZo */
			if (type == BMC_32) {
				bmc_unit[type] = (BYTE)(bmc_pagereg[type] >> 10);
			}
			else {
				bmc_unit[type] = (BYTE)(bmc_pagereg[type] >> 14);
			}
			return TRUE;

		/* y[WAhXWX^L(BPGADL) */
		case 0xfd15:
		case 0xfdd5:
			bmc_pagereg[type] &= 0xff00;
			bmc_pagereg[type] |= dat;
			return TRUE;

		/* y[WJEgWX^H(BPGCTH) */
		case 0xfd16:
		case 0xfdd6:
			bmc_countreg[type] &= 0x00ff;
			bmc_countreg[type] |= (WORD)(dat << 8);
			return TRUE;

		/* y[WJEgWX^L(BPGCTL) */
		case 0xfd17:
		case 0xfdd7:
			bmc_countreg[type] &= 0xff00;
			bmc_countreg[type] |= dat;
			return TRUE;
	}

	return FALSE;
}

/*
 *	ou Rg[
 *	Z[u
 */
BOOL FASTCALL bmc_save(int fileh)
{
	int i, j, bmc_units;
	DWORD size;

	for (i=0; i<2; i++) {
		if (i == BMC_32) {
			bmc_units = BMC_UNITS_32;
			size = (DWORD)bmc_page_size[i];
		}
		else {
			bmc_units = BMC_UNITS_128;
			size = (DWORD)bmc_page_size[i];
		}

		/* t@C֌WɎĂ */
		for (j=0; j<bmc_units; j++) {
			if (!file_byte_write(fileh, bmc_ready[i][j])) {
				return FALSE;
			}
		}
		for (j=0; j<bmc_units; j++) {
			if (!file_write(fileh, (BYTE*)bmc_fname[i][j], 256 + 1)) {
				return FALSE;
			}
		}
		for (j=0; j<bmc_units; j++) {
			if (!file_byte_write(fileh, bmc_media[i][j])) {
				return FALSE;
			}
		}

		for (j=0; j<bmc_units; j++) {
			if (!file_bool_write(fileh, bmc_teject[i][j])) {
				return FALSE;
			}
		}

		/* t@CXe[^X */
		if (!file_write(fileh, bmc_buffer[i], size)) {
			return FALSE;
		}

		/* bmc_dataptr͊Ɉˑf[^|C^ */
		if (!bmc_dataptr[i]) {
			if (!file_word_write(fileh, (WORD)size)) {
				return FALSE;
			}
		}
		else {
			if (!file_word_write(fileh, (WORD)(bmc_dataptr[i] - &bmc_buffer[i][0]))) {
				return FALSE;
			}
		}

		if (!file_dword_write(fileh, bmc_offset[i])) {
			return FALSE;
		}

		/* I/O */
		if (!file_byte_write(fileh, bmc_datareg[i])) {
			return FALSE;
		}
		if (!file_byte_write(fileh, bmc_command[i])) {
			return FALSE;
		}
		if (!file_byte_write(fileh, bmc_status[i])) {
			return FALSE;
		}
		if (!file_byte_write(fileh, bmc_errorreg[i])) {
			return FALSE;
		}
		if (!file_word_write(fileh, bmc_pagereg[i])) {
			return FALSE;
		}
		if (!file_word_write(fileh, bmc_countreg[i])) {
			return FALSE;
		}

		/* ̑ */
		if (!file_word_write(fileh, bmc_totalcnt[i])) {
			return FALSE;
		}
		if (!file_word_write(fileh, bmc_nowcnt[i])) {
			return FALSE;
		}
		for (j=0; j<bmc_units; j++) {
			if (!file_byte_write(fileh, bmc_access[i][j])) {
				return FALSE;
			}
		}

		/* LE`FbN */
		if (!file_bool_write(fileh, bmc_enable[i])) {
			return FALSE;
		}
		if (!file_bool_write(fileh, bmc_use[i])) {
			return FALSE;
		}
	}

#ifdef FDDSND
	if (!file_bool_write(fileh, bmc_wait)) {
		return FALSE;
	}
#else
	if (!file_bool_write(fileh, FALSE)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	ou Rg[
 *	[h
 */
BOOL FASTCALL bmc_load(int fileh, int ver)
{
	int i, j, bmc_units, bmc_types;
	BYTE ready[2][BMC_UNITS_128];
	char fname[2][BMC_UNITS_128][256 + 1];
	BYTE media[2][BMC_UNITS_128];
	WORD offset;
	DWORD size;
#ifndef FDDSND
	BOOL tmp;
#endif
	int pathlen;

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	if ((ver < 307) ||
		((ver >= 400) && (ver < 403))) {
		bmc_init();
		bmc_reset();
		return TRUE;
	}

	/* 128Kou̗L */
	if (ver >= 400) {
		bmc_types = 2;
	}
	else {
		bmc_types = 1;
	}

	/* t@C̍ő啶 */
	if (ver != 403) {
		pathlen = 256;
	}
	else {
		pathlen = 128;
	}

	for (i=0; i<bmc_types; i++) {
		if (i == BMC_32) {
			bmc_units = BMC_UNITS_32;
			size = (DWORD)bmc_page_size[i];
		}
		else {
			bmc_units = BMC_UNITS_128;
			size = (DWORD)bmc_page_size[i];
		}

		/* t@C֌WɎĂ */
		for (j=0; j<bmc_units; j++) {
			if (!file_byte_read(fileh, &ready[i][j])) {
				return FALSE;
			}
		}
		for (j=0; j<bmc_units; j++) {
			if (!file_read(fileh, (BYTE*)fname[i][j], pathlen + 1)) {
				return FALSE;
			}
		}
		if (((ver >= 308) && (ver <= 399)) ||
			(ver >= 409)) {
			for (j=0; j<bmc_units; j++) {
				if (!file_byte_read(fileh, &media[i][j])) {
					return FALSE;
				}
			}
		}
		else {
			for (j=0; j<bmc_units; j++) {
				media[i][j] = 0;
			}
		}

		/* ă}Eg݂ */
		for (j=0; j<bmc_units; j++) {
			bmc_setfile(i, j, NULL);
			if (ready[i][j] != BMC_TYPE_NOTREADY) {
				bmc_setfile(i, j, fname[i][j]);
				if (bmc_ready[i][j] != BMC_TYPE_NOTREADY) {
					if (bmc_medias[i][j] >= (media[i][j] + 1)) {
						bmc_setmedia(i, j, media[i][j]);
					}
				}
			}
		}

		for (j=0; j<bmc_units; j++) {
			if (!file_bool_read(fileh, &bmc_teject[i][j])) {
				return FALSE;
			}
		}

		/* t@CXe[^X */
		if (!file_read(fileh, bmc_buffer[i], size)) {
			return FALSE;
		}

		/* bmc_dataptr͊Ɉˑf[^|C^ */
		if (!file_word_read(fileh, &offset)) {
			return FALSE;
		}
		if (offset >= (WORD)size) {
			bmc_dataptr[i] = NULL;
		}
		else {
			bmc_dataptr[i] = &bmc_buffer[i][offset];
		}

		if (!file_dword_read(fileh, &bmc_offset[i])) {
			return FALSE;
		}

		/* I/O */
		if (!file_byte_read(fileh, &bmc_datareg[i])) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &bmc_command[i])) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &bmc_status[i])) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &bmc_errorreg[i])) {
			return FALSE;
		}
		if (!file_word_read(fileh, &bmc_pagereg[i])) {
			return FALSE;
		}
		if (!file_word_read(fileh, &bmc_countreg[i])) {
			return FALSE;
		}

		/* ̑ */
		if (!file_word_read(fileh, &bmc_totalcnt[i])) {
			return FALSE;
		}
		if (!file_word_read(fileh, &bmc_nowcnt[i])) {
			return FALSE;
		}
		for (j=0; j<bmc_units; j++) {
			if (!file_byte_read(fileh, &bmc_access[i][j])) {
				return FALSE;
			}
		}

		/* y[W烆jbgZo */
		if (i == BMC_32) {
			bmc_unit[i] = (BYTE)(bmc_pagereg[i] >> 10);
		}
		else {
			bmc_unit[i] = (BYTE)(bmc_pagereg[i] >> 14);
		}

		/* LE`FbN */
		if (!file_bool_read(fileh, &bmc_enable[i])) {
			return FALSE;
		}

		if (((ver >= 307) && (ver <= 399)) ||
			(ver >= 405)) {
			if (!file_bool_read(fileh, &bmc_use[i])) {
				return FALSE;
			}
		}
		else {
			bmc_use[i] = bmc_enable[i];
		}
	}

	if (((ver >= 307) && (ver <= 399)) ||
		(ver >= 404)) {
#ifdef FDDSND
		if (!file_bool_read(fileh, &bmc_wait)) {
			return FALSE;
		}
#else
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
#endif
	}
#ifdef FDDSND
	else {
		/* EFCg[hFALSE */
		bmc_wait = FALSE;
	}
#endif

	return TRUE;
}
#else
/*
 *	O[o [N
 */

/* WX^ */
BYTE bmc_datareg;						/* f[^WX^ */
BYTE bmc_command;						/* R}hWX^ */
BYTE bmc_status;						/* Xe[^XWX^ */
BYTE bmc_errorreg;						/* G[Xe[^XWX^ */
WORD bmc_pagereg;						/* y[WWX^ */
WORD bmc_countreg;						/* y[WJEgWX^ */

/* O[N */
WORD bmc_totalcnt;						/* g[^JE^ */
WORD bmc_nowcnt;						/* JgJE^ */
BYTE bmc_unit;							/* jbg */
BYTE bmc_ready[BMC_UNITS_32];			/* fB */
BOOL bmc_teject[BMC_UNITS_32];			/* ꎞCWFNg */
BOOL bmc_writep[BMC_UNITS_32];			/* ݋֎~ */

char bmc_fname[BMC_UNITS_32][256+1];	/* t@C */
char bmc_name[BMC_UNITS_32][BMC_MEDIAS][17];	/* C[W */
BOOL bmc_fwritep[BMC_UNITS_32];			/* CgveNg */
BYTE bmc_header[BMC_UNITS_32][0x20];	/* B77t@Cwb_ */
BYTE bmc_medias[BMC_UNITS_32];			/* fBA */
BYTE bmc_media[BMC_UNITS_32];			/* fBAZNg */
BYTE bmc_access[BMC_UNITS_32];			/* ANZXLED */

BOOL bmc_enable;						/* LEtO */
BOOL bmc_use;							/* gptO */

/*
 *	X^eBbN [N
 */
static BYTE bmc_buffer[BMC_PSIZE_32];	/* 32bytef[^obt@ */
static BYTE *bmc_dataptr;				/* f[^|C^ */
static DWORD bmc_offset;				/* ItZbg */
static DWORD bmc_foffset[BMC_UNITS_32][BMC_MEDIAS];
static DWORD bmc_fsize[BMC_UNITS_32];
#ifdef FDDSND
static BOOL bmc_wait;					/* EFCg[hstO */
#endif

/*
 *	ou Rg[
 *	
 */
BOOL FASTCALL bmc_init(void)
{
	/* t@C֌WZbg */
	bmc_totalcnt = 0;
	bmc_nowcnt = 0;

	memset(bmc_ready, BMC_TYPE_NOTREADY, sizeof(bmc_ready));
	memset(bmc_teject, FALSE, sizeof(bmc_teject));
	memset(bmc_writep, FALSE, sizeof(bmc_writep));
	memset(bmc_fname, 0, sizeof(bmc_fname));
	memset(bmc_fwritep, FALSE, sizeof(bmc_fwritep));
	memset(bmc_medias, 0, sizeof(bmc_medias));

	/* t@CItZbgSăNA */
	memset(bmc_foffset, 0, sizeof(bmc_foffset));

	/* ftHg͖ */
	bmc_enable = FALSE;

	/* EFCg}[htO */
#ifdef FDDSND
	bmc_wait = FALSE;
#endif

	return TRUE;
}

/*
 *	ou Rg[
 *	N[Abv
 */
void FASTCALL bmc_cleanup(void)
{
	/* t@C֌WZbg */
	memset(bmc_ready, 0, sizeof(bmc_ready));
}

/*
 *	ou Rg[
 *	Zbg
 */
void FASTCALL bmc_reset(void)
{
	/* WX^Zbg */
	bmc_datareg = 0;
	bmc_command = 0xff;
	bmc_status = 0;
	bmc_errorreg = 0;
	bmc_pagereg = 0;
	bmc_countreg = 0;

	memset(bmc_access, 0, sizeof(bmc_access));
	bmc_unit = 0;
	bmc_dataptr = 0;
	bmc_offset = 0;

	bmc_use = FALSE;
}

/*-[ t@CǗ ]---------------------------------------------------------*/

/*
 *	y[WݏI
 */
static BOOL FASTCALL bmc_write_page(void)
{
	DWORD offset;
	int handle;

	/* assert */
	ASSERT(bmc_ready[bmc_unit] != BMC_TYPE_NOTREADY);
	ASSERT(bmc_dataptr);
	ASSERT(bmc_totalcnt > 0);

	/* ItZbgZo */
	offset = (DWORD)((bmc_pagereg & BMC_MAXADDR_32) * BMC_PSIZE_32);
	if (bmc_ready[bmc_unit] == BMC_TYPE_B77) {
		if (bmc_fsize[bmc_unit] < offset + BMC_PSIZE_32 + 0x0020) {
			return FALSE;
		}
		offset += *(DWORD *)(&bmc_header[bmc_unit][0x0014]);
	}

	/*  */
	handle = file_open(bmc_fname[bmc_unit], OPEN_RW);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_write(handle, bmc_dataptr, bmc_totalcnt)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	return TRUE;
}

/*
 *	y[Wǂݍ
 */
static BOOL FASTCALL bmc_readbuf(void)
{
	int handle;
	DWORD offset;

	/* y[WAhX`FbN */
	if (bmc_unit >= BMC_UNITS_32) {
		memset(bmc_buffer, 0, BMC_PSIZE_32);
		return FALSE;
	}

	/* fB`FbN */
	if (bmc_ready[bmc_unit] == BMC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* ItZbgZo */
	offset = (DWORD)(bmc_pagereg & BMC_MAXADDR_32);
	offset *= BMC_PSIZE_32;
	if (bmc_ready[bmc_unit] == BMC_TYPE_B77) {
		if (bmc_fsize[bmc_unit] < offset + BMC_PSIZE_32 + 0x0020) {
			return FALSE;
		}
		offset += *(DWORD *)(&bmc_header[bmc_unit][0x0014]);
	}
	bmc_offset = offset;

	/* ǂݍ */
	memset(bmc_buffer, 0, BMC_PSIZE_32);
	handle = file_open(bmc_fname[bmc_unit], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	file_read(handle, bmc_buffer, BMC_PSIZE_32);
	file_close(handle);
	return TRUE;
}

/*
 *	B77t@C wb_ǂݍ
 */
static BOOL FASTCALL bmc_readhead(int unit, int index)
{
	DWORD offset;
	DWORD temp;
	int handle;

	/* assert */
	ASSERT((unit >= 0) && (unit < BMC_UNITS_32));
	ASSERT((index >= 0) && (index < BMC_MEDIAS));
	ASSERT(bmc_ready[unit] == BMC_TYPE_B77);

	/* ItZbg */
	offset = bmc_foffset[unit][index];

	/* V[NAǂݍ */
	handle = file_open(bmc_fname[unit], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_read(handle, bmc_header[unit], 0x20)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	/* JZbgTCY */
	temp = 0;
	temp |= bmc_header[unit][0x001c + 3];
	temp *= 256;
	temp |= bmc_header[unit][0x001c + 2];
	temp *= 256;
	temp |= bmc_header[unit][0x001c + 1];
	temp *= 256;
	temp |= bmc_header[unit][0x001c + 0];
	bmc_fsize[unit] = temp;

	/* ^Cv`FbN */
	if (bmc_header[unit][0x001b] != 0x80) {
		/* 32KBłȂ */
		return FALSE;
	}

	/* CgveNgݒ */
	if (bmc_fwritep[unit]) {
		bmc_writep[unit] = TRUE;
	}
	else {
		if (bmc_header[unit][0x001a] & 0x10) {
			bmc_writep[unit] = TRUE;
		}
		else {
			bmc_writep[unit] = FALSE;
		}
	}

	/* ItZbg */
	*(DWORD *)(&bmc_header[unit][0x0014]) = offset + 0x0020;

	return TRUE;
}

/*
 *	݂̃fBÃCgveNg؂ւ
 */
BOOL FASTCALL bmc_setwritep(int unit, BOOL writep)
{
	BYTE header[0x2b0];
	DWORD offset;
	int handle;

	/* assert */
	ASSERT((unit >= 0) && (unit < BMC_UNITS_32));
	ASSERT((writep == TRUE) || (writep == FALSE));

	/* fBłȂ΂ȂȂ */
	if (bmc_ready[unit] == BMC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* t@CݕsȂ_ */
	if (bmc_fwritep[unit]) {
		return FALSE;
	}

	if (bmc_ready[unit] == BMC_TYPE_B77) {
		offset = bmc_foffset[unit][bmc_media[unit]];
		handle = file_open(bmc_fname[unit], OPEN_RW);
		if (handle == -1) {
			return FALSE;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_read(handle, header, 0x20)) {
			file_close(handle);
			return FALSE;
		}
		if (writep) {
			header[0x001a] |= 0x10;
		}
		else {
			header[0x001a] &= ~0x10;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_write(handle, header, 0x20)) {
			file_close(handle);
			return FALSE;
		}

		file_close(handle);
	}

	/*  */
	bmc_writep[unit] = writep;

	return TRUE;
}


/*
 *	fBAԍݒ
 */
BOOL FASTCALL bmc_setmedia(int unit, int index)
{
	/* assert */
	ASSERT((unit >= 0) && (unit <= 1));
	ASSERT((index >= 0) && (index < BMC_MEDIAS));

	/* fBԂ */
	if (bmc_ready[unit] == BMC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* 32KBt@C̏ꍇAindex = 0 */
	if ((bmc_ready[unit] == BMC_TYPE_32) && (index != 0)) {
		return FALSE;
	}

	/* index > 0 ȂAbmc_foffset𒲂ׂ>0Kv */
	if (index > 0) {
		if (bmc_foffset[unit][index] == 0) {
			return FALSE;
		}
	}

	/* B77t@C̏ꍇAwb_ǂݍ */
	if (bmc_ready[unit] == BMC_TYPE_B77) {
		/* CgveNg͓Őݒ */
		if (!bmc_readhead(unit, index)) {
			return FALSE;
		}
	}
	else {
		/* 32KBt@CȂAt@Cɏ] */
		bmc_writep[unit] = bmc_fwritep[unit];
	}

	/* fBAꂽꍇAꎞCWFNg */
	if (bmc_media[unit] != (BYTE)index) {
		bmc_media[unit] = (BYTE)index;
		bmc_teject[unit] = FALSE;
	}

	return TRUE;
}

/*
 *	B77t@ĆAfBAі̎擾
 */
static int FASTCALL bmc_chkb77(int unit)
{
	int i;
	int handle;
	int count;
	DWORD offset;
	DWORD len;
	BYTE buf[0x20];

	/*  */
	for (i=0; i<BMC_MEDIAS; i++) {
		bmc_foffset[unit][i] = 0;
		bmc_name[unit][i][0] = '\0';
	}
	count = 0;
	offset = 0;

	/* t@CI[v */
	handle = file_open(bmc_fname[unit], OPEN_R);
	if (handle == -1) {
		return count;
	}

	/* fBA[v */
	while (count < BMC_MEDIAS) {
		/* V[N */
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return count;
		}

		/* ǂݍ */
		if (!file_read(handle, buf, 0x0020)) {
			file_close(handle);
			return count;
		}

		/* ^Cv`FbNB32KB̂ݑΉ */
		if (buf[0x001b] != 0x80) {
			file_close(handle);
			return count;
		}

		/* ok,t@CAItZbgi[ */
		buf[17] = '\0';
		memcpy(bmc_name[unit][count], buf, 17);
		bmc_foffset[unit][count] = offset;

		/* next */
		len = 0;
		len |= buf[0x1f];
		len *= 256;
		len |= buf[0x1e];
		len *= 256;
		len |= buf[0x1d];
		len *= 256;
		len |= buf[0x1c];
		offset += len;
		count++;
	}

	/* ő僁fBAɒB */
	file_close(handle);
	return count;
}

/*
 *	t@Cݒ
 */
int FASTCALL bmc_setfile(int unit, char *fname)
{
	BOOL writep;
	int handle;
	DWORD fsize;
	int count;

	ASSERT((unit >= 0) && (unit < BMC_UNITS_32));

	/* mbgfBɂꍇ */
	if (fname == NULL) {
		bmc_ready[unit] = BMC_TYPE_NOTREADY;
		bmc_fname[unit][0] = '\0';
		return 1;
	}

	/* t@CI[vAt@CTCY𒲂ׂ */
	if (strlen(fname) >= sizeof(bmc_fname[unit])) {
		bmc_ready[unit] = BMC_TYPE_NOTREADY;
		bmc_fname[unit][0] = '\0';
		return 1;
	}
	writep = FALSE;
	handle = file_open(fname, OPEN_RW);
	if (handle == -1) {
		handle = file_open(fname, OPEN_R);
		if (handle == -1) {
			bmc_ready[unit] = BMC_TYPE_NOTREADY;
			return 0;
		}
		writep = TRUE;
	}
	strcpy(bmc_fname[unit], fname);
	fsize = file_getsize(handle);
	file_close(handle);

	/*
	 * 32KBt@C
	 */
	if (fsize == 32768) {
		/* ^CvAݑݒ */
		bmc_ready[unit] = BMC_TYPE_32;
		bmc_fwritep[unit] = writep;

		/* fBAݒ */
		if (!bmc_setmedia(unit, 0)) {
			bmc_ready[unit] = BMC_TYPE_NOTREADY;
			bmc_fname[unit][0] = '\0';
			return 0;
		}

		/* BꎞCWFNg */
		bmc_teject[unit] = FALSE;
		bmc_medias[unit] = 1;
		return 1;
	}

	/*
	 * B77t@C
	 */
	bmc_ready[unit] = BMC_TYPE_B77;
	bmc_fwritep[unit] = writep;

	/* t@C */
	count = bmc_chkb77(unit);
	if (count != 0){
		/* fBAݒ */
		if (bmc_setmedia(unit, 0)) {
			/* BꎞCWFNg */
			bmc_teject[unit] = FALSE;
			bmc_medias[unit] = (BYTE)count;
			return count;
		}
	}

	bmc_ready[unit] = BMC_TYPE_NOTREADY;
	bmc_fname[unit][0] = '\0';
	return 0;
}

/*-[ BMCR}h ]----------------------------------------------------------*/

/*
 *	Xe[^X쐬
 */
static void FASTCALL bmc_make_stat(void)
{
	/* Lȃjbg */
	if (bmc_ready[bmc_unit] == BMC_TYPE_NOTREADY) {
		bmc_status |= (BYTE)BMC_ST_NOTREADY;
		bmc_status |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg |= (BYTE)BMC_ES_EJECT;
	}
	else {
		bmc_status &= (BYTE)(~BMC_ST_NOTREADY);
	}

	/* ꎞCWFNg */
	if (bmc_teject[bmc_unit]) {
		bmc_status |= (BYTE)BMC_ST_NOTREADY;
		bmc_status |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg |= (BYTE)BMC_ES_EJECT;
	}

	/* CgveNg */
	if (bmc_writep[bmc_unit]) {
		bmc_status |= (BYTE)BMC_ST_WRITEP;
	}
	else {
		bmc_status &= (BYTE)(~BMC_ST_WRITEP);
	}
}

/*
 *	ou Rg[
 */
static void FASTCALL bmc_initialize(void)
{
	/*  */
	bmc_datareg = (BYTE)0;
	bmc_status = (BYTE)0;
	bmc_errorreg = (BYTE)0;
	bmc_pagereg = 0;
	bmc_countreg = 0;
	bmc_totalcnt = 0;
	bmc_nowcnt = 0;

	/* Xe[^Xݒ肷 */
	bmc_status = (BYTE)BMC_ST_CME;

	/* ANZX(READY) */
	bmc_access[bmc_unit] = (BYTE)BMC_ACCESS_READY;

	bmc_unit = 0;
	bmc_dataptr = NULL;
}

/*
 *	READ/WRITE Tu
 */
static BOOL FASTCALL bmc_rw_sub(void)
{
	bmc_status = (BYTE)0;
	bmc_errorreg = (BYTE)0;

	/* y[WAhX`FbN */
	if ((bmc_unit >= BMC_UNITS_32) ||
		(bmc_pagereg >= (BMC_MAXADDR_32 + 1) * 2)) {
		bmc_status |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg |= (BYTE)BMC_ES_PAGEOVER;
		return FALSE;
	}

	/* NOT READY`FbN */
	if ((bmc_ready[bmc_unit] == BMC_TYPE_NOTREADY) ||
		bmc_teject[bmc_unit]) {
		bmc_make_stat();

		/* Xe[^Xݒ肷 */
		bmc_status &= (BYTE)(~BMC_ST_BUSY);
		bmc_status |= (BYTE)BMC_ST_CME;
		return FALSE;
	}

	return TRUE;
}

/*
 *	y[Wǂݍ݊Jn
 */
static void FASTCALL bmc_read_data(void)
{
	/* {`FbN */
	if (!bmc_rw_sub()) {
		return;
	}

	/* JE^ݒ */
	bmc_totalcnt = BMC_PSIZE_32;
	bmc_nowcnt = 0;

	/* Xe[^Xݒ */
	bmc_status |= (BYTE)BMC_ST_BUSY;
	bmc_status &= (BYTE)(~BMC_ST_CME);

	/* f[^|C^ݒ */
	bmc_dataptr = bmc_buffer;
	bmc_offset = 0;

	/* ANZX(READ) */
	bmc_access[bmc_unit] = (BYTE)BMC_ACCESS_READ;

	/* f[^obt@ǂݍ */
	if (!bmc_readbuf()) {
		bmc_status |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
		bmc_errorreg |= (BYTE)BMC_ES_NOMAKER;
		return;
	}

	/* ŏ̃f[^ݒ */
	bmc_datareg = bmc_dataptr[0];

#ifdef FDDSND
	if (!bmc_wait) {
		bmc_status |= (BYTE)BMC_ST_RDA;
	}
#else
	bmc_status |= (BYTE)BMC_ST_RDA;
#endif
}

/*
 *	y[Wݍ݊Jn
 */
static void FASTCALL bmc_write_data(void)
{
	/* {`FbN */
	if (!bmc_rw_sub()) {
		return;
	}

	/* JE^ݒ */
	bmc_totalcnt = BMC_PSIZE_32;
	bmc_nowcnt = 0;

	/* Xe[^Xݒ */
	bmc_status |= (BYTE)BMC_ST_BUSY;
	bmc_status &= (BYTE)(~BMC_ST_CME);

	/* WRITE PROTECT`FbN */
	if (bmc_writep[bmc_unit] != 0) {
		bmc_make_stat();
		bmc_status &= (BYTE)(~BMC_ST_BUSY);
		bmc_status |= (BYTE)BMC_ST_CME;
		return;
	}

	/* f[^|C^ݒ */
	bmc_dataptr = bmc_buffer;
	bmc_offset = 0;

	/* ANZX(WRITE) */
	bmc_access[bmc_unit] = (BYTE)BMC_ACCESS_WRITE;

#ifdef FDDSND
	if (!bmc_wait) {
		bmc_status |= (BYTE)BMC_ST_TDRA;
	}
#else
	bmc_status |= (BYTE)BMC_ST_TDRA;
#endif
}

/*
 *	R}h
 */
static void FASTCALL bmc_process_cmd(void)
{
	/* f[^]sĂ΁A~߂ */
	bmc_dataptr = NULL;

#ifdef FDDSND
	/* EFCgtOݒ */
	bmc_wait = fdc_waitmode;
#endif

	/*  */
	switch (bmc_command) {
		/* bubble read */
		case 0x01:
			bmc_read_data();
			break;
		/* bubble write */
		case 0x02:
			bmc_write_data();
			break;
		/* initialize */
		case 0x04:
		case 0x0f:
			bmc_initialize();
			break;
		/* ȊO */
		default:
			/* Xe[^Xݒ肷 */
			bmc_status = (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
			bmc_errorreg = (BYTE)BMC_ES_UNDEF;
			ASSERT(FALSE);
			break;
	}
}

/*
 *	ou Rg[
 *	PoCgǂݏo
 */
BOOL FASTCALL bmc_readb(WORD addr, BYTE *dat)
{
	/* LE`FbN */
	if (!bmc_enable) {
		return FALSE;
	}

	/* FM-8[h */
	if (fm_subtype != FMSUB_FM8) {
		return FALSE;
	}

	/* AhX`FbN */
	if ((addr & 0xfff8) != 0xfd10) {
		return FALSE;
	}

	bmc_use = TRUE;

	switch (addr) {
		/* f[^WX^(BDATA) */
		case 0xfd10:
			*dat = bmc_datareg;
			/* JE^ */
			if (bmc_dataptr) {
				bmc_nowcnt++;
				if (bmc_nowcnt == bmc_totalcnt) {
					bmc_status &= (BYTE)(~BMC_ST_BUSY);
					bmc_status &= (BYTE)(~BMC_ST_RDA);
					bmc_status |= (BYTE)BMC_ST_CME;

					bmc_countreg--;
					if ((bmc_countreg > 0) &&
						((bmc_pagereg & BMC_MAXADDR_32) < BMC_MAXADDR_32)) {
						/* }`y[W */
						bmc_status |= (BYTE)BMC_ST_BUSY;
						bmc_status &= (BYTE)(~BMC_ST_CME);
						bmc_pagereg++;
						bmc_read_data();
						bmc_nowcnt = 0;
						return TRUE;
					}
					/* VOy[W or }`y[WI */
					bmc_dataptr = NULL;

					/* ANZX(READY) */
					bmc_access[bmc_unit] = (BYTE)BMC_ACCESS_READY;
				}
				else {
					bmc_datareg = bmc_dataptr[bmc_nowcnt];
#ifdef FDDSND
					if (bmc_wait) {
						bmc_status &= (BYTE)(~BMC_ST_RDA);
					}
#endif
				}
			}
			return TRUE;

		/* Xe[^XWX^(BSTAT) */
		case 0xfd12:
			bmc_make_stat();
			*dat = bmc_status;
			/* BUSY */
			if ((bmc_status & BMC_ST_BUSY) &&
				(bmc_dataptr == NULL)) {
				/* BUSYtO𗎂Ƃ */
				bmc_status &= (BYTE)(~BMC_ST_BUSY);
				bmc_access[bmc_unit] = (BYTE)BMC_ACCESS_READY;
			}
			/* CAN READ */
			if ((bmc_command == 0x01) &&
				(bmc_dataptr) &&
				!(bmc_status & BMC_ST_RDA)) {
				/* CAN READtO𗧂Ă */
				bmc_status |= (BYTE)BMC_ST_RDA;
			}
			/* CAN WRITE */
			if ((bmc_command == 0x02) &&
				(bmc_dataptr) &&
				!(bmc_status & BMC_ST_TDRA)) {
				/* CAN WRITEtO𗧂Ă */
				bmc_status |= (BYTE)BMC_ST_TDRA;
			}
			return TRUE;

		/* G[Xe[^XWX^(BERRST) */
		case 0xfd13:
			*dat = bmc_errorreg;
			return TRUE;
	}

	return FALSE;
}

/*
 *	ou Rg[
 *	PoCg
 */
BOOL FASTCALL bmc_writeb(WORD addr, BYTE dat)
{
	/* LE`FbN */
	if (!bmc_enable) {
		return FALSE;
	}

	/* 32KBFM-8[h */
	if (fm_subtype != FMSUB_FM8) {
		return FALSE;
	}

	/* AhX`FbN */
	if ((addr & 0xfff8) != 0xfd10) {
		return FALSE;
	}

	bmc_use = TRUE;

	switch (addr) {
		/* f[^WX^(BDATA) */
		case 0xfd10:
			bmc_datareg = dat;
			/* JE^ */
			if (bmc_dataptr) {
				bmc_dataptr[bmc_nowcnt] = bmc_datareg;
				bmc_nowcnt++;
				if (bmc_nowcnt == bmc_totalcnt) {
					bmc_status &= (BYTE)(~BMC_ST_BUSY);
					bmc_status &= (BYTE)(~BMC_ST_TDRA);
					bmc_status |= (BYTE)BMC_ST_CME;

					/* Cgy[W */
					if (!bmc_write_page()) {
						bmc_status |= (BYTE)(BMC_ST_ERROR | BMC_ST_CME);
						bmc_errorreg |= (BYTE)BMC_ES_NOMAKER;
					}

					bmc_countreg--;
					if ((bmc_countreg > 0) &&
						((bmc_pagereg & BMC_MAXADDR_32) < BMC_MAXADDR_32)) {
						/* }`y[W */
						bmc_status |= (BYTE)BMC_ST_BUSY;
#ifdef FDDSND
						if (!bmc_wait) {
							bmc_status |= (BYTE)BMC_ST_TDRA;
						}
#else
						bmc_status |= (BYTE)BMC_ST_TDRA;
#endif
						bmc_status &= (BYTE)(~BMC_ST_CME);
						bmc_pagereg++;
						bmc_nowcnt = 0;
						return TRUE;
					}
					/* VOy[W or }`y[WI */
					bmc_dataptr = NULL;

					/* ANZX(READY) */
					bmc_access[bmc_unit] = (BYTE)BMC_ACCESS_READY;
				}
				else {
#ifdef FDDSND
					if (bmc_wait) {
						bmc_status &= (BYTE)(~BMC_ST_TDRA);
					}
#endif
				}
			}
			return TRUE;

		/* R}hWX^(BCMD) */
		case 0xfd11:
			bmc_command = (BYTE)(dat & 0x0f);
			bmc_process_cmd();
			return TRUE;

		/* y[WAhXWX^H(BPGADH) */
		case 0xfd14:
			bmc_pagereg &= 0x00ff;
			bmc_pagereg |= (WORD)(dat << 8);

			/* y[W烆jbgZo */
			bmc_unit = (BYTE)(bmc_pagereg >> 10);
			return TRUE;

		/* y[WAhXWX^L(BPGADL) */
		case 0xfd15:
			bmc_pagereg &= 0xff00;
			bmc_pagereg |= dat;
			return TRUE;

		/* y[WJEgWX^H(BPGCTH) */
		case 0xfd16:
			bmc_countreg &= 0x00ff;
			bmc_countreg |= (WORD)(dat << 8);
			return TRUE;

		/* y[WJEgWX^L(BPGCTL) */
		case 0xfd17:
			bmc_countreg &= 0xff00;
			bmc_countreg |= dat;
			return TRUE;
	}

	return FALSE;
}

/*
 *	ou Rg[
 *	Z[u
 */
BOOL FASTCALL bmc_save(int fileh)
{
	int i, bmc_units;
	DWORD size;

	bmc_units = BMC_UNITS_32;
	size = (DWORD)BMC_PSIZE_32;

	/* t@C֌WɎĂ */
	for (i=0; i<bmc_units; i++) {
		if (!file_byte_write(fileh, bmc_ready[i])) {
			return FALSE;
		}
	}
	for (i=0; i<bmc_units; i++) {
		if (!file_write(fileh, (BYTE*)bmc_fname[i], 256 + 1)) {
			return FALSE;
		}
	}
	for (i=0; i<bmc_units; i++) {
		if (!file_byte_write(fileh, bmc_media[i])) {
			return FALSE;
		}
	}

	for (i=0; i<bmc_units; i++) {
		if (!file_bool_write(fileh, bmc_teject[i])) {
			return FALSE;
		}
	}

	/* t@CXe[^X */
	if (!file_write(fileh, bmc_buffer, size)) {
		return FALSE;
	}

	/* bmc_dataptr͊Ɉˑf[^|C^ */
	if (!bmc_dataptr) {
		if (!file_word_write(fileh, (WORD)size)) {
			return FALSE;
		}
	}
	else {
		if (!file_word_write(fileh, (WORD)(bmc_dataptr - &bmc_buffer[0]))) {
			return FALSE;
		}
	}

	if (!file_dword_write(fileh, bmc_offset)) {
		return FALSE;
	}

	/* I/O */
	if (!file_byte_write(fileh, bmc_datareg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, bmc_command)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, bmc_status)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, bmc_errorreg)) {
		return FALSE;
	}
	if (!file_word_write(fileh, bmc_pagereg)) {
		return FALSE;
	}
	if (!file_word_write(fileh, bmc_countreg)) {
		return FALSE;
	}

	/* ̑ */
	if (!file_word_write(fileh, bmc_totalcnt)) {
		return FALSE;
	}
	if (!file_word_write(fileh, bmc_nowcnt)) {
		return FALSE;
	}
	for (i=0; i<bmc_units; i++) {
		if (!file_byte_write(fileh, bmc_access[i])) {
			return FALSE;
		}
	}

	/* LE`FbN */
	if (!file_bool_write(fileh, bmc_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, bmc_use)) {
		return FALSE;
	}

#ifdef FDDSND
	if (!file_bool_write(fileh, bmc_wait)) {
		return FALSE;
	}
#else
	if (!file_bool_write(fileh, FALSE)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	ou Rg[
 *	[h
 */
BOOL FASTCALL bmc_load(int fileh, int ver)
{
	int i, bmc_units;
	BYTE ready[BMC_UNITS_32];
	char fname[BMC_UNITS_32][256 + 1];
	BYTE media[BMC_UNITS_32];
	WORD offset;
	DWORD size;
#ifndef FDDSND
	BOOL tmp;
#endif

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	if (ver < 307) {
		bmc_init();
		bmc_reset();
		return TRUE;
	}

	bmc_units = BMC_UNITS_32;
	size = (DWORD)BMC_PSIZE_32;

	/* t@C֌WɎĂ */
	for (i=0; i<bmc_units; i++) {
		if (!file_byte_read(fileh, &ready[i])) {
			return FALSE;
		}
	}
	for (i=0; i<bmc_units; i++) {
		if (!file_read(fileh, (BYTE*)fname[i], 256 + 1)) {
			return FALSE;
		}
	}
	if (ver >= 308) {
		for (i=0; i<bmc_units; i++) {
			if (!file_byte_read(fileh, &media[i])) {
				return FALSE;
			}
		}
	}
	else {
		for (i=0; i<bmc_units; i++) {
			media[i] = 0;
		}
	}

	/* ă}Eg݂ */
	for (i=0; i<bmc_units; i++) {
		bmc_setfile(i, NULL);
		if (ready[i] != BMC_TYPE_NOTREADY) {
			bmc_setfile(i, fname[i]);
			if (bmc_ready[i] != BMC_TYPE_NOTREADY) {
				if (bmc_medias[i] >= (media[i] + 1)) {
					bmc_setmedia(i, media[i]);
				}
			}
		}
	}

	for (i=0; i<bmc_units; i++) {
		if (!file_bool_read(fileh, &bmc_teject[i])) {
			return FALSE;
		}
	}

	/* t@CXe[^X */
	if (!file_read(fileh, bmc_buffer, size)) {
		return FALSE;
	}

	/* bmc_dataptr͊Ɉˑf[^|C^ */
	if (!file_word_read(fileh, &offset)) {
		return FALSE;
	}
	if (offset >= (WORD)size) {
		bmc_dataptr = NULL;
	}
	else {
		bmc_dataptr = &bmc_buffer[offset];
	}

	if (!file_dword_read(fileh, &bmc_offset)) {
		return FALSE;
	}

	/* I/O */
	if (!file_byte_read(fileh, &bmc_datareg)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &bmc_command)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &bmc_status)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &bmc_errorreg)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &bmc_pagereg)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &bmc_countreg)) {
		return FALSE;
	}

	/* ̑ */
	if (!file_word_read(fileh, &bmc_totalcnt)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &bmc_nowcnt)) {
		return FALSE;
	}
	for (i=0; i<bmc_units; i++) {
		if (!file_byte_read(fileh, &bmc_access[i])) {
			return FALSE;
		}
	}

	/* y[W烆jbgZo */
	bmc_unit = (BYTE)(bmc_pagereg >> 10);

	/* LE`FbN */
	if (!file_bool_read(fileh, &bmc_enable)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &bmc_use)) {
		return FALSE;
	}

#ifdef FDDSND
	if (!file_bool_read(fileh, &bmc_wait)) {
		return FALSE;
	}
#else
	if (!file_bool_read(fileh, &tmp)) {
		return FALSE;
	}
#endif

	return TRUE;
}
#endif

#endif	/* XM7_VER == 1 && defined(BUBBLE) */
