/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ {J[h / gTuVXeROM ]
 */

#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "xm7.h"
#include "mmr.h"
#include "device.h"
#include "kanji.h"
#include "jcard.h"

/*
 *	O[o [N
 */
BYTE *dicrom;							/* ROM $40000 */
BYTE *dicram;							/* wKRAM $2000 */
BYTE *extram_b;							/* gRAM $10000 */

BYTE dicrom_bank;						/* ROMoNԍ */
#if XM7_VER >= 3
BOOL extrom_sel;						/* ROM/gROMItO */
#endif
BOOL dicrom_en;							/* ROMANeBu */
BOOL dicram_en;							/* wKRAMANeBu */

#if XM7_VER >= 3
BYTE *extrom;							/* gROM $20000 */
#endif

#if XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH)
BOOL jcard_available;					/* {J[hgp\tO */
BOOL jcard_enable;						/* {J[hLtO */
#endif

/*
 *	{J[h
 *	
 */
BOOL FASTCALL jcard_init(void)
{
	/* [NGA */
	extram_b = NULL;
	dicrom = NULL;
	dicram = NULL;
#if XM7_VER >= 3
	extrom = NULL;
#endif
#if XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH)
	jcard_available = TRUE;
	jcard_enable = TRUE;
#endif

	/* { gRAM */
	extram_b = (BYTE *)malloc(0x10000);
	if (extram_b == NULL) {
		return FALSE;
	}

	/* ROM */
	dicrom = (BYTE *)malloc(0x40000);
	if (dicrom == NULL) {
		return FALSE;
	}
	if (!file_load(DICT_ROM, dicrom, 0x40000)) {
#if XM7_VER == 2 && defined(JCARD) && defined(XM7DASH)
		jcard_available = FALSE;
#else
		return FALSE;
#endif
	}

	/* wKRAMǂݍ */
	dicram = (BYTE *)malloc(0x2000);
	if (dicram == NULL) {
		return FALSE;
	}
	if (!file_load(DICT_RAM, dicram, 0x2000)) {
		/* t@C݂ȂB */
		memset(dicram, 0xff, 0x2000);
	}

#if XM7_VER >= 3
	/* gROM */
	extrom = (BYTE *)malloc(0x20000);
	if (extrom == NULL) {
		return FALSE;
	}
	memset(extrom, 0xff, 0x20000);
	if (!file_load(EXTSUB_ROM, extrom, 0xc000)) {
		return FALSE;
	}
	else {
		/* oN56`63BASIC ROMABu[gROM̓eRs[ */
		memcpy(&extrom[0x18000], basic_rom, 0x7c00);
		if (available_mmrboot) {
			memcpy(&extrom[0x1fe00], boot_mmr, 0x200);
		}
		else {
			memcpy(&extrom[0x1fe00], &init_rom[0x1a00], 0x1e0);
			memset(&extrom[0x1ffe0], 0, 32);
			extrom[0x1fffe] = 0xfe;
			extrom[0x1ffff] = 0x00;
		}
	}
#endif

	return TRUE;
}

/*
 *	{J[h
 *	N[Abv
 */
void FASTCALL jcard_cleanup(void)
{
#ifndef XM7DASH
	ASSERT(extram_b);
	ASSERT(dicram);
	ASSERT(dicrom);
	ASSERT(extrom);
#endif

	/* rŎsꍇl */
#if XM7_VER >= 3
	if (extrom) {
		free(extrom);
	}
#endif
	if (dicram) {
		/* wKRAM̓et@Cɏo */
		file_save(DICT_RAM, dicram, 0x2000);
		free(dicram);
	}
	if (dicrom) {
		free(dicrom);
	}
	if (extram_b) {
		free(extram_b);
	}
}

/*
 *	{J[h
 *	Zbg
 */
void FASTCALL jcard_reset(void)
{
	dicrom_bank = 0;
	dicram_en = FALSE;
	dicrom_en = FALSE;
#if XM7_VER >= 3
	extrom_sel = FALSE;
#endif
}

/*
 *	{J[h
 *	PoCgǂݍ
 */
BYTE FASTCALL jcard_readb(WORD addr)
{
	DWORD dicrom_addr;

#if defined(JCARD) && defined(XM7DASH)
	/* FM77AVFM77AV40EXT|[g */
	if (fm7_ver == 1 ||
		(fm7_ver <= 2 && !(jcard_available && jcard_enable))) {
#else
	/* FM77AV40EX̂݃T|[g */
	if (fm7_ver < 3) {
#endif
		return 0xff;
	}

	/* $28000-$29FFF : wKRAM */
	if ((addr >= 0x8000) && (addr < 0xa000)) {
		if (dicram_en) {
			return dicram[addr - 0x8000];
		}
	}

	/* $2E000-$2EFFF : ROM or gROM */
	if ((addr >= 0xe000) && (addr < 0xf000)) {
		/* ROML */
		if (dicrom_en) {
			addr &= (WORD)0x0fff;
			dicrom_addr = (dicrom_bank << 12);

#if XM7_VER == 2 && defined(JCARD) && defined(XM7DASH)
			/* ROM */
			return dicrom[addr | dicrom_addr];
#else
			/* gROML */
			if (extrom_sel) {
				/* oN0`31 : 1(JIS78) */
				if ((dicrom_bank >= 6) && (dicrom_bank < 8) &&
#ifdef XM7DASH
					(kanji_rom_mode || !kanji_patch_disable) &&
#endif
					!kanji_asis_flag) {
					return (BYTE)(addr & 1);
				}
				if (dicrom_bank < 32) {
					return kanji_rom_jis78[addr | dicrom_addr];
				}

				/* oN32`43 : gTuVXeROM(extsub.rom) */
				/* oN56`63 : F-BASIC V3.0 ROM ($8000-$EFFF) */
				/* oN63     : DOS[hBOOT ($FE00-$FFDF) */
				/* oN63     : 荞݃xN^̈ ($FFE0-$FFFF) */
				return extrom[addr | (dicrom_addr - 0x20000)];
			}
			else {
				/* ROM */
				return dicrom[addr | dicrom_addr];
			}
#endif
		}
	}

	/* gRAM */
	return extram_b[addr];
}

/*
 *	{J[h
 *	PoCg
 */
void FASTCALL jcard_writeb(WORD addr, BYTE dat)
{
#if defined(JCARD) && defined(XM7DASH)
	/* FM77AVFM77AV40EXT|[g */
	if (fm7_ver == 1 ||
		(fm7_ver <= 2 && !(jcard_available && jcard_enable))) {
#else
	/* FM77AV40EX̂݃T|[g */
	if (fm7_ver < 3) {
#endif
		return;
	}

	/* $28000-$29FFF : wKRAM */
	if ((addr >= 0x8000) && (addr < 0xa000)) {
		if (dicram_en) {
			dicram[addr - 0x8000] = dat;
			return;
		}
	}

	/* gRAM (ROM̑IԂɊւ炸݉\) */
	extram_b[addr] = dat;
}

/*
 *	{J[h
 *	Z[u
 */
BOOL FASTCALL jcard_save(int fileh)
{
	if (!file_byte_write(fileh, dicrom_bank)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (!file_bool_write(fileh, extrom_sel)) {
		return FALSE;
	}
#endif
	if (!file_bool_write(fileh, dicrom_en)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, dicram_en)) {
		return FALSE;
	}
	if (!file_write(fileh, extram_b, 0x10000)) {
		return FALSE;
	}
#if defined(JCARD) && defined(XM7DASH)
	if (!file_bool_write(fileh, jcard_enable)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	{J[h
 *	[h
 */
#if XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH)
BOOL FASTCALL jcard_load(int fileh, int ver, BOOL bflag)
#else
BOOL FASTCALL jcard_load(int fileh, int ver)
#endif
{
	/* o[W`FbN */
#if XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH)
	if (ver < 800 && !bflag) {
#else
	if (ver < 800) {
#endif
		dicrom_bank = 0;
#if XM7_VER >= 3
		extrom_sel = FALSE;
#endif
		dicram_en = FALSE;
		dicrom_en = FALSE;
		return TRUE;
	}

	if (!file_byte_read(fileh, &dicrom_bank)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (!file_bool_read(fileh, &extrom_sel)) {
		return FALSE;
	}
#endif
	if (!file_bool_read(fileh, &dicrom_en)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &dicram_en)) {
		return FALSE;
	}
	if (!file_read(fileh, extram_b, 0x10000)) {
		return FALSE;
	}
#if XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH)
	if (bflag) {
		if (!file_bool_read(fileh, &jcard_enable)) {
			return FALSE;
		}
	}
	else {
		jcard_enable = TRUE;
	}
	if (!jcard_available) {
		/* {J[hɃJ[hgp̃Xe[g͓ǂ߂Ȃ */
		if (dicrom_bank != 0 || dicram_en || dicrom_en) {
			return FALSE;
		}
	}
#endif
#if XM7_VER >= 3 && defined(XM7DASH)
	/* AV20EX */
	if (bDownGradeMode && extrom_sel) {
		return FALSE;
	}
#endif

	return TRUE;
}

#endif	/* XM7_VER >= 3 */
