/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ {ʐMJ[h ]
 */

#ifdef JSUB

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "jsubsys.h"
#include "kanji.h"
#include "mainetc.h"
#include "device.h"
#include "event.h"

/*
 *	O[o [N
 */
cpu6809_t jsubcpu;
BYTE *jsubrom;						/* {Tuj^ROM */
BYTE *jdicrom;						/* ROM({ʐMJ[h) */
BYTE *jsub_sram;					/* [NSRAM */
BOOL jsub_available;				/* {TuVXegp\tO */
BOOL jsub_enable;					/* {TuVXegptO */
BOOL jsub_haltflag;					/* {TuVXeHALTtO */
BYTE jsub_dicbank;					/* ROMoN */
BYTE jsub_address;					/* RCBAhXJE^ */
WORD jsub_kanji_addr;				/* ROMAhXWX^ */
#ifdef XM7DASH
BOOL jsub_kanji_disable;			/* ROM({ʐMJ[h) */
#endif

/*
 *	vg^Cv錾
 */
void jsub_reset(void);
void jsub_line(void);
void jsub_exec(void);
BOOL jsubcpu_event(void);

/*
 *	{TuCPU
 *	
 */
BOOL FASTCALL jsubsys_init(void)
{
	/* xASăNA */
	jsubrom = NULL;
	jdicrom = NULL;
	jsub_sram = NULL;
	jsub_available = TRUE;
	jsub_enable = TRUE;
	jsub_address = 0;
#ifdef XM7DASH
	jsub_kanji_disable = FALSE;
#endif

	/* m */
	jsubrom = (BYTE *)malloc(0x4000);
	if (jsubrom == NULL) {
		return FALSE;
	}
	jdicrom = (BYTE *)malloc(0x40000);
	if (jdicrom == NULL) {
		return FALSE;
	}
	jsub_sram = (BYTE *)malloc(0x2000);
	if (jsub_sram == NULL) {
		return FALSE;
	}

	/* ROMt@Cǂݍ */
	if (!file_load(JSUBSYS_ROM, jsubrom, 0x4000)) {
		jsub_available = FALSE;
	}
	if (!file_load(DICT_ROM, jdicrom, 0x40000)) {
		jsub_available = FALSE;
	}

	/* ANZX֐ݒ */
	jsubcpu.readmem = jsubmem_readb;
	jsubcpu.writemem = jsubmem_writeb;

	return TRUE;
}

/*
 *	{TuCPU
 *	N[Abv
 */
void FASTCALL jsubsys_cleanup(void)
{
#ifndef XM7DASH
	ASSERT(jsubrom);
	ASSERT(jdicrom);
	ASSERT(jsub_sram);
#endif

	if (jsubrom) {
		free(jsubrom);
	}
	if (jdicrom) {
		free(jdicrom);
	}
	if (jsub_sram) {
		free(jsub_sram);
	}
}

/*
 *	{TuCPU
 *	Zbg
 */
void FASTCALL jsubsys_reset(void)
{
	jsub_address = 0;
	jsub_haltflag = FALSE;
	jsub_kanji_addr = 0;

	if (jsub_available) {
		jsub_reset();
	}
}

/*
 *	{TuCPU
 *	Pss
 */
void FASTCALL jsubcpu_execline(void)
{
	if (jsub_available && jsub_enable) {
		jsub_line();
	}
}

/*
 *	{TuCPU
 *	s
 */
void FASTCALL jsubcpu_exec(void)
{
	if (jsub_available && jsub_enable) {
		jsub_exec();
	}
}

/*
 *	{TuCPU
 *	NMI荞ݐݒ
 */
void FASTCALL jsubcpu_nmi(void)
{
}

/*
 *	{TuCPU
 *	FIRQ荞ݐݒ
 */
void FASTCALL jsubcpu_firq(void)
{
}

/*
 *	{TuCPU
 *	IRQ荞ݐݒ
 */
void FASTCALL jsubcpu_irq(void)
{
}

/*
 *	{TuCPU
 *	PoCg擾
 */
BYTE FASTCALL jsubmem_readb(WORD addr)
{
	/* {Tuj^ROM */
	if (addr >= 0xc000) {
		return jsubrom[addr - 0xc000];
	}

	/* ROM */
	if ((addr >= 0xa000) && (addr <= 0xafff)) {
		return jdicrom[(addr & 0xfff) | (jsub_dicbank << 12)];
	}

	/* SRAM */
	if ((addr >= 0x8000) && (addr <= 0x9ffe)) {
		return jsub_sram[addr - 0x8000];
	}

	/* oNEtOWX^ */
	if (addr == 0x9fff) {
		return jsub_dicbank;
	}

	return 0xff;
}

/*
 *	{TuCPU
 *	PoCg擾(I/OȂ)
 */
BYTE FASTCALL jsubmem_readbnio(WORD addr)
{
	/* {Tuj^ROM */
	if (addr >= 0xc000) {
		return jsubrom[addr - 0xc000];
	}

	/* ROM */
	if ((addr >= 0xa000) && (addr <= 0xafff)) {
		return jdicrom[(addr & 0xfff) | (jsub_dicbank << 12)];
	}

	/* SRAM */
	if ((addr >= 0x8000) && (addr <= 0x9ffe)) {
		return jsub_sram[addr - 0x8000];
	}

	/* oNEtOWX^ */
	if (addr == 0x9fff) {
		return (BYTE)jsub_dicbank;
	}

#ifdef XM7DASH
	if (jsubcpu.intr & INTR_SLOAD) {
		ASSERT(FALSE);
	}
#else
	ASSERT(FALSE);
#endif
	return 0xff;
}

/*
 *	{TuCPU
 *	PoCg
 */
void FASTCALL jsubmem_writeb(WORD addr, BYTE dat)
{
	/* oNEtOWX^ */
	if (addr == 0x9fff) {
		if (dat & 0x80) {
			jsub_haltflag = FALSE;
		}
		else {
			jsub_haltflag = TRUE;
		}

		jsub_dicbank = (BYTE)(dat & 0x3f);
		return;
	}

	/* SRAM */
	if ((addr >= 0x8000) && (addr <= 0x9ffe)) {
		jsub_sram[addr - 0x8000] = dat;
		return;
	}

	return;
}

/*
 *	{TuVXe
 *	CRCBPoCgǂݍ
 */
BYTE FASTCALL jsub_readrcb(void)
{
	return jsub_sram[0x1f00 | (jsub_address++)];
}

/*
 *	{TuVXe
 *	CRCBPoCg
 */
void FASTCALL jsub_writercb(BYTE dat)
{
	jsub_sram[0x1f00 | (jsub_address++)] = dat;
}

/*
 *	{TuVXe
 *	AhXJE^NA
 */
void FASTCALL jsub_clear_address(void)
{
	jsub_address = 0;
}

/*
 *	{TuVXeC^tF[X
 *	PoCgǂݏo
 */
BOOL FASTCALL jsub_readb(WORD addr, BYTE *dat)
{
	int offset;

#if XM7_VER >= 3
	/* FM77AV40EX[hE{ʐMJ[h͖ */
	if ((fm7_ver == 3) || !jsub_enable) {
#elif XM7_VER >= 2
	/* {ʐMJ[h͖ */
	if (!jsub_enable) {
#else
	/* FM-8[hE{ʐMJ[h͖ */
	if ((fm_subtype == FMSUB_FM8) || !jsub_enable) {
#endif
		return FALSE;
	}

	switch (addr) {
		/* {TuVXetO */
		case 0xfd28:
			*dat = 0xff;
			if (jsub_haltflag) {
				*dat &= (BYTE)~0x80;
			}
			return TRUE;
	
		/* RCBf[^ǂݍ */
		case 0xfd29:
			if (jsub_haltflag) {
				*dat = jsub_readrcb();
			}
			else {
				*dat = 0xff;
			}
			return TRUE;

		/* 1f[^ */
		case 0xfd2a:		/* 1f[^LEFT */
		case 0xfd2b:		/* 1f[^RIGHT */
#ifdef XM7DASH
#if XM7_VER >= 2
			if (fm7_ver == 2 && !jsub_kanji_disable) {
#else
			if (fm_subtype == FMSUB_FM77 && !jsub_kanji_disable) {
#endif
#else
#if XM7_VER >= 2
			if (fm7_ver == 2) {
#else
			if (fm_subtype == FMSUB_FM77) {
#endif
#endif
				offset = jsub_kanji_addr << 1;
				if ((offset >= 0x6000) && (offset < 0x8000) &&
#ifdef XM7DASH
					(kanji_rom_mode || !kanji_patch_disable) &&
#endif
					!kanji_asis_flag) {
					/* $6000`$7FFF͖`̈ */
					*dat = (BYTE)(addr & 1);
				}
				else {
					/* ʏ̈ */
#if XM7_VER >= 3
					*dat = kanji_rom_jis78[offset + (addr & 1)];
#else
					*dat = kanji_rom[offset + (addr & 1)];
#endif
				}
				return TRUE;
			}

			return FALSE;
	}

	return FALSE;
}

/*
 *	{TuVXeC^tF[X
 *	PoCg
 */
BOOL FASTCALL jsub_writeb(WORD addr, BYTE dat)
{
#if XM7_VER >= 3
	/* FM77AV40EX[hE{ʐMJ[h͖ */
	if ((fm7_ver == 3) || !jsub_enable) {
#elif XM7_VER >= 2
	/* {ʐMJ[h͖ */
	if (!jsub_enable) {
#else
	/* FM-8[hE{ʐMJ[h͖ */
	if ((fm_subtype == FMSUB_FM8) || !jsub_enable) {
#endif
		return FALSE;
	}

	switch (addr) {
		/* AhX */
		case 0xfd28:
#ifdef XM7DASH
			if (!jsub_kanji_disable) {
				jsub_kanji_addr &= 0x00ff;
				jsub_kanji_addr |= (WORD)(dat << 8);
				return TRUE;
			}
			return FALSE;
#else
			jsub_kanji_addr &= 0x00ff;
			jsub_kanji_addr |= (WORD)(dat << 8);
			return TRUE;
#endif

		/* AhX */
		case 0xfd29:
#ifdef XM7DASH
			if (!jsub_kanji_disable) {
				jsub_kanji_addr &= 0xff00;
				jsub_kanji_addr |= dat;
				return TRUE;
			}
			return FALSE;
#else
			jsub_kanji_addr &= 0xff00;
			jsub_kanji_addr |= dat;
			return TRUE;
#endif

		/* {TuVXetO */
		case 0xfd2a:
			if (dat & 0x80) {
				jsub_haltflag = FALSE;
			}
			else {
				jsub_haltflag = TRUE;
				jsub_clear_address();
			}
			return TRUE;

		/* RCBf[^ */
		case 0xfd2b:
			if (jsub_haltflag) {
				jsub_writercb(dat);
			}
			return TRUE;
	}

	return FALSE;
}

/*
 *	{TuVXe
 *	Z[u
 */
BOOL FASTCALL jsubsys_save(int fileh)
{
	/* vbgtH[Ƃ̃pbLO邽߁A */
	if (!file_byte_write(fileh, jsubcpu.cc)) {
		return FALSE;
	}

	if (!file_byte_write(fileh, jsubcpu.dp)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.acc.d)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.x)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.y)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.u)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.s)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.pc)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.intr)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.cycle)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.total)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, jsub_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, jsub_haltflag)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, jsub_dicbank)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, jsub_address)) {
		return FALSE;
	}
	if (!file_write(fileh, jsub_sram, 0x2000)) {
		return FALSE;
	}
	if (!file_word_write(fileh, jsub_kanji_addr)) {
		return FALSE;
	}

#ifdef XM7DASH
	if (!file_bool_write(fileh, jsub_kanji_disable)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	{TuVXe
 *	[h
 */
#if XM7_VER >= 2 && defined(XM7DASH)
BOOL FASTCALL jsubsys_load(int fileh, int ver, BOOL bflag)
#else
BOOL FASTCALL jsubsys_load(int fileh, int ver)
#endif
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

#if XM7_VER >= 2 && defined(XM7DASH)
	if (bflag) {
#else
	if (ver >= 301 && ver <= 499) {
#endif
		/* vbgtH[Ƃ̃pbLO邽߁A */
		if (!file_byte_read(fileh, &jsubcpu.cc)) {
			return FALSE;
		}

		if (!file_byte_read(fileh, &jsubcpu.dp)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.acc.d)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.x)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.y)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.u)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.s)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.pc)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.intr)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.cycle)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.total)) {
			return FALSE;
		}

		if (!file_bool_read(fileh, &jsub_enable)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &jsub_haltflag)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &jsub_dicbank)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &jsub_address)) {
			return FALSE;
		}
		if (!file_read(fileh, jsub_sram, 0x2000)) {
			return FALSE;
		}
#if XM7_VER >= 2
		if (!file_word_read(fileh, &jsub_kanji_addr)) {
			return FALSE;
		}
#else
#ifdef XM7DASH
		if ((ver >= 305 && ver <= 399) || (ver >= 404 && ver <= 499)) {
#else
		if (ver >= 305 && ver <= 499) {
#endif
			if (!file_word_read(fileh, &jsub_kanji_addr)) {
				return FALSE;
			}
		}
#endif

#ifdef XM7DASH
#if XM7_VER >= 2
		if (!file_bool_read(fileh, &jsub_kanji_disable)) {
			return FALSE;
		}
#else
		if (ver >= 404 && ver <= 499) {
			if (!file_bool_read(fileh, &jsub_kanji_disable)) {
				return FALSE;
			}
		}
#endif
#endif
	}

	/* {Tugps̏ꍇfBZ[u */
	if (!jsub_available) {
		jsub_enable = FALSE;
	}

	return TRUE;
}

#endif
