/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ ROM (񊿎EꐅE񐅏) ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "kanji.h"
#include "subctrl.h"
#include "device.h"
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
#include "jcard.h"
#endif

/*
 *	O[o [N
 */
WORD kanji_addr;						/* AhXWX^() */
BYTE *kanji_rom;						/* PROM */
BYTE *kanji_rom2;						/* QROM */
#if XM7_VER >= 3
BYTE *kanji_rom_jis78;					/* PROM (JIS78) */
#endif
BOOL kanji_asis_flag;					/* JIS78G~[VtO */
#ifdef XM7DASH
BOOL kanji_rom_mode;					/* ROMG~[V[h */
BOOL kanji_patch_disable;				/* `̈pb`tO */
#endif

/*
 *	X^eBbN [N
 */
static BOOL kanji_rom_available;		/* ROMgp\tO */
#ifndef XM7DASH
static BOOL kanji_rom_mode;				/* ROMG~[V[h */
#if XM7_VER <= 2
static BOOL kanji_rom2_flag;				/* 2ROMgp\tO */
#endif
#endif

/*
 *	vg^Cv錾
 */
static void FASTCALL kanji_make_jis78(BYTE *rom);


/*
 *	ROM
 *	
 */
BOOL FASTCALL kanji_init(void)
{
	/* ̂܂܃[hItɂ */
	kanji_asis_flag = FALSE;
#ifdef XM7DASH
	kanji_rom_mode = FALSE;
#endif

	/* m */
#if XM7_VER >= 3
	kanji_rom_jis78 = (BYTE *)malloc(0x20000);
	if (!kanji_rom_jis78) {
		return FALSE;
	}
	kanji_rom = (BYTE *)malloc(0x40000);
#else
#if (XM7_VER == 1 || !defined(JCARD)) && defined(XM7DASH)
	kanji_rom = (BYTE *)malloc(0x20000);
#else
	kanji_rom = (BYTE *)malloc(0x40000);
#endif
#endif
	if (!kanji_rom) {
		return FALSE;
	}

	/* 1(JIS78) t@Cǂݍ */
#if XM7_VER >= 3
	if (!file_load(KANJI_ROM_J78, kanji_rom_jis78, 0x20000)) {
#else
	if (!file_load(KANJI_ROM_J78, kanji_rom, 0x20000)) {
#endif
		kanji_rom_available = FALSE;
	}
	else {
		kanji_rom_available = TRUE;
#ifndef XM7DASH
		kanji_rom_mode = TRUE;
#endif
	}

#if XM7_VER >= 3
	/* 1(JIS83) t@Cǂݍ */
	if (!file_load(KANJI_ROM, kanji_rom, 0x20000)) {
		return FALSE;
	}
	else {
		/* ROMG~[VJIS78tHg̗LŔf */
#ifdef XM7DASH
		if (!kanji_rom_available) {
			kanji_rom_available = TRUE;
			kanji_rom_mode = TRUE;
		}
#else
		if (kanji_rom_available) {
			kanji_rom_mode = FALSE;
		}
		else {
			kanji_rom_mode = TRUE;
		}
#endif
	}

	/* 2 t@Cǂݍ */
	kanji_rom2 = (BYTE *)(kanji_rom + 0x20000);
	if (!file_load(KANJI_ROM2, kanji_rom2, 0x20000)) {
		return FALSE;
	}
#else
	/* 2 t@Cǂݍ */
#ifdef XM7DASH
#if XM7_VER >= 2 && defined(JCARD)
	kanji_rom2 = (BYTE *)(kanji_rom + 0x20000);
	if (!file_load(KANJI_ROM2, kanji_rom2, 0x20000)) {
		kanji_rom2 = NULL;
	}
#else
	kanji_rom2 = (BYTE *)malloc(0x20000);
	if (!file_load(KANJI_ROM2, kanji_rom2, 0x20000)) {
		free(kanji_rom2);
		kanji_rom2 = NULL;
	}
#endif
#else
	kanji_rom2 = (BYTE *)(kanji_rom + 0x20000);
	if (!file_load(KANJI_ROM2, kanji_rom2, 0x20000)) {
		kanji_rom2_flag = FALSE;
	}
	else {
		kanji_rom2_flag = TRUE;
	}
#endif

	/* 1(JIS83) t@Cǂݍ */
	if (!kanji_rom_available) {
		if (file_load(KANJI_ROM, kanji_rom, 0x20000)) {
			/* IJIS78ROMG~[Vs */
			kanji_rom_available = TRUE;
			kanji_rom_mode = TRUE;
		}
#if XM7_VER >= 2
		else {
			return FALSE;
		}
#endif
	}
#endif

	/* JIS78ROMf[^ */
#if XM7_VER >= 3
	if (kanji_rom_mode) {
		memcpy(kanji_rom_jis78, kanji_rom, 0x20000);
		kanji_make_jis78(kanji_rom_jis78);
	}
#else
	if (kanji_rom_available && kanji_rom_mode) {
		kanji_make_jis78(kanji_rom);
	}
#endif

#ifdef XM7DASH
#if XM7_VER == 1 || (XM7_VER <= 2 && !defined(JCARD))
	if (kanji_rom2) {
		free(kanji_rom2);
		kanji_rom2 = NULL;
	}
#endif
	kanji_patch_disable = FALSE;
#endif

	return TRUE;
}

/*
 *	ROM
 *	N[Abv
 */
void FASTCALL kanji_cleanup(void)
{
#ifndef XM7DASH
	ASSERT(kanji_rom);
#endif
#if XM7_VER >= 3
#ifndef XM7DASH
	ASSERT(kanji_rom_jis78);
#endif

	if (kanji_rom_jis78) {
		free(kanji_rom_jis78);
		kanji_rom_jis78 = NULL;
	}
#endif
	if (kanji_rom) {
		free(kanji_rom);
		kanji_rom = NULL;
		kanji_rom2 = NULL;
	}
}

/*
 *	ROM
 *	Zbg
 */
void FASTCALL kanji_reset(void)
{
	kanji_addr = 0;
}

/*
 *	ROM
 *	PoCgǂݏo
 */
BOOL FASTCALL kanji_readb(WORD addr, BYTE *dat)
{
	int offset;

	switch (addr) {
		/* AhX(1E2)ʁE */
		case 0xfd20:
		case 0xfd21:
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
		case 0xfd2c:
		case 0xfd2d:
#endif
			*dat = 0xff;
			return TRUE;

		/* 1f[^ */
		case 0xfd22:		/* 1f[^LEFT */
		case 0xfd23:		/* 1f[^RIGHT */
#if XM7_VER == 1
			if (!kanji_rom_available) {
				*dat = 0xff;
				return TRUE;
			}
#endif

#if XM7_VER >= 3 || (XM7_VER == 1 && defined(L4CARD))
#if XM7_VER >= 3
			if (subkanji_flag) {
#else
			if (subkanji_flag && (fm_subtype == FMSUB_FM77)) {
#endif
				*dat = 0xff;
				return TRUE;
			}
#endif

			offset = kanji_addr << 1;
#if XM7_VER >= 3
			if (fm7_ver >= 3) {
				/* FM77AV40EX[hJIS83 */
				*dat = kanji_rom[offset + (addr & 1)];
			}
			else if ((fm7_ver <= 1) && (offset >= 0x6000) &&
					 (offset < 0x8000) &&
#ifdef XM7DASH
					(kanji_rom_mode || !kanji_patch_disable) &&
#endif
					 !kanji_asis_flag) {
				/* FM-7[h$6000`$7FFF͖`̈ */
				*dat = (BYTE)(addr & 1);
			}
			else {
				/* FM-7/FM77AV[hJIS78 */
				*dat = kanji_rom_jis78[offset + (addr & 1)];
			}
#else
 #if XM7_VER >= 2
			if ((fm7_ver <= 1) && (offset >= 0x6000) && (offset < 0x8000) &&
 #else
			if ((offset >= 0x6000) && (offset < 0x8000) &&
 #endif
#ifdef XM7DASH
				(kanji_rom_mode || !kanji_patch_disable) &&
#endif
				!kanji_asis_flag) {
				/* FM-7[h$6000`$7FFF͖`̈ */
				*dat = (BYTE)(addr & 1);
			}
			else {
				/* ʏ̈ */
				*dat = kanji_rom[offset + (addr & 1)];
			}
#endif
			return TRUE;

#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
		/* 2f[^ */
		case 0xfd2e:		/* 2f[^LEFT */
		case 0xfd2f:		/* 2f[^RIGHT */
#if defined(JCARD) && defined(XM7DASH)
#if XM7_VER >= 3
			if ((fm7_ver == 1 ||
				 (fm7_ver <= 2 && !(jcard_available && jcard_enable))) ||
				subkanji_flag) {
#else
			if (fm7_ver == 1 ||
				(fm7_ver <= 2 && !(jcard_available && jcard_enable)) ||
				kanji_rom2 == NULL) {
#endif
#else
			if ((fm7_ver < 3) || subkanji_flag) {
#endif
				*dat = 0xff;
			}
			else {
				offset = kanji_addr << 1;
				*dat = kanji_rom2[offset + (addr & 1)];
			}
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	ROM
 *	PoCg
 */
BOOL FASTCALL kanji_writeb(WORD addr, BYTE dat)
{
	switch (addr) {
		/* AhX(1E2) */
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
		case 0xfd2c:
#if defined(JCARD) && defined(XM7DASH)
			if (fm7_ver == 1 ||
				(fm7_ver <= 2 && !(jcard_available && jcard_enable))) {
#else
			if (fm7_ver < 3) {
#endif
				return TRUE;
			}
#endif
		case 0xfd20:
			kanji_addr &= 0x00ff;
			kanji_addr |= (WORD)(dat << 8);
			return TRUE;

		/* AhX(1E2) */
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
		case 0xfd2d:
#if defined(JCARD) && defined(XM7DASH)
			if (fm7_ver == 1 ||
				(fm7_ver <= 2 && !(jcard_available && jcard_enable))) {
#else
			if (fm7_ver < 3) {
#endif
				return TRUE;
			}
#endif
		case 0xfd21:
			kanji_addr &= 0xff00;
			kanji_addr |= dat;
			return TRUE;

		/* f[^ */
		case 0xfd22:
		case 0xfd23:
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
		case 0xfd2f:
#endif
			return TRUE;

#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
		/* {J[h oNZNg */
		case 0xfd2e:
#if defined(JCARD) && defined(XM7DASH)
			if (fm7_ver == 1 ||
				(fm7_ver <= 2 && !(jcard_available && jcard_enable))) {
#else
			if (fm7_ver < 3) {
#endif
				return TRUE;
			}

			/* bit7:wKRAM */
			if (dat & 0x80) {
				dicram_en = TRUE;
			}
			else {
				dicram_en = FALSE;
			}

			/* bit6:ROMCl[u */
			if (dat & 0x40) {
				dicrom_en = TRUE;
			}
			else {
				dicrom_en = FALSE;
			}

			/* bit0-5:ROMoN */
			dicrom_bank = (BYTE)(dat & 0x3f);
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	ROM
 *	JIS78̈쐬
 */
static void FASTCALL kanji_make_jis78(BYTE *rom)
{
	/* JIS78`e[u(FM77AV) */
	static const DWORD jis78_table[] = {
		0xffffffff, 0x00000001, 0xffff8001, 0xfc00ffff,
		0x00000001, 0x00000001, 0xfe000001, 0x00000001,
		0xffffffff, 0x80000000, 0xffffffff, 0xf8000001,
		0xfff00000, 0xff800000, 0xffffffff, 0xfffc0000,
		0xffffffff, 0x00000000, 0xffffffff, 0xf8000001,
		0x00000000, 0x00000000, 0xfe000001, 0x0001fffc,
	};

	/* JIS781JIS832֊e[u */
	static const DWORD convert_table[][2] = {
		/* 1code addr		  2code addr */
		{/* 0x3033*/0x4130,	/* 0x724D*/0xE4D0},
		{/* 0x3229*/0x4490,	/* 0x7274*/0xD540},
		{/*a 0x3342*/0x6620,	/*y 0x695A*/0x93A0},
		{/*h 0x3349*/0x6690,	/* 0x5978*/0x5380},
		{/* 0x3376*/0x8760,	/*} 0x635E*/0x87E0},
		{/* 0x3443*/0x6830,	/* 0x5E75*/0x5D50},
		{/* 0x3452*/0x6920,	/*| 0x6B5D*/0x97D0},
		{/*z 0x375B*/0x6FB0,	/* 0x7074*/0xD140},
		{/*{ 0x395C*/0x73C0,	/* 0x6268*/0xA480},
		{/* 0x3C49*/0x7890,	/*A 0x6922*/0x7220},
		{/*x 0x3F59*/0x7F90,	/* 0x7057*/0xE170},
		{/*G 0x4128*/0xA280,	/* 0x6C4D*/0x98D0},
		{/* 0x445B*/0xC9B0,	/* 0x5464*/0x4840},
		{/*v 0x4557*/0xCB70,	/* 0x626A*/0xA4A0},
		{/* 0x456E*/0xEAE0,	/* 0x5B6D*/0x56D0},
		{/* 0x4573*/0xEB30,	/* 0x5E39*/0x1D90},
		{/* 0x4676*/0xED60,	/* 0x6D6E*/0xBAE0},
		{/* 0x4768*/0xEE80,	/* 0x6A24*/0x7440},
		{/*O 0x4930*/0xB300,	/*w 0x5B58*/0x3780},
		{/* 0x4B79*/0xF790,	/* 0x5056*/0x2160},
		{/* 0x4C79*/0xF990,	/*M 0x692E*/0x72E0},
		{/*U 0x4F36*/0xBF60,	/* 0x6446*/0x8860},
		{/* 0x3646*/0x6C60,	/* 0x7421*/0xC810},
		{/* 0x4B6A*/0xF6A0,	/* 0x7422*/0xC820},
		{/*y 0x4D5A*/0xDBA0,	/* 0x7423*/0xC830},
		{0, 0},
	};

	DWORD addr;
	DWORD i;

	/* pb`ĕsvtOĂꍇ͏XLbv */
	if (kanji_asis_flag) {
		return;
	}

	/* JIS78ROMł͖`ɂȂĂ镔𖢒`Ƃ */
	for (addr = 0; addr < 0x6000; addr += 32) {
		if (jis78_table[addr >> 10] & (1 << (addr >> 5))) {
			/* `tHg쐬 */
			for (i = 0; i < 32; i += 2) {
				rom[addr + i + 0] = 0x00;
				rom[addr + i + 1] = 0x01;
			}
		}
	}

#if XM7_VER <= 2
	/* KANJI2.ROMȂꍇ͂̂܂܋A */
#ifdef XM7DASH
	if (!kanji_rom2) {
#else
	if (!kanji_rom2_flag) {
#endif
		return;
	}
#endif

	/* JIS83œւꂽ2ROM㏑ */
	for (addr = 0; convert_table[addr][0]; addr ++) {
		/* tHgRs[ */
		for (i = 0; i < 32; i ++) {
			rom[convert_table[addr][0] * 2 + i] =
			kanji_rom2[convert_table[addr][1] * 2 + i];
		}
	}
}

/*
 *	ROM
 *	Z[u
 */
BOOL FASTCALL kanji_save(int fileh)
{
	if (!file_word_write(fileh, kanji_addr)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	ROM
 *	[h
 */
BOOL FASTCALL kanji_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	if (!file_word_read(fileh, &kanji_addr)) {
		return FALSE;
	}

	return TRUE;
}
