/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ CCPU ]
 */

#include "xm7.h"
#include "subctrl.h"
#include "keyboard.h"
#include "mainetc.h"
#include "tapelp.h"
#include "device.h"

/*
 *	O[o [N
 */
cpu6809_t maincpu;

/*
 *	vg^Cv錾
 */
void main_reset(void);
void main_line(void);
void main_exec(void);

/*
 *	CCPU
 *	
 */
BOOL FASTCALL maincpu_init(void)
{
	maincpu.readmem = mainmem_readb;
	maincpu.writemem = mainmem_writeb;

	return TRUE;
}

/*
 *	CCPU
 *	N[Abv
 */
void FASTCALL maincpu_cleanup(void)
{
	return;
}

/*
 *	CCPU
 *	Zbg
 */
void FASTCALL maincpu_reset(void)
{
	main_reset();
}

/*
 *	CCPU
 *	Pss
 */
void FASTCALL maincpu_execline(void)
{
	DWORD cnt;

#if XM7_VER >= 2
	cnt = 0x100;
#else
	if (lowspeed_mode) {
		cnt = 0x09e;
	}
	else {
		cnt = 0x100;
	}
#endif

	/* 1ss */
	main_line();

	/* e[vJE^ */
	if (tape_motor) {
		tape_subcnt += (DWORD)(maincpu.cycle << 4);
		if (tape_subcnt >= cnt) {
			tape_subcnt -= cnt;
			tape_count++;
			if (tape_count == 0) {
				tape_count = 0xffff;
			}
		}
	}
}

/*
 *	CCPU
 *	s
 */
void FASTCALL maincpu_exec(void)
{
	DWORD cnt;

#if XM7_VER >= 2
	cnt = 0x100;
#else
	if (lowspeed_mode) {
		cnt = 0x09e;
	}
	else {
		cnt = 0x100;
	}
#endif

	/* s */
	main_exec();

	/* e[vJE^ */
	if (tape_motor) {
		tape_subcnt += (DWORD)(maincpu.cycle << 4);
		if (tape_subcnt >= cnt) {
			tape_subcnt -= cnt;
			tape_count++;
			if (tape_count == 0) {
				tape_count = 0xffff;
			}
		}
	}
}

/*
 *	CCPU
 *	NMI荞ݐݒ
 */
void FASTCALL maincpu_nmi(void)
{
	/* FM-771MB FDDΉ(IRQ)ŎgĂ炵 */
	if (maincpu.intr & INTR_SLOAD){
		maincpu.intr |= INTR_NMI;
	}
}

/*
 *	CCPU
 *	FIRQ荞ݐݒ
 */
void FASTCALL maincpu_firq(void)
{
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
	/* SFDC̍f[^]pDRQPULSEM */
	if (sfd_firq_mode) {
		if (sfd_drq_flag) {
			maincpu.intr |= INTR_FIRQ_PLS;
		}
		else {
			maincpu.intr &= ~INTR_FIRQ_PLS;
		}

		/* SFD荞݃[h͂܂ */
		return;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */

	/* BREAKL[yсATuCPŨAeV荞 */
	if (break_flag || subattn_flag) {
		maincpu.intr |= INTR_FIRQ;
	}
	else {
		maincpu.intr &= ~INTR_FIRQ;
	}
}

/*
 *	CCPU
 *	IRQ荞ݐݒ
 */
void FASTCALL maincpu_irq(void)
{
#if XM7_VER == 1
	/* FM-8[hIRQ荞ݐݒ */
	if (fm_subtype == FMSUB_FM8) {
#ifdef XM7DASH
		if (mfd_irq_flag ||
 #ifdef SFDC
			sfd_irq_flag ||
			dma_irq_flag ||
 #endif	/* SFDC */
			txrdy_irq_flag ||
#else	/* XM7DASH */
		if (txrdy_irq_flag ||
#endif	/* XM7DASH */
			rxrdy_irq_flag ||
			syndet_irq_flag) {
			maincpu.intr |= INTR_IRQ;
		}
		else {
			maincpu.intr &= ~INTR_IRQ;
		}
		return;
	}
#endif

	/* IRQ荞ݐݒ */
	if ((key_irq_flag && !(key_irq_mask)) ||
		timer_irq_flag ||
		lp_irq_flag ||
		mfd_irq_flag ||
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
		sfd_irq_flag ||
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */
		txrdy_irq_flag ||
		rxrdy_irq_flag ||
		syndet_irq_flag ||
#if XM7_VER >= 3
		dma_irq_flag ||
#endif
		opn_irq_flag ||
		whg_irq_flag ||
		thg_irq_flag) {
		maincpu.intr |= INTR_IRQ;
	}
	else {
		maincpu.intr &= ~INTR_IRQ;
	}
}

/*
 *	CCPU
 *	Z[u
 */
BOOL FASTCALL maincpu_save(int fileh)
{
	/* vbgtH[Ƃ̃pbLO邽߁A */
	if (!file_byte_write(fileh, maincpu.cc)) {
		return FALSE;
	}

	if (!file_byte_write(fileh, maincpu.dp)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.acc.d)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.x)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.y)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.u)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.s)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.pc)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.intr)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.cycle)) {
		return FALSE;
	}

	if (!file_word_write(fileh, maincpu.total)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	CCPU
 *	[h
 */
BOOL FASTCALL maincpu_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	/* vbgtH[Ƃ̃pbLO邽߁A */
	if (!file_byte_read(fileh, &maincpu.cc)) {
		return FALSE;
	}

	if (!file_byte_read(fileh, &maincpu.dp)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.acc.d)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.x)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.y)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.u)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.s)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.pc)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.intr)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.cycle)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &maincpu.total)) {
		return FALSE;
	}

	return TRUE;
}
