/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ CCPUeI/O ]
 */

#include <string.h>
#include "xm7.h"
#include "mainetc.h"
#include "keyboard.h"
#include "opn.h"
#include "device.h"
#include "event.h"

/*
 *	O[o [N
 */
BOOL key_irq_flag;					/* L[{[h荞 v */
BOOL key_irq_mask;					/* L[{[h荞 }XN */
BOOL lp_irq_flag;					/* v^荞 v */
BOOL lp_irq_mask;					/* v^荞 }XN */
BOOL timer_irq_flag;				/* ^C}[荞 v */
BOOL timer_irq_mask;				/* ^C}[荞 }XN */

BOOL mfd_irq_flag;					/* FDC荞 tO */
BOOL mfd_irq_mask;					/* FDC荞 }XN */
BOOL txrdy_irq_flag;				/* TxRDY荞 tO */
BOOL txrdy_irq_mask;				/* TxRDY荞 }XN */
BOOL rxrdy_irq_flag;				/* RxRDY荞 tO */
BOOL rxrdy_irq_mask;				/* RxRDY荞 }XN */
BOOL syndet_irq_flag;				/* SYNDET荞 tO */
BOOL syndet_irq_mask;				/* SYNDET荞 }XN */

BOOL opn_irq_flag;					/* OPN荞 tO */
BOOL whg_irq_flag;					/* WHG荞 tO */
BOOL thg_irq_flag;					/* THG荞 tO */
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
BOOL dma_irq_flag;					/* DMA荞 tO */
#endif
BOOL beep_flag;						/* BEEPtO */
BOOL speaker_flag;					/* Xs[JtO */
#if XM7_VER == 1
BOOL banksel_en;					/* oN؂芷Cl[utO */
#endif
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
BOOL sfd_drq_flag;					/* SFD DRQpXtO */
BOOL sfd_firq_mode;					/* SFD FIRQ荞݃[h؂芷tO */
BOOL sfd_irq_flag;					/* SFDC荞 tO */
BOOL sfd_irq_mask;					/* SFDC荞 }XN */
BOOL sfdc_irq;						/* SFDC荞ݏLEtO */
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */
#if XM7_VER == 1 && defined(XM7DASH)
BOOL mfdc_irq;						/* MFDC荞ݏLEtO */
#endif	/* XM7_VER == 1 && defined(XM7DASH) */

/*
 *	X^eBbN [N
 */
#if XM7_VER == 1 && defined(BUBBLE) && !defined(XM7DASH)
static int boot_mode_reset;			/* Zbg̃u[gROMoN */
#endif

/*
 *	vg^Cv錾
 */
BOOL FASTCALL mainetc_event(void);	/* 2.03ms ^C}[Cxg */

/*
 *	CCPU I/O
 *	
 */
BOOL FASTCALL mainetc_init(void)
{
#if XM7_VER == 1
	/* FM-8[h̕W̓oN؂芷 */
	banksel_en = FALSE;
#if defined(BUBBLE) && !defined(XM7DASH)
	boot_mode_reset = BOOT_BASIC;
#endif
#endif

#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
	/* ftHgSFDC荞ݏ */
	sfdc_irq = FALSE;
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */

#if XM7_VER == 1 && defined(XM7DASH)
	/* ftHgMFDC荞ݏ */
	mfdc_irq = FALSE;
#endif	/* XM7_VER == 1 && defined(XM7DASH) */
	return TRUE;
}

/*
 *	CCPU I/O
 *	N[Abv
 */
void FASTCALL mainetc_cleanup(void)
{
}

/*
 *	CCPU I/O
 *	Zbg
 */
void FASTCALL mainetc_reset(void)
{
	/* 荞݃tO */
	key_irq_flag = FALSE;
	key_irq_mask = TRUE;
	lp_irq_flag = FALSE;
	lp_irq_mask = TRUE;
	timer_irq_flag = FALSE;
	timer_irq_mask = TRUE;

	mfd_irq_flag = FALSE;
	mfd_irq_mask = TRUE;
	txrdy_irq_flag = FALSE;
	txrdy_irq_mask = TRUE;
	rxrdy_irq_flag = FALSE;
	rxrdy_irq_mask = TRUE;
	syndet_irq_flag = FALSE;
	syndet_irq_mask = TRUE;

	opn_irq_flag = FALSE;
	whg_irq_flag = FALSE;
	thg_irq_flag = FALSE;
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	dma_irq_flag = FALSE;
#endif

	/* BEEPtO */
	beep_flag = FALSE;
	speaker_flag = FALSE;

#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
	sfd_drq_flag = FALSE;
	sfd_firq_mode = FALSE;
	sfd_irq_flag = FALSE;
	sfd_irq_mask = TRUE;
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */

#if XM7_VER == 1 && defined(XM7DASH)
	if (fm_subtype == FMSUB_FM8) {
		txrdy_irq_mask = FALSE;
		rxrdy_irq_mask = FALSE;
		syndet_irq_mask = FALSE;
		mfd_irq_mask = !mfdc_irq;
 #ifdef SFDC
		sfd_irq_mask = !sfdc_irq;
 #endif
	}
#endif	/* XM7_VER == 1 && defined(XM7DASH) */

#if XM7_VER == 1
	/* Cxgǉ(FM-8[h) */
	if (fm_subtype != FMSUB_FM8) {
		schedule_setevent(EVENT_MAINTIMER, 2034, mainetc_event);
	}
	else {
		schedule_delevent(EVENT_MAINTIMER);
	}
#else
	/* Cxgǉ */
	schedule_setevent(EVENT_MAINTIMER, 2034, mainetc_event);
#endif

	/* ʒm */
	beep_notify();

	/* Zbg̃u[g[hL */
#if XM7_VER == 1 && defined(BUBBLE) && !defined(XM7DASH)
	boot_mode_reset = boot_mode;
#endif
}

/*
 *	BEEPICxg
 */
BOOL FASTCALL mainetc_beep(void)
{
	/* BEEP~ */
	beep_flag = FALSE;

	/* ȃCxg폜́ATRUEɂ */
	schedule_delevent(EVENT_BEEP);

	/* ʒm */
	beep_notify();

	return TRUE;
}

/*
 *	^C}[荞݃Cxg
 */
static BOOL FASTCALL mainetc_event(void)
{
	/* 2.03msƂCLKŁAmask̔]DFFœ͂ */
	timer_irq_flag = !timer_irq_mask;
	maincpu_irq();

	/* CLK4.9152MHzȂ̂ŁA2.0345msPʂŔ */
	if (event[EVENT_MAINTIMER].reload == 2034) {
		schedule_setevent(EVENT_MAINTIMER, 2035, mainetc_event);
	}
	else {
		schedule_setevent(EVENT_MAINTIMER, 2034, mainetc_event);
	}

	return TRUE;
}

/*
 *	FDC荞
 *	(fdc.cR}hEُIAtH[XC^vgŌĂ΂)
 */
void FASTCALL mainetc_fdc(void)
{
#if XM7_VER == 1 && !defined(XM7DASH)
	/* }XNĂ΁AȂ */
	if ((fm_subtype == FMSUB_FM8) || mfd_irq_mask) {
		return;
	}
#else
	/* }XNĂ΁AȂ */
	if (mfd_irq_mask) {
		return;
	}
#endif

	/* CCPUIRQ荞݂ */
	mfd_irq_flag = TRUE;

	/*  */
	maincpu_irq();
}

#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
/*
 *	SFDC荞
 *	(fdc.cR}hEُIAtH[XC^vgŌĂ΂)
 */
void FASTCALL mainetc_sfdc(void)
{
	/* }XNĂ΁AȂ */
	if (sfd_irq_mask) {
		return;
	}

	/* CCPUIRQ荞݂ */
	sfd_irq_flag = TRUE;

	/*  */
	maincpu_irq();
}
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */

/*
 *	LP荞
 *	(tapelp.cv^f[^o͎ɌĂ΂)
 */
void FASTCALL mainetc_lp(void)
{
#if XM7_VER == 1
	/* }XNĂ΁AȂ */
	if ((fm_subtype == FMSUB_FM8) || lp_irq_mask) {
		return;
	}
#else
	/* }XNĂ΁AȂ */
	if (lp_irq_mask) {
		return;
	}
#endif

	/* CCPUIRQ荞݂ */
	lp_irq_flag = TRUE;

	/*  */
	maincpu_irq();
}

/*
 *	CCPU I/O
 *	PoCgǂݏo
 */
BOOL FASTCALL mainetc_readb(WORD addr, BYTE *dat)
{
	BYTE ret;

	switch (addr) {
		/* L[{[h  */
		case 0xfd00:
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				ret = 0xff;
				if (lowspeed_mode) {
					ret &= (BYTE)0xfe;
				}

				*dat = ret;
				return TRUE;
			}
#endif

			if (key_fm7 & 0x0100) {
				ret = 0xff;
			}
			else {
				ret = 0x7f;
			}
#if XM7_VER == 1
			if (lowspeed_mode) {
				ret &= (BYTE)0xfe;
			}
#endif
			*dat = ret;
			return TRUE;

		/* L[{[h  */
		case 0xfd01:
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return FALSE;
			}
#endif

			*dat = (BYTE)(key_fm7 & 0xff);
			key_irq_flag = FALSE;
			maincpu_irq();
			subcpu_firq();
			return TRUE;

		/* IRQv */
		case 0xfd03:
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return FALSE;
			}
#endif

			ret = 0xff;
			if ((key_irq_flag) && !(key_irq_mask)) {
				ret &= ~0x01;
			}
			if (lp_irq_flag) {
				ret &= ~0x02;
				lp_irq_flag = FALSE;
			}
			if (timer_irq_flag) {
				ret &= ~0x04;
				timer_irq_flag = FALSE;
			}
			if (mfd_irq_flag ||
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
				sfd_irq_flag ||
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */
				txrdy_irq_flag ||
				rxrdy_irq_flag ||
				syndet_irq_flag ||
#if XM7_VER >= 3
				dma_irq_flag ||
#endif
				opn_irq_flag ||
				whg_irq_flag ||
				thg_irq_flag) {
				ret &= ~0x08;
			}
			*dat = ret;
			maincpu_irq();
			return TRUE;

		/* BASIC ROM */
		case 0xfd0f:
#if XM7_VER == 1
			if ((fm_subtype == FMSUB_FM8) && !banksel_en) {
				return FALSE;
			}
#endif
			*dat = 0xff;
			basicrom_en = TRUE;

#if XM7_VER == 1 && !defined(XM7DASH)
			/* FM-8[h̓u[g[h؂ւdl */
			if (fm_subtype == FMSUB_FM8) {
				boot_mode = BOOT_BASIC;
			}
#endif
			return TRUE;
	}

	return FALSE;
}

/*
 *	CCPU I/O
 *	PoCg
 */
BOOL FASTCALL mainetc_writeb(WORD addr, BYTE dat)
{
	switch (addr) {
		/* 荞݃}XN */
		case 0xfd02:
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return FALSE;
			}
#endif

			if (dat & 0x80) {
				syndet_irq_mask = FALSE;
			}
			else {
				syndet_irq_mask = TRUE;
			}
			if (dat & 0x40) {
				rxrdy_irq_mask = FALSE;
			}
			else {
				rxrdy_irq_mask = TRUE;
			}
			if (dat & 0x20) {
				txrdy_irq_mask = FALSE;
			}
			else {
				txrdy_irq_mask = TRUE;
			}
			if (dat & 0x10) {
				mfd_irq_mask = FALSE;
			}
			else {
				mfd_irq_mask = TRUE;
			}
			if (dat & 0x04) {
				timer_irq_mask = FALSE;
			}
			else {
				timer_irq_mask = TRUE;
			}
			if (dat & 0x02) {
				lp_irq_mask = FALSE;
			}
			else {
				lp_irq_mask = TRUE;
			}
			if (dat & 0x01) {
				key_irq_mask = FALSE;
			}
			else {
				key_irq_mask = TRUE;
			}
			maincpu_irq();
			subcpu_firq();
			return TRUE;

		/* BEEP */
		case 0xfd03:
			/* Xs[JtȌ */
			if (dat & 0x01) {
				speaker_flag = TRUE;
			}
			else {
				speaker_flag = FALSE;
			}
			if (dat & 0x40) {
				/* PBEEP */
				beep_flag = TRUE;
				schedule_setevent(EVENT_BEEP, 205000, mainetc_beep);

				/* ʒm */
				beep_notify();
			}
			else {
				if (dat & 0x80) {
					/* ABEEP */
					beep_flag = TRUE;
				}
				else {
					/* BEEP OFF */
					beep_flag = FALSE;
				}

				/* ʒm */
				beep_notify();
			}
			return TRUE;

		/* BASIC ROM */
		case 0xfd0f:
#if XM7_VER == 1
			if ((fm_subtype == FMSUB_FM8) && !banksel_en) {
				return FALSE;
			}
#endif
			basicrom_en = FALSE;

#if XM7_VER == 1 && !defined(XM7DASH)
			/* FM-8[h̓u[g[h؂ւdl */
			if (fm_subtype == FMSUB_FM8) {
#if defined(BUBBLE)
				if (boot_mode_reset == BOOT_BASIC) {
				boot_mode = BOOT_DOS;
			}
				else {
					boot_mode = boot_mode_reset;
				}
#else
				boot_mode = BOOT_DOS;
#endif
			}
#endif
			return TRUE;
	}

	return FALSE;
}

/*
 *	CCPU I/O
 *	Z[u
 */
BOOL FASTCALL mainetc_save(int fileh)
{
	if (!file_bool_write(fileh, key_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, key_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, timer_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, timer_irq_mask)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, mfd_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, mfd_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, txrdy_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, txrdy_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, rxrdy_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, rxrdy_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, syndet_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, syndet_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, whg_irq_flag)) {
		return FALSE;
	}

	/* Ver6,8ǉ */
	if (!file_bool_write(fileh, thg_irq_flag)) {
		return FALSE;
	}
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	/* Ver8ǉ */
	if (!file_bool_write(fileh, dma_irq_flag)) {
		return FALSE;
	}
#endif

	if (!file_bool_write(fileh, beep_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, speaker_flag)) {
		return FALSE;
	}

#if XM7_VER == 1
	if (!file_bool_write(fileh, banksel_en)) {
		return FALSE;
	}
#endif

#if XM7_VER == 1 && defined(XM7DASH)
 #ifdef SFDC
	/* XM7 V1.2 / 1MB FDC֘A */
	if (!file_bool_write(fileh, sfd_drq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, sfd_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, sfd_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, sfdc_irq)) {
		return FALSE;
	}
 #endif
	if (!file_bool_write(fileh, mfdc_irq)) {
		return FALSE;
	}
#endif	/* XM7_VER == 1 && defined(XM7DASH) */

	return TRUE;
}

/*
 *	CCPU I/O
 *	[h
 */
BOOL FASTCALL mainetc_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &key_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &key_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &timer_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &timer_irq_mask)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &mfd_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &mfd_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &txrdy_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &txrdy_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &rxrdy_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &rxrdy_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &syndet_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &syndet_irq_mask)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &opn_irq_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &whg_irq_flag)) {
		return FALSE;
	}

	/* Ver6ǉ */
#if XM7_VER >= 2
	if (ver >= 600) {
#else
	if (ver >= 300) {
#endif
		if (!file_bool_read(fileh, &thg_irq_flag)) {
			return FALSE;
		}
	}

#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	/* Ver8ǉ */
 #if XM7_VER >= 3
	if (ver >= 800) {
 #else	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */
	if (ver >= 405) {
 #endif
		if (!file_bool_read(fileh, &dma_irq_flag)) {
			return FALSE;
		}
	}
#endif

	if (!file_bool_read(fileh, &beep_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &speaker_flag)) {
		return FALSE;
	}

#if XM7_VER == 1
#ifdef XM7DASH
	if ((ver >= 303 && ver <= 399) || ver >= 404) {
#else	/* XM7DASH */
	if (ver >= 303) {
#endif	/* XM7DASH */
		if (!file_bool_read(fileh, &banksel_en)) {
			return FALSE;
		}
	}
#endif

#if XM7_VER == 1 && defined(XM7DASH)
 #ifdef SFDC
	/* XM7 V1.2 - Ver4g 1MB FDC֘A */
	if (ver >= 401) {
		if (!file_bool_read(fileh, &sfd_drq_flag)) {
			return FALSE;
		}
	}
	else {
		sfd_drq_flag = FALSE;
	}
	if (ver >= 405) {
		if (!file_bool_read(fileh, &sfd_irq_flag)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &sfd_irq_mask)) {
			return FALSE;
		}
	}
	else {
		sfd_irq_flag = FALSE;
		sfd_irq_mask = TRUE;
	}
 #endif
	if (ver >= 404) {
 #ifdef SFDC
		if (!file_bool_read(fileh, &sfdc_irq)) {
			return FALSE;
		}
 #endif
		if (!file_bool_read(fileh, &mfdc_irq)) {
			return FALSE;
		}
	}
	else {
 #ifdef SFDC
		sfdc_irq = FALSE;
 #endif
		mfdc_irq = FALSE;
	}
#endif	/* XM7_VER == 1 && defined(XM7DASH) */

	/* Cxg */
	schedule_handle(EVENT_MAINTIMER, mainetc_event);
	schedule_handle(EVENT_BEEP, mainetc_beep);

	/* ʒm */
	beep_notify();

	return TRUE;
}
