/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ OPN/WHG/THG (YM2203) / OPNA (YM2608) ]
 */

#include <string.h>
#include "xm7.h"
#include "opn.h"
#include "device.h"
#include "mainetc.h"
#include "event.h"
#include "mouse.h"

/*
 *	O[o [N
 */
BOOL opn_enable;						/* OPNLEtO(7 only) */
BOOL opn_use;							/* OPNgptO */
BOOL whg_enable;						/* WHGLEtO */
BOOL whg_use;							/* WHGgptO */
BOOL thg_enable;						/* THGLEtO */
BOOL thg_use;							/* THGgptO */
#if defined(XM7DASH) && defined(OPNAMODE)
BOOL opna_enable[3];					/* OPNALEtO */
BOOL opna_use[3];						/* OPNAgptO */
#endif

#if defined(XM7DASH) && defined(OPNAMODE)
BYTE opn_reg[3][512];					/* OPNWX^ */
BOOL opn_key[3][6];						/* OPNL[ItO */
#else
BYTE opn_reg[3][256];					/* OPNWX^ */
BOOL opn_key[3][4];						/* OPNL[ItO */
#endif
BOOL opn_timera[3];						/* ^C}[AtO */
BOOL opn_timerb[3];						/* ^C}[BtO */
DWORD opn_timera_tick[3];				/* ^C}[AԊu */
DWORD opn_timerb_tick[3];				/* ^C}[BԊu */
BYTE opn_scale[3];						/* vXP[ */
#if defined(XM7DASH) && defined(OPNAMODE)
BYTE opna_exreg[3];						/* OPNAgWX^tO */
#endif

/*
 *	X^eBbN [N
 */
static BYTE opn_pstate[3];				/* |[g */
#if defined(XM7DASH) && defined(OPNAMODE)
static WORD opn_selreg[3];				/* ZNgWX^ */
#else
static BYTE opn_selreg[3];				/* ZNgWX^ */
#endif
static BYTE opn_seldat[3];				/* ZNgf[^ */
static BOOL opn_timera_int[3];			/* ^C}[AI[o[t[ */
static BOOL opn_timerb_int[3];			/* ^C}[BI[o[t[ */
static BOOL opn_timera_en[3];			/* ^C}[ACl[u */
static BOOL opn_timerb_en[3];			/* ^C}[BCl[u */

/*
 *	vg^Cv錾
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static void FASTCALL opn_notify_no(int no, WORD reg, BYTE dat);
#else
static void FASTCALL opn_notify_no(int no, BYTE reg, BYTE dat);
#endif
static void FASTCALL opn_cleanup_common(int no);
static void FASTCALL opn_reset_common(int no);

static BOOL FASTCALL opn_timera_event(void);
static BOOL FASTCALL whg_timera_event(void);
static BOOL FASTCALL thg_timera_event(void);
static BOOL FASTCALL opn_timerb_event(void);
static BOOL FASTCALL whg_timerb_event(void);
static BOOL FASTCALL thg_timerb_event(void);
static void FASTCALL opn_timera_calc(int no);
static void FASTCALL opn_timerb_calc(int no);


/*
 *	OPN
 *	
 */
BOOL FASTCALL opn_init(void)
{
	memset(opn_reg, 0, sizeof(opn_reg));

	/* ftHgOPNL */
	opn_enable = TRUE;
#if XM7_VER == 1
	opn_use = FALSE;
#else
	if (fm7_ver >= 2) {
		opn_use = TRUE;
	}
	else {
		opn_use = FALSE;
	}
#endif
#if defined(XM7DASH) && defined(OPNAMODE)
	opna_enable[OPN_STD] = TRUE;
	opna_use[OPN_STD] = FALSE;
#endif

	return TRUE;
}

/*
 *	OPN
 *	N[Abv
 */
void FASTCALL opn_cleanup(void)
{
	opn_cleanup_common(OPN_STD);
}

/*
 *	OPN
 *	Zbg
 */
void FASTCALL opn_reset(void)
{
	/* 샂[hɂĎgptOݒ肷 */
#if XM7_VER == 1
	opn_use = FALSE;
#else
	if (fm7_ver >= 2) {
		opn_use = TRUE;
	}
	else {
		opn_use = FALSE;
	}
#endif

	/* Zbg */
	opn_reset_common(OPN_STD);
}

/*
 *	WHG
 *	
 */
BOOL FASTCALL whg_init(void)
{
	/* WHGL */
	whg_enable = TRUE;
	whg_use = FALSE;
#if defined(XM7DASH) && defined(OPNAMODE)
	opna_enable[OPN_WHG] = TRUE;
	opna_use[OPN_WHG] = FALSE;
#endif

	return TRUE;
}

/*
 *	WHG
 *	N[Abv
 */
void FASTCALL whg_cleanup(void)
{
	opn_cleanup_common(OPN_WHG);
}

/*
 *	WHG
 *	Zbg
 */
void FASTCALL whg_reset(void)
{
	/* gptO */
	whg_use = FALSE;

	/* Zbg */
	opn_reset_common(OPN_WHG);
}

/*
 *	THG
 *	
 */
BOOL FASTCALL thg_init(void)
{
	/* THGL */
	thg_enable = FALSE;
	thg_use = FALSE;
#if defined(XM7DASH) && defined(OPNAMODE)
	opna_enable[OPN_THG] = TRUE;
	opna_use[OPN_THG] = FALSE;
#endif

	return TRUE;
}

/*
 *	THG
 *	N[Abv
 */
void FASTCALL thg_cleanup(void)
{
	opn_cleanup_common(OPN_THG);
}

/*
 *	THG
 *	Zbg
 */
void FASTCALL thg_reset(void)
{
	/* gptO */
	thg_use = FALSE;

	/* Zbg */
	opn_reset_common(OPN_THG);
}

/*-[ ʏ(Reset/Cleanup) ]-----------------------------------------------*/

/*
 *	OPN
 *	N[Abvʏ
 */
static void FASTCALL opn_cleanup_common(int no)
{
	BYTE i;

	/* PSG */
	for (i=0; i<6; i++) {
		opn_notify_no(no, i, 0);
	}
	opn_notify_no(no, 7, 0xff);

	/* TL=$7F */
	for (i=0x40; i<0x50; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0x7f);
	}

	/* L[It */
#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<6; i++) {
		opn_notify_no(no, 0x28, i + ((i < 3) ? (0) : (1)));
#else
	for (i=0; i<3; i++) {
		opn_notify_no(no, 0x28, i);
#endif
		opn_key[no][i] = FALSE;
	}
}

/*
 *	OPN
 *	Zbgʏ
 */
static void FASTCALL opn_reset_common(int no)
{
#if defined(XM7DASH) && defined(OPNAMODE)
	WORD i;
#else
	BYTE i;
#endif

	/* WX^NA */
	memset(opn_reg[no], 0, sizeof(opn_reg[no]));

	/* ^C}[OFF */
	opn_timera[no] = FALSE;
	opn_timerb[no] = FALSE;

	/* I/O */
	opn_pstate[no] = OPN_INACTIVE;
	opn_selreg[no] = 0;
	opn_seldat[no] = 0;

	/* foCX */
	opn_timera_int[no] = FALSE;
	opn_timerb_int[no] = FALSE;
	opn_timera_tick[no] = 0;
	opn_timerb_tick[no] = 0;
	opn_timera_en[no] = FALSE;
	opn_timerb_en[no] = FALSE;
	opn_scale[no] = 3;

	/* PSG */
	for (i=0; i<14;i++) {
		if (i == 7) {
			opn_notify_no(no, i, 0xff);
			opn_reg[no][i] = 0xff;
		}
		else {
			opn_notify_no(no, i, 0);
		}
	}

	/* MUL,DT */
	for (i=0x30; i<0x40; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0);
	}

	/* TL=$7F */
	for (i=0x40; i<0x50; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0x7f);
		opn_reg[no][i] = 0x7f;
	}

	/* AR=$1F */
	for (i=0x50; i<0x60; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0x1f);
		opn_reg[no][i] = 0x1f;
	}

	/* ̑ */
	for (i=0x60; i<0xb4; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0);
	}

	/* SL,RR */
	for (i=0x80; i<0x90; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0xff);
		opn_reg[no][i] = 0xff;
	}

#if defined(XM7DASH) && defined(OPNAMODE)
	/* MUL,DT */
	for (i=0x130; i<0x140; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0);
	}

	/* TL=$7F */
	for (i=0x140; i<0x150; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0x7f);
		opn_reg[no][i] = 0x7f;
	}

	/* AR=$1F */
	for (i=0x150; i<0x160; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0x1f);
		opn_reg[no][i] = 0x1f;
	}

	/* ̑ */
	for (i=0x160; i<0x1b4; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0);
	}

	/* SL,RR */
	for (i=0x180; i<0x190; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0xff);
		opn_reg[no][i] = 0xff;
	}

	/* o */
	for (i=0xb4; i<0xb8; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0xc0);
		opn_reg[no][i] = 0xc0;
	}
	for (i=0x1b4; i<0x1b8; i++) {
		if ((i & 0x03) == 3) {
			continue;
		}
		opn_notify_no(no, i, 0xc0);
		opn_reg[no][i] = 0xc0;
	}
#endif

	/* L[It */
#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<6; i++) {
		opn_notify_no(no, 0x28, i + ((i < 3) ? (0) : (1)));
#else
	for (i=0; i<3; i++) {
		opn_notify_no(no, 0x28, i);
#endif
		opn_key[no][i] = FALSE;
	}

	/* [h */
	opn_notify_no(no, 0x27, 0);

#if defined(XM7DASH) && defined(OPNAMODE)
	/* OPNAgWX^tO */
	opna_use[no] = FALSE;
	opna_exreg[no] = 0;
#endif
}

/*-[ ʏ(VM I/F) ]------------------------------------------------------*/

/*
 *	OPN
 *	WX^ݒʒm
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static void FASTCALL opn_notify_no(int no, WORD reg, BYTE dat)
#else
static void FASTCALL opn_notify_no(int no, BYTE reg, BYTE dat)
#endif
{
	/* e̒ʒm֐ĂԂ */
	switch (no) {
		case OPN_STD:
			opn_notify(reg, dat);
			break;
		case OPN_WHG:
			whg_notify(reg, dat);
			break;
		case OPN_THG:
			thg_notify(reg, dat);
			break;
		default:
			ASSERT(FALSE);
			break;
	}
}

/*-[ ʏ(^C}[) ]----------------------------------------------------*/

/*
 *	OPN
 *	IRQtOʏ
 */
static void FASTCALL opn_set_irq_flag(int no, BOOL flag)
{
	switch (no) {
		case OPN_STD:
			opn_irq_flag = flag;
			break;
		case OPN_WHG:
			whg_irq_flag = flag;
			break;
		case OPN_THG:
			thg_irq_flag = flag;
			break;
		default:
			ASSERT(FALSE);
			break;
	}

#if XM7_VER == 1
	if (fm_subtype == FMSUB_FM8) {
		/* 肦Ȃ */
		return;
	}
#endif
	maincpu_irq();
}

/*
 *	OPN
 *	^C}[AI[ot[ʏ
 */
static BOOL FASTCALL opn_timera_event_main(int no)
{
	ASSERT ((no >= 0) && (no <= 2));

	/* Cl[u */
	if (opn_timera_en[no]) {
		/* I[o[t[ANVL */
		if (opn_timera[no]) {
			opn_timera[no] = FALSE;
			opn_timera_int[no] = TRUE;

			/* 荞݂ */
			opn_set_irq_flag(no, TRUE);
		}
	}

	/* CSM[hł̃L[I */
	opn_notify_no(no, 0xff, 0);

	/* ^C}[Đݒ */
	opn_timera_calc(no);

	/* ^C}[͉񂵑 */
	return TRUE;
}

/*
 *	OPN
 *	^C}[BI[ot[ʏ
 */
static BOOL FASTCALL opn_timerb_event_main(int no)
{
	ASSERT ((no >= 0) && (no <= 2));

	/* Cl[u */
	if (opn_timerb_en[no]) {
		/* I[o[t[ANVL */
		if (opn_timerb[no]) {
			/* tOύX */
			opn_timerb[no] = FALSE;
			opn_timerb_int[no] = TRUE;

			/* 荞݂ */
			opn_set_irq_flag(no, TRUE);
		}
	}

	/* ^C}[Đݒ */
	opn_timerb_calc(no);

	/* ^C}[͉񂵑 */
	return TRUE;
}

/*
 *	OPN
 *	^C}[AC^[oZoʏ
 */
static void FASTCALL opn_timera_calc(int no)
{
	DWORD t;
	BYTE temp;

	ASSERT ((no >= 0) && (no <= 2));

	t = opn_reg[no][0x24];
	t *= 4;
	temp = (BYTE)(opn_reg[no][0x25] & 3);
	t |= temp;
	t &= 0x3ff;
	t = (1024 - t);
	t *= opn_scale[no];
	t *= 12;
	t *= 10000;
	t /= OPN_CLOCK;

	/* ^C}[lݒ */
	if (t != opn_timera_tick[no]) {
		opn_timera_tick[no] = t;
		switch (no) {
			case OPN_STD:
				schedule_setevent(EVENT_OPN_A, opn_timera_tick[OPN_STD],
					opn_timera_event);
				break;
			case OPN_WHG:
				schedule_setevent(EVENT_WHG_A, opn_timera_tick[OPN_WHG],
					whg_timera_event);
				break;
			case OPN_THG:
				schedule_setevent(EVENT_THG_A, opn_timera_tick[OPN_THG],
					thg_timera_event);
				break;
			default:
				ASSERT(FALSE);
				break;
		}
	}
}

/*
 *	OPN
 *	^C}[BC^[oZoʏ
 */
static void FASTCALL opn_timerb_calc(int no)
{
	DWORD t;

	ASSERT ((no >= 0) && (no <= 2));

	t = opn_reg[no][0x26];
	t = (256 - t);
	t *= 192;
	t *= opn_scale[no];
	t *= 10000;
	t /= OPN_CLOCK;

	/* ^C}[lݒ */
	if (t != opn_timerb_tick[no]) {
		opn_timerb_tick[no] = t;
		switch (no) {
			case OPN_STD:
				schedule_setevent(EVENT_OPN_B, opn_timerb_tick[OPN_STD],
					opn_timerb_event);
				break;
			case OPN_WHG:
				schedule_setevent(EVENT_WHG_B, opn_timerb_tick[OPN_WHG],
					whg_timerb_event);
				break;
			case OPN_THG:
				schedule_setevent(EVENT_THG_B, opn_timerb_tick[OPN_THG],
					thg_timerb_event);
				break;
			default:
				ASSERT(FALSE);
				break;
		}
	}
}

/*-[ ʏ(WX^AC) ]----------------------------------------------*/

/*
 *	OPN
 *	WX^ACǂݏo
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static BYTE FASTCALL opn_readreg(int no, WORD reg)
#else
static BYTE FASTCALL opn_readreg(int no, BYTE reg)
#endif
{
	/* WX^rbg}XNf[^(PSG) */
	static const BYTE opn_bitmask[16] = {
		0xff,	0x0f,	0xff,	0x0f,
		0xff,	0x0f,	0x1f,	0xff,
		0x1f,	0x1f,	0x1f,	0xff,
		0xff,	0x0f,	0xff,	0xff,
	};

	ASSERT ((no >= 0) && (no <= 2));

	/* FM͓ǂݏoȂ */
	if (reg >= 0x10) {
		return 0xff;
	}

	return (BYTE)(opn_reg[no][reg] & opn_bitmask[reg]);
}

/*
 *	OPN
 *	WX^AC֏
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static void FASTCALL opn_writereg(int no, WORD reg, BYTE dat)
#else
static void FASTCALL opn_writereg(int no, BYTE reg, BYTE dat)
#endif
{
#ifdef MOUSE
	BOOL strobe;
	BYTE mask;
#endif

	ASSERT ((no >= 0) && (no <= 2));

	if (no == OPN_STD) {
		/* WOPNtOI */
		opn_use = TRUE;
	}

	if (no == OPN_WHG) {
		/* WHGtOI */
		whg_use = TRUE;
	}

	/* ^C}[ */
	/* ̃WX^͔ɓBǂȂ܂܈Ăl唼ł́H */
	if (reg == 0x27) {
		/* I[o[t[tÕNA */
		if (dat & 0x10) {
			opn_timera_int[no] = FALSE;
		}
		if (dat & 0x20) {
			opn_timerb_int[no] = FALSE;
		}

		/* A荞݂𗎂Ƃ */
		if (!opn_timera_int[no] && !opn_timerb_int[no]) {
			opn_set_irq_flag(no, FALSE);
		}

		/* ^C}[A */
		if (dat & 0x01) {
			/* 01Ń^C}[l[hAȊOł^C}[on */
			if ((opn_reg[no][0x27] & 0x01) == 0) {
				if (!opn_timera[no]) {
					/* I[o[t[ς݂̏ꍇAĐݒ */
					opn_timera_tick[no] = 0;
				}
				opn_timera_calc(no);
			}
			opn_timera_en[no] = TRUE;
		}
		else {
			opn_timera_en[no] = FALSE;
		}
		if (dat & 0x04) {
			opn_timera[no] = TRUE;
		}
		else {
			opn_timera[no] = FALSE;
		}

		/* ^C}[B */
		if (dat & 0x02) {
			/* 01Ń^C}[l[hAȊOł^C}[on */
			if ((opn_reg[no][0x27] & 0x02) == 0) {
				if (!opn_timerb[no]) {
					/* I[o[t[ς݂̏ꍇAĐݒ */
					opn_timerb_tick[no] = 0;
				}
				opn_timerb_calc(no);
			}
			opn_timerb_en[no] = TRUE;
		}
		else {
			opn_timerb_en[no] = FALSE;
		}
		if (dat & 0x08) {
			opn_timerb[no] = TRUE;
		}
		else {
			opn_timerb[no] = FALSE;
		}
	}

	/* f[^L */
	opn_reg[no][reg] = dat;

	switch (reg) {
#ifdef MOUSE
		/* o̓|[g */
		case 0x0f:
			/* }EXG~[V͕WOPN̂ݗL */
			if ((no == OPN_STD) && mos_capture) {
				if (mos_port == 1) {
					mask = 0x10;
				}
				else {
					mask = 0x20;
				}

				if (opn_reg[OPN_STD][15] & mask) {
					strobe = TRUE;
				}
				else {
					strobe = FALSE;
				}
				mos_strobe_signal(strobe);
			}
			return;
#endif

		/* vXP[P */
		case 0x2d:
			if (opn_scale[no] != 3) {
				opn_scale[no] = 6;
				opn_timera_calc(no);
				opn_timerb_calc(no);
			}
			return;

		/* vXP[Q */
		case 0x2e:
			opn_scale[no] = 3;
			opn_timera_calc(no);
			opn_timerb_calc(no);
			return;

		/* vXP[R */
		case 0x2f:
			opn_scale[no] = 2;
			opn_timera_calc(no);
			opn_timerb_calc(no);
			return;

		/* ^C}[A */
		case 0x24:
		case 0x25:
			opn_timera_tick[no] = 0;
#if defined(XM7DASH) && defined(FMDLL)
			break;
#else
			return;
#endif

		/* ^C}[B */
		case 0x26:
			opn_timerb_tick[no] = 0;
#if defined(XM7DASH) && defined(FMDLL)
			break;
#else
			return;
#endif

#if defined(XM7DASH) && defined(OPNAMODE)
		/* SCH */
		case 0x29:
			if (dat & 0x80) {
				opna_use[no] = TRUE;
				opn_notify_no(no, 0xb4, opn_reg[no][0xb4]);
				opn_notify_no(no, 0xb5, opn_reg[no][0xb5]);
				opn_notify_no(no, 0xb6, opn_reg[no][0xb6]);
			}
			else {
				opna_use[no] = FALSE;
				opn_notify_no(no, 0xb4, opn_reg[no][0xb4] | 0xc0);
				opn_notify_no(no, 0xb5, opn_reg[no][0xb5] | 0xc0);
				opn_notify_no(no, 0xb6, opn_reg[no][0xb6] | 0xc0);
			}
			break;

		/* LR */
		case 0xb4:
		case 0xb5:
		case 0xb6:
			if (!opna_use[no]) {
				dat |= 0xc0;
			}
			break;
#endif
	}

	/* o͐i */
#if defined(XM7DASH) && defined(OPNAMODE)
	if ((reg >= 0x0e) && (reg <= 0x0f)) {
		return;
	}
	if ((reg >= 0x10) && (reg <= 0x11) && !opna_enable[no]) {
		return;
	}
	if (reg == 0x12) {
		return;
	}
	if ((reg >= 0x13) && (reg <= 0x1d) && !opna_enable[no]) {
		return;
	}
	if ((reg >= 0x1e) && (reg <= 0x21)) {
		return;
	}
	if ((reg == 0x22) && !opna_enable[no]) {
		return;
	}
#if defined(XM7DASH) && defined(FMDLL)
	if (reg == 0x23) {						/* ODLL */
#else
	if ((reg >= 0x23) && (reg <= 0x26)) {
#endif
		return;
	}
	if ((reg == 0x28) && (dat & 0x04) && !opna_enable[no]) {
		return;
	}
	if ((reg >= 0x2a) && (reg <= 0x2c)) {
		return;
	}
#else
#if defined(XM7DASH) && defined(FMDLL)
	if ((reg >= 14) && (reg <= 0x23)) {		/* ODLL */
#else
	if ((reg >= 14) && (reg <= 0x26)) {
#endif
		return;
	}
	if ((reg >= 0x29) && (reg <= 0x2c)) {
		return;
	}
#endif

	/* L[I */
	if (reg == 0x28) {
		if (dat >= 16) {
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_key[no][(dat & 0x07) - (((dat & 0x07) >= 3) ? (1) : (0))] = TRUE;
#else
			opn_key[no][dat & 0x03] = TRUE;
#endif
		}
		else {
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_key[no][(dat & 0x07) - (((dat & 0x07) >= 3) ? (1) : (0))] = FALSE;
#else
			opn_key[no][dat & 0x03] = FALSE;
#endif
		}
	}

	/* o */
	opn_notify_no(no, reg, dat);
}

/*-[ ʏ(I/O|[g) ]---------------------------------------------------*/

/*
 *	OPN
 *	f[^WX^ǂݍ݋ʏ
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static void FASTCALL opn_read_data_reg(int no, BYTE *dat, BYTE exreg)
#else
static void FASTCALL opn_read_data_reg(int no, BYTE *dat)
#endif
{
#ifdef MOUSE
	BYTE port, trigger;
#endif

	ASSERT ((no >= 0) && (no <= 2));
#if defined(XM7DASH) && defined(OPNAMODE)
	ASSERT (exreg <= 1);
#endif

	switch (opn_pstate[no]) {
		/* ʏR}h */
		case OPN_INACTIVE:
		case OPN_READDAT:
		case OPN_WRITEDAT:
		case OPN_ADDRESS:
			*dat = opn_seldat[no];
			break;

		/* Xe[^Xǂݏo */
		case OPN_READSTAT:
			*dat = 0;
			if (opn_timera_int[no]) {
				*dat |= 0x01;
			}
			if (opn_timerb_int[no]) {
				*dat |= 0x02;
			}
			break;

		/* WCXeBbNǂݎ */
		case OPN_JOYSTICK:
#if defined(XM7DASH) && defined(OPNAMODE)
			if (exreg) {
				*dat = 0xff;
				break;
			}
#endif
			if (opn_selreg[no] == 14) {
				/* WCXeBbN|[gG~[V͕WOPN̂ */
				if (no == OPN_STD) {
#ifdef MOUSE
					if (mos_capture) {
						/* }EX */
						if (mos_port == 1) {
							port = 0x00;
							trigger = (BYTE)(opn_reg[no][15] & 0x03);
						}
						else {
							port = 0x40;
							trigger = (BYTE)((opn_reg[no][15] >> 2) & 0x03);
						}

						if ((opn_reg[no][15] & 0xc0) == port) {
							*dat = (BYTE)(mos_readdata(trigger) | 0xc0);
							break;
						}
					}
#endif
					if ((opn_reg[no][15] & 0xf0) == 0x20) {
						/* WCXeBbNP */
						*dat = (BYTE)(~joy_request(0) | 0xc0);
						break;
					}
					if ((opn_reg[no][15] & 0xf0) == 0x50) {
						/* WCXeBbNQ */
						*dat = (BYTE)(~joy_request(1) | 0xc0);
						break;
					}
				}

				/* ȊO */
				*dat = 0xff;
			}
			else {
				/* WX^14łȂ΁AFFȊOԂ */
				/* HOW MANY ROBOT΍ */
				*dat = 0;
			}
			break;
	}
}

/*
 *	OPN
 *	荞݃Xe[^Xǂݍ݋ʏ
 */
static void FASTCALL opn_read_interrupt_reg(int no, BYTE *dat)
{
	ASSERT ((no >= 0) && (no <= 2));

	if (opn_timera_int[no] || opn_timerb_int[no]) {
		*dat = 0xf7;
	}
	else {
		*dat = 0xff;
	}
}

/*
 *	OPN
 *	R}hWX^݋ʏ
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static void FASTCALL opn_write_command_reg(int no, BYTE dat, BYTE exreg)
#else
static void FASTCALL opn_write_command_reg(int no, BYTE dat)
#endif
{
	ASSERT ((no >= 0) && (no <= 2));

	switch (dat & 0x0f) {
		/* CANeBu(`ȂAf[^WX^) */
		case OPN_INACTIVE:
			opn_pstate[no] = OPN_INACTIVE;
			break;
		/* f[^ǂݏo */
		case OPN_READDAT:
			opn_pstate[no] = OPN_READDAT;
			opn_seldat[no] = opn_readreg(no, opn_selreg[no]);
			break;
		/* f[^ */
		case OPN_WRITEDAT:
			opn_pstate[no] = OPN_WRITEDAT;
			opn_writereg(no, opn_selreg[no], opn_seldat[no]);
			break;
		/* b`AhX */
		case OPN_ADDRESS:
			opn_pstate[no] = OPN_ADDRESS;
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_selreg[no] = (WORD)opn_seldat[no] + 0x100 * (WORD)exreg;
#else
			opn_selreg[no] = opn_seldat[no];
#endif

			/* vXP[̓AhXŵ݂ok */
			if ((opn_selreg[no] >= 0x2d) && (opn_selreg[no] <= 0x2f)) {
				opn_seldat[no] = 0;
				opn_writereg(no, opn_selreg[no], opn_seldat[no]);
			}
			break;
		/* [hXe[^X */
		case OPN_READSTAT:
			opn_pstate[no] = OPN_READSTAT;
			break;
		/* WCXeBbNǂݎ */
		case OPN_JOYSTICK:
			opn_pstate[no] = OPN_JOYSTICK;
			break;
	}
}

/*
 *	OPN
 *	f[^WX^݋ʏ
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static void FASTCALL opn_write_data_reg(int no, BYTE dat, BYTE exreg)
#else
static void FASTCALL opn_write_data_reg(int no, BYTE dat)
#endif
{
	ASSERT ((no >= 0) && (no <= 2));
#if defined(XM7DASH) && defined(OPNAMODE)
	ASSERT (exreg <= 1);
#endif

	/* f[^L */
	opn_seldat[no] = dat;

	/* CANeBuȊȌꍇ́A̓s */
	switch (opn_pstate[no]) {
		/* f[^ */
		case OPN_WRITEDAT:
			opn_writereg(no, opn_selreg[no], opn_seldat[no]);
			break;
		/* b`AhX */
		case OPN_ADDRESS:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_selreg[no] = (WORD)opn_seldat[no] + 0x100 * (WORD)exreg;
#else
			opn_selreg[no] = opn_seldat[no];
#endif

			/* vXP[̓AhXŵ݂ok */
			if ((opn_selreg[no] >= 0x2d) && (opn_selreg[no] <= 0x2f)) {
				opn_seldat[no] = 0;
				opn_writereg(no, opn_selreg[no], opn_seldat[no]);
			}
			break;
	}
}

/*-[ ʏ(Xe[g) ]----------------------------------------------------*/

/*
 *	OPN
 *	Xe[gZ[uʏ
 */
static BOOL FASTCALL opn_save_common(int no, int fileh)
{
	ASSERT ((no >= 0) && (no <= 2));

#if defined(XM7DASH) && defined(OPNAMODE)
	if (!file_write(fileh, opn_reg[no], 512)) {
#else
	if (!file_write(fileh, opn_reg[no], 256)) {
#endif
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_timera[no])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_timerb[no])) {
		return FALSE;
	}
	if (!file_dword_write(fileh, opn_timera_tick[no])) {
		return FALSE;
	}
	if (!file_dword_write(fileh, opn_timerb_tick[no])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, opn_scale[no])) {
		return FALSE;
	}

	if (!file_byte_write(fileh, opn_pstate[no])) {
		return FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (!file_word_write(fileh, opn_selreg[no])) {
#else
	if (!file_byte_write(fileh, opn_selreg[no])) {
#endif
		return FALSE;
	}
	if (!file_byte_write(fileh, opn_seldat[no])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_timera_int[no])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_timerb_int[no])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_timera_en[no])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_timerb_en[no])) {
		return FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (!file_byte_write(fileh, opna_exreg[no])) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	OPN
 *	Xe[g[hʏ
 */
#if defined(XM7DASH) && defined(OPNAMODE)
static BOOL FASTCALL opn_load_common(int no, int fileh, BOOL bflag)
#else
static BOOL FASTCALL opn_load_common(int no, int fileh)
#endif
{
#if defined(XM7DASH) && defined(OPNAMODE)
	BOOL flag = bflag;
#endif

	ASSERT ((no >= 0) && (no <= 2));

#if defined(XM7DASH) && defined(OPNAMODE)
	if (flag) {
		if (!file_read(fileh, opn_reg[no], 512)) {
			return FALSE;
		}
	}
	else {
		memset(opn_reg[no], 0, sizeof(opn_reg[no]));
		if (!file_read(fileh, opn_reg[no], 256)) {
			return FALSE;
		}
		opn_reg[no][0xb4] = 0xc0;
		opn_reg[no][0xb5] = 0xc0;
		opn_reg[no][0xb6] = 0xc0;
	}
#else
	if (!file_read(fileh, opn_reg[no], 256)) {
		return FALSE;
	}
#endif
	if (!file_bool_read(fileh, &opn_timera[no])) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &opn_timerb[no])) {
		return FALSE;
	}
	if (!file_dword_read(fileh, &opn_timera_tick[no])) {
		return FALSE;
	}
	if (!file_dword_read(fileh, &opn_timerb_tick[no])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &opn_scale[no])) {
		return FALSE;
	}

	if (!file_byte_read(fileh, &opn_pstate[no])) {
		return FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (flag) {
		if (!file_word_read(fileh, &opn_selreg[no])) {
			return FALSE;
		}
	}
	else {
		BYTE tmp;
		if (!file_byte_read(fileh, &tmp)) {
			return FALSE;
		}
		opn_selreg[no] = (WORD)tmp;
	}
#else
	if (!file_byte_read(fileh, &opn_selreg[no])) {
		return FALSE;
	}
#endif
	if (!file_byte_read(fileh, &opn_seldat[no])) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &opn_timera_int[no])) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &opn_timerb_int[no])) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &opn_timera_en[no])) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &opn_timerb_en[no])) {
		return FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (flag) {
		if (!file_byte_read(fileh, &opna_exreg[no])) {
			return FALSE;
		}
	}
	else {
		opna_exreg[no] = 0;
	}
	if (opn_reg[no][0x29] & 0x80) {
		opna_use[no] = TRUE;
	}
	else {
		opna_use[no] = FALSE;
	}
#endif

	return TRUE;
}

/*-[ WOPN ]---------------------------------------------------------------*/

/*
 *	OPN
 *	^C}[AI[ot[
 */
static BOOL FASTCALL opn_timera_event(void)
{
	return opn_timera_event_main(OPN_STD);
}

/*
 *	OPN
 *	^C}[BI[ot[
 */
static BOOL FASTCALL opn_timerb_event(void)
{
	return opn_timerb_event_main(OPN_STD);
}

/*
 *	OPN
 *	PoCgǂݏo
 */
BOOL FASTCALL opn_readb(WORD addr, BYTE *dat)
{
#if defined(XM7DASH) && defined(OPNAMODE)
	BYTE exreg = opna_exreg[OPN_STD];
#endif

	/* FM-7[hOPNL`FbN */
#if XM7_VER >= 2
	if ((fm7_ver == 1) && (!opn_enable)) {
#else
	if ((fm_subtype == FMSUB_FM8) || (!opn_enable)) {
#endif
		return FALSE;
	}

	switch (addr) {
		/* R}hWX^͓ǂݏo֎~ */
#if XM7_VER >= 2
		case 0xfd0d:
			/* FM-7[hłTHG SSGgp邽ߖ */
			if (fm7_ver == 1) {
				return FALSE;
			}
#endif
		case 0xfd15:
			*dat = 0xff;
			return TRUE;

		/* f[^WX^ */
#if XM7_VER >= 2
		case 0xfd0e:
			/* FM-7[hłTHG SSGgp邽ߖ */
			if (fm7_ver == 1) {
				return FALSE;
			}
#if defined(XM7DASH) && defined(OPNAMODE)
			exreg = 0;
#endif
#endif
		case 0xfd16:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_read_data_reg(OPN_STD, dat, exreg);
#else
			opn_read_data_reg(OPN_STD, dat);
#endif
			return TRUE;

		/* g荞݃Xe[^X */
		case 0xfd17:
			opn_read_interrupt_reg(OPN_STD, dat);
			return TRUE;
	}

	return FALSE;
}

/*
 *	OPN
 *	PoCg
 */
BOOL FASTCALL opn_writeb(WORD addr, BYTE dat)
{
#if defined(XM7DASH) && defined(OPNAMODE)
	BYTE exreg = opna_exreg[OPN_STD];
#endif

	/* FM-7[hOPNL`FbN */
#if XM7_VER >= 2
	if ((fm7_ver == 1) && (!opn_enable)) {
#else
	if ((fm_subtype == FMSUB_FM8) || (!opn_enable)) {
#endif
		return FALSE;
	}

	switch (addr) {
#if defined(XM7DASH) && defined(OPNAMODE)
		case 0xfd14:
			opna_exreg[OPN_STD] = (dat & 1);
			return TRUE;
#endif

		/* OPNR}hWX^ */
#if XM7_VER >= 2
		case 0xfd0d:
			/* FM-7[hłTHG SSGgp邽ߖ */
			if (fm7_ver == 1) {
				return FALSE;
			}
			/* PSGANZX͉2bit̂ݗL */
			dat &= 0x03;
#if defined(XM7DASH) && defined(OPNAMODE)
			exreg = 0;
#endif
#endif
		case 0xfd15:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_write_command_reg(OPN_STD, dat, exreg);
#else
			opn_write_command_reg(OPN_STD, dat);
#endif
			return TRUE;

		/* f[^WX^ */
#if XM7_VER >= 2
		case 0xfd0e:
			/* FM-7[hłTHG SSGgp邽ߖ */
			if (fm7_ver == 1) {
				return FALSE;
			}
#if defined(XM7DASH) && defined(OPNAMODE)
			exreg = 0;
#endif
#endif
		case 0xfd16:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_write_data_reg(OPN_STD, dat, exreg);
#else
			opn_write_data_reg(OPN_STD, dat);
#endif
			return TRUE;
	}

	return FALSE;
}

/*
 *	OPN
 *	Z[u
 */
BOOL FASTCALL opn_save(int fileh)
{
	if (!file_bool_write(fileh, opn_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, opn_use)) {
		return FALSE;
	}
	if (!opn_save_common(OPN_STD, fileh)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	OPN
 *	[h
 */
#if defined(XM7DASH) && defined(OPNAMODE)
BOOL FASTCALL opn_load(int fileh, int ver, BOOL bflag)
#else
BOOL FASTCALL opn_load(int fileh, int ver)
#endif
{
	int i;

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	/* Ver9.16/7.16/3.06g */
#if XM7_VER >= 3
	if ((ver >= 916) || ((ver >= 716) && (ver <= 799))) {
#elif XM7_VER >= 2
	if ((ver >= 716) && (ver <= 799)) {
#elif defined(XM7DASH)
	if (((ver >= 404) && (ver <= 499)) || ((ver >= 306) && (ver <= 399))) {
#else
	if (ver >= 306) {
#endif
		if (!file_bool_read(fileh, &opn_enable)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &opn_use)) {
			return FALSE;
		}
	}
#ifdef XM7DASH
	else {
		opn_enable = TRUE;
#if XM7_VER == 1
		opn_use = FALSE;
#else
		if (fm7_ver >= 2) {
			opn_use = TRUE;
		}
		else {
			opn_use = FALSE;
		}
#endif
	}
#endif

#if defined(XM7DASH) && defined(OPNAMODE)
	if (!opn_load_common(OPN_STD, fileh, bflag)) {
#else
	if (!opn_load_common(OPN_STD, fileh)) {
#endif
		return FALSE;
	}

	/* OPNWX^ */
	opn_notify(0x27, (BYTE)(opn_reg[OPN_STD][0x27]));

#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<6; i++) {
		opn_notify(0x28, (BYTE)(i + ((i < 3) ? (0) : (1))));
#else
	for (i=0; i<3; i++) {
		opn_notify(0x28, (BYTE)i);
#endif
		opn_key[OPN_STD][i] = FALSE;
	}

#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<13; i++) {
		if ((i < 8) || (i > 10)) {
			opn_notify((WORD)i, opn_reg[OPN_STD][i]);
		}
		else {
			opn_notify((WORD)i, 0);
		}
	}

	for (i=0x10; i<0x1d; i++) {
		if (i == 0x10) {
			opn_notify((WORD)i, opn_reg[OPN_STD][i] | 0x80);
		}
		else if ((i >= 0x12) && (i <= 0x17)) {
			continue;
		}
		else {
			opn_notify((WORD)i, opn_reg[OPN_STD][i]);
		}
	}

	opn_notify(0x22, (BYTE)(opn_reg[OPN_STD][0x22]));
	opn_notify(0x29, (BYTE)(opn_reg[OPN_STD][0x29]));

	for (i=0x30; i<0xb8; i++) {
		if (!(opn_reg[OPN_STD][0x29] & 0x80) && (i >= 0xb4)) {
			opn_notify((WORD)i, opn_reg[OPN_STD][i] | 0xc0);
		}
		else {
			opn_notify((WORD)i, opn_reg[OPN_STD][i]);
		}
	}
	for (i=0x130; i<0x1b8; i++) {
		opn_notify((WORD)i, opn_reg[OPN_STD][i]);
	}
#else
	for (i=0; i<13; i++) {
		if ((i < 8) || (i > 10)) {
			opn_notify((BYTE)i, opn_reg[OPN_STD][i]);
		}
		else {
			opn_notify((BYTE)i, 0);
		}
	}

	for (i=0x30; i<0xb4; i++) {
		opn_notify((BYTE)i, opn_reg[OPN_STD][i]);
	}
#endif

	/* Cxg */
	schedule_handle(EVENT_OPN_A, opn_timera_event);
	schedule_handle(EVENT_OPN_B, opn_timerb_event);

	return TRUE;
}

/*-[ WHG OPN ]---------------------------------------------------------------*/

/*
 *	WHG
 *	^C}[AI[ot[
 */
static BOOL FASTCALL whg_timera_event(void)
{
	return opn_timera_event_main(OPN_WHG);
}

/*
 *	WHG
 *	^C}[BI[ot[
 */
static BOOL FASTCALL whg_timerb_event(void)
{
	return opn_timerb_event_main(OPN_WHG);
}

/*
 *	WHG
 *	PoCgǂݏo
 */
BOOL FASTCALL whg_readb(WORD addr, BYTE *dat)
{
	/* LtO`FbNAȂ牽Ȃ */
	if (!whg_enable) {
		return FALSE;
	}
#if XM7_VER == 1
	if (fm_subtype == FMSUB_FM8) {
		return FALSE;
	}
#endif

	switch (addr) {
		/* R}hWX^͓ǂݏo֎~ */
		case 0xfd45:
			*dat = 0xff;
			return TRUE;

		/* f[^WX^ */
		case 0xfd46:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_read_data_reg(OPN_WHG, dat, opna_exreg[OPN_WHG]);
#else
			opn_read_data_reg(OPN_WHG, dat);
#endif
			return TRUE;

		/* g荞݃Xe[^X */
		case 0xfd47:
			opn_read_interrupt_reg(OPN_WHG, dat);
			return TRUE;
	}

	return FALSE;
}

/*
 *	WHG
 *	PoCg
 */
BOOL FASTCALL whg_writeb(WORD addr, BYTE dat)
{
	/* LtO`FbNAȂ牽Ȃ */
	if (!whg_enable) {
		return FALSE;
	}
#if XM7_VER == 1
	if (fm_subtype == FMSUB_FM8) {
		return FALSE;
	}
#endif

	switch (addr) {
#if defined(XM7DASH) && defined(OPNAMODE)
		case 0xfd44:
			opna_exreg[OPN_WHG] = (dat & 1);
			return TRUE;
#endif

		/* WHG R}hWX^ */
		case 0xfd45:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_write_command_reg(OPN_WHG, dat, opna_exreg[OPN_WHG]);
#else
			opn_write_command_reg(OPN_WHG, dat);
#endif
			return TRUE;

		/* WHG f[^WX^ */
		case 0xfd46:
#if defined(XM7DASH) && defined(OPNAMODE)
			opn_write_data_reg(OPN_WHG, dat, opna_exreg[OPN_WHG]);
#else
			opn_write_data_reg(OPN_WHG, dat);
#endif
			return TRUE;
	}

	return FALSE;
}

/*
 *	WHG
 *	Z[u
 */
BOOL FASTCALL whg_save(int fileh)
{
	if (!file_bool_write(fileh, whg_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, whg_use)) {
		return FALSE;
	}
	if (!opn_save_common(OPN_WHG, fileh)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	WHG
 *	[h
 */
#if defined(XM7DASH) && defined(OPNAMODE)
BOOL FASTCALL whg_load(int fileh, int ver, BOOL bflag)
#else
BOOL FASTCALL whg_load(int fileh, int ver)
#endif
{
	int i;

	/* t@Co[W3Œǉ */
	if (ver < 300) {
		whg_use = FALSE;
		return TRUE;
	}

	if (!file_bool_read(fileh, &whg_enable)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &whg_use)) {
		return FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (!opn_load_common(OPN_WHG, fileh, bflag)) {
#else
	if (!opn_load_common(OPN_WHG, fileh)) {
#endif
		return FALSE;
	}

	/* WHGWX^ */
	whg_notify(0x27, (BYTE)(opn_reg[OPN_WHG][0x27]));

#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<6; i++) {
		whg_notify(0x28, (BYTE)(i + ((i < 3) ? (0) : (1))));
#else
	for (i=0; i<3; i++) {
		whg_notify(0x28, (BYTE)i);
#endif
		opn_key[OPN_WHG][i] = FALSE;
	}

#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<13; i++) {
		if ((i < 8) || (i > 10)) {
			whg_notify((WORD)i, opn_reg[OPN_WHG][i]);
		}
		else {
			whg_notify((WORD)i, 0);
		}
	}

	for (i=0x10; i<0x1d; i++) {
		if (i == 0x10) {
			whg_notify((WORD)i, opn_reg[OPN_WHG][i] | 0x80);
		}
		else if ((i >= 0x12) && (i <= 0x17)) {
			continue;
		}
		else {
			whg_notify((WORD)i, opn_reg[OPN_WHG][i]);
		}
	}

	whg_notify(0x22, (BYTE)(opn_reg[OPN_WHG][0x22]));
	whg_notify(0x29, (BYTE)(opn_reg[OPN_WHG][0x29]));

	for (i=0x30; i<0xb8; i++) {
		if (!(opn_reg[OPN_WHG][0x29] & 0x80) && (i >= 0xb4)) {
			whg_notify((WORD)i, opn_reg[OPN_WHG][i] | 0xc0);
		}
		else {
			whg_notify((WORD)i, opn_reg[OPN_WHG][i]);
		}
	}
	for (i=0x130; i<0x1b8; i++) {
		whg_notify((WORD)i, opn_reg[OPN_WHG][i]);
	}
#else
	for (i=0; i<13; i++) {
		if ((i < 8) || (i > 10)) {
			whg_notify((BYTE)i, opn_reg[OPN_WHG][i]);
		}
		else {
			whg_notify((BYTE)i, 0);
		}
	}

	for (i=0x30; i<0xb4; i++) {
		whg_notify((BYTE)i, opn_reg[OPN_WHG][i]);
	}
#endif

	/* Cxg */
	schedule_handle(EVENT_WHG_A, whg_timera_event);
	schedule_handle(EVENT_WHG_B, whg_timerb_event);

	return TRUE;
}

/*-[ THG OPN ]---------------------------------------------------------------*/

/*
 *	THG
 *	^C}[AI[ot[
 */
static BOOL FASTCALL thg_timera_event(void)
{
	return opn_timera_event_main(OPN_THG);
}

/*
 *	THG
 *	^C}[BI[ot[
 */
static BOOL FASTCALL thg_timerb_event(void)
{
	return opn_timerb_event_main(OPN_THG);
}

/*
 *	THG
 *	PoCgǂݏo
 */
BOOL FASTCALL thg_readb(WORD addr, BYTE *dat)
{
#if defined(XM7DASH) && defined(OPNAMODE)
	BYTE exreg = opna_exreg[OPN_THG];
#endif

	switch (addr) {
		/* R}hWX^͓ǂݏo֎~ */
		case 0xfd0d:
#if XM7_VER >= 2
			/* FM-7[ĥݗL */
			if (fm7_ver != 1) {
				return FALSE;
			}
#endif
		case 0xfd51:
			/* LtO`FbNAȂ牽Ȃ */
			if ((!thg_enable) && (addr == 0xfd51)) {
				return FALSE;
			}
#if XM7_VER == 1
			if ((fm_subtype == FMSUB_FM8) && (addr == 0xfd51)) {
				return FALSE;
			}
#endif

			*dat = 0xff;
			return TRUE;

		/* f[^WX^ */
		case 0xfd0e:
#if XM7_VER >= 2
			/* FM-7[ĥݗL */
			if (fm7_ver != 1) {
				return FALSE;
			}
#endif
#if defined(XM7DASH) && defined(OPNAMODE)
			exreg = 0;
#endif
		case 0xfd52:
			/* LtO`FbNAȂ牽Ȃ */
			if ((!thg_enable) && (addr == 0xfd52)) {
				return FALSE;
			}
#if XM7_VER == 1
			if ((fm_subtype == FMSUB_FM8) && (addr == 0xfd52)) {
				return FALSE;
			}
#endif

#if defined(XM7DASH) && defined(OPNAMODE)
			opn_read_data_reg(OPN_THG, dat, exreg);
#else
			opn_read_data_reg(OPN_THG, dat);
#endif
			return TRUE;

		/* g荞݃Xe[^X */
		case 0xfd53:
#ifdef XM7DASH
			/* LtO`FbNAȂ牽Ȃ */
			if (!thg_enable) {
				return FALSE;
			}
#endif
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return FALSE;
			}
#endif
			opn_read_interrupt_reg(OPN_THG, dat);
			return TRUE;
	}

	return FALSE;
}

/*
 *	THG
 *	PoCg
 */
BOOL FASTCALL thg_writeb(WORD addr, BYTE dat)
{
#if defined(XM7DASH) && defined(OPNAMODE)
	BYTE exreg = opna_exreg[OPN_THG];
#endif

	switch (addr) {
#if defined(XM7DASH) && defined(OPNAMODE)
		case 0xfd50:
#if XM7_VER == 1
			if (fm_subtype == FMSUB_FM8) {
				return FALSE;
			}
#endif
			opna_exreg[OPN_THG] = (dat & 1);
			return TRUE;
#endif

		/* THGR}hWX^ */
		case 0xfd0d:
#if XM7_VER >= 2
			/* FM-7[ĥݗL */
			if (fm7_ver != 1) {
				return FALSE;
			}
#endif
#if defined(XM7DASH) && defined(OPNAMODE)
			exreg = 0;
#endif
			/* PSGANZX͉2bit̂ݗL */
			dat &= 0x03;
		case 0xfd51:
			/* LtO`FbNAȂ牽Ȃ */
			if ((!thg_enable) && (addr == 0xfd51)) {
				return FALSE;
			}
#if XM7_VER == 1
			if ((fm_subtype == FMSUB_FM8) && (addr == 0xfd51)) {
				return FALSE;
			}
#endif

			/* THGtOONEPSGWX^}XN */
			if ((dat & 0x0f) == OPN_WRITEDAT) {
				if (addr == 0xfd51) {
					/* THG I/OANZX̓tOI */
					thg_use = TRUE;
				}
				else if (opn_selreg[OPN_THG] >= 0x10) {
					/* PSG I/OANZXAWX^0x10ȏȂ牽Ȃ */
					return TRUE;
				}
			}

#if defined(XM7DASH) && defined(OPNAMODE)
			opn_write_command_reg(OPN_THG, dat, exreg);
#else
			opn_write_command_reg(OPN_THG, dat);
#endif
			return TRUE;

		/* f[^WX^ */
		case 0xfd0e:
#if XM7_VER >= 2
			/* FM-7[ĥݗL */
			if (fm7_ver != 1) {
				return FALSE;
			}
#endif
#if defined(XM7DASH) && defined(OPNAMODE)
			exreg = 0;
#endif
		case 0xfd52:
			/* LtO`FbNAȂ牽Ȃ */
			if ((!thg_enable) && (addr == 0xfd52)) {
				return FALSE;
			}
#if XM7_VER == 1
			if ((fm_subtype == FMSUB_FM8) && (addr == 0xfd52)) {
				return FALSE;
			}
#endif

			/* THGtOONEPSGWX^}XN */
			if (opn_pstate[OPN_THG] == OPN_WRITEDAT) {
				if (addr == 0xfd51) {
					/* THG I/OANZX̓tOI */
					thg_use = TRUE;
				}
				else if (opn_selreg[OPN_THG] >= 0x10) {
					/* PSG I/OANZXAWX^0x10ȏȂ牽Ȃ */
					return TRUE;
				}
			}

#if defined(XM7DASH) && defined(OPNAMODE)
			opn_write_data_reg(OPN_THG, dat, exreg);
#else
			opn_write_data_reg(OPN_THG, dat);
#endif
			return TRUE;
	}

	return FALSE;
}

/*
 *	THG
 *	Z[u
 */
BOOL FASTCALL thg_save(int fileh)
{
	if (!file_bool_write(fileh, thg_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, thg_use)) {
		return FALSE;
	}
	if (!opn_save_common(OPN_THG, fileh)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	THG
 *	[h
 */
#if defined(XM7DASH) && defined(OPNAMODE)
BOOL FASTCALL thg_load(int fileh, int ver, BOOL bflag)
#else
BOOL FASTCALL thg_load(int fileh, int ver)
#endif
{
	int i;

	/* t@Co[W6Œǉ */
	if (ver < 200) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (ver < 600) {
		thg_use = FALSE;
		return TRUE;
	}
#endif

	if (!file_bool_read(fileh, &thg_enable)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &thg_use)) {
		return FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (!opn_load_common(OPN_THG, fileh, bflag)) {
#else
	if (!opn_load_common(OPN_THG, fileh)) {
#endif
		return FALSE;
	}

	/* THGWX^ */
	thg_notify(0x27, (BYTE)(opn_reg[OPN_THG][0x27]));

#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<6; i++) {
		thg_notify(0x28, (BYTE)(i + ((i < 3) ? (0) : (1))));
#else
	for (i=0; i<3; i++) {
		thg_notify(0x28, (BYTE)i);
#endif
		opn_key[OPN_THG][i] = FALSE;
	}

#if defined(XM7DASH) && defined(OPNAMODE)
	for (i=0; i<13; i++) {
		if ((i < 8) || (i > 10)) {
			thg_notify((WORD)i, opn_reg[OPN_THG][i]);
		}
		else {
			thg_notify((WORD)i, 0);
		}
	}

	for (i=0x10; i<0x1d; i++) {
		if (i == 0x10) {
			thg_notify((WORD)i, opn_reg[OPN_THG][i] | 0x80);
		}
		else if ((i >= 0x12) && (i <= 0x17)) {
			continue;
		}
		else {
			thg_notify((WORD)i, opn_reg[OPN_THG][i]);
		}
	}

	thg_notify(0x22, (BYTE)(opn_reg[OPN_THG][0x22]));
	thg_notify(0x29, (BYTE)(opn_reg[OPN_THG][0x29]));

	for (i=0x30; i<0xb8; i++) {
		if (!(opn_reg[OPN_STD][0x29] & 0x80) && (i >= 0xb4)) {
			thg_notify((WORD)i, opn_reg[OPN_THG][i] | 0xc0);
		}
		else {
			thg_notify((WORD)i, opn_reg[OPN_THG][i]);
		}
	}
	for (i=0x130; i<0x1b8; i++) {
		thg_notify((WORD)i, opn_reg[OPN_THG][i]);
	}
#else
	for (i=0; i<13; i++) {
		if ((i < 8) || (i > 10)) {
			thg_notify((BYTE)i, opn_reg[OPN_THG][i]);
		}
		else {
			thg_notify((BYTE)i, 0);
		}
	}

	for (i=0x30; i<0xb4; i++) {
		thg_notify((BYTE)i, opn_reg[OPN_THG][i]);
	}
#endif

	/* Cxg */
	schedule_handle(EVENT_THG_A, thg_timera_event);
	schedule_handle(EVENT_THG_B, thg_timerb_event);

	return TRUE;
}