/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ VXeǗ ]
 *
 *	RHG
 *	  2001.08.07		zbgZbg@\ǉ
 *	  2001.11.19		Xe[gt@C(Cxg16^Cv)̃[hɑΉ
 *						V2.5L21ō쐬ꂽXe[gt@C̓ǂݍ݂ɑΉ
 *	  2002.03.06		V3ł̃Xe[gt@CVer.7̃[hɑΉ
 *						Xe[gt@Co[WSʓI3
 *	  2002.12.04		zbgZbg̕@ύX
 *	  2004.05.04		Xe[gt@CVer.5[hłȂC
 *	  2004.12.01		F-BASIC V3.5ł̃zbgZbgɑΉ
 *	  2006.11.08		{TuVXẽXe[gZ[uE[hɑΉ
 */

#include <string.h>
#include <stdlib.h>
#include "xm7.h"
#include "display.h"
#include "ttlpalet.h"
#include "subctrl.h"
#include "keyboard.h"
#include "fdc.h"
#include "mainetc.h"
#include "multipag.h"
#include "kanji.h"
#include "tapelp.h"
#include "display.h"
#include "opn.h"
#include "mmr.h"
#include "aluline.h"
#include "apalet.h"
#include "rtc.h"
#include "mouse.h"
#include "rs232c.h"
#include "jsubsys.h"
#include "bubble.h"
#ifdef MIDI
#include "midi.h"
#endif
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
#include "jcard.h"
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
#include "dmac.h"
#endif
#include "device.h"
#include "event.h"

/*
 *	O[o [N
 */
int fm7_ver;							/* n[hEFAo[W */
int boot_mode;							/* N[h BASIC/DOS */
#if XM7_VER == 1
BYTE fm_subtype;						/* n[hEFATuo[W */
BOOL lowspeed_mode;						/* NbN[h */
BOOL available_fm8roms;					/* FM-8 ROMgp\tO */
BOOL available_fm7roms;					/* FM-7 ROMgp\tO */
#ifdef XM7DASH
BOOL available_bblboot8;				/* FM-8 BOOTROM(BUBBLE)gp\tO */
BOOL available_sfdboot8;				/* FM-8 BOOTROM(8"DOS)gp\tO */
BOOL available_dbgboot8;				/* FM-8 BOOTROM(DEBUG)gp\tO */
BOOL available_1mbboot;					/* FM-77 BOOTROM(1MB DOS)gp\tO */
BOOL available_bblboot;					/* FM-77 BOOTROM(BUBBLE)gp\tO */
#endif
#endif
#if (XM7_VER == 1) || (XM7_VER >= 3)
BOOL available_mmrboot;					/* Bu[gROMgp\tO */
#endif
#if XM7_VER >= 3 && defined(XM7DASH)
BOOL bDownGradeMode;					/* FM77AV20EX[htO */
#endif

BOOL hotreset_flag;						/* CzbgZbgtO */
BOOL reset_flag;						/* VXeZbgtO */
BYTE fetch_op;							/* OɃtFb` */
#ifdef XM7DASH
BOOL init_flag;							/* VXetO */
#endif

/*
 *	Xe[gt@Cwb_
 */
#if XM7_VER >= 3
#ifdef XM7DASH
#if defined(OPNAMODE) && (defined(JCARD) || defined(JSUB))
const char *state_header = "XM7'VM'STATE'918";
#elif defined(OPNAMODE)
const char *state_header = "XM7'VM STATE'918";
#elif defined(JCARD) || defined(JSUB)
const char *state_header = "XM7'VM'STATE 918";
#else
const char *state_header = "XM7'VM STATE 918";
#endif
#else
const char *state_header = "XM7 VM STATE 918";
#endif
#elif XM7_VER >= 2
#ifdef XM7DASH
#if defined(OPNAMODE) && (defined(JCARD) || defined(JSUB))
const char *state_header = "XM7'VM'STATE'718";
#elif defined(OPNAMODE)
const char *state_header = "XM7'VM STATE'718";
#elif defined(JCARD) || defined(JSUB)
const char *state_header = "XM7'VM'STATE 718";
#else
const char *state_header = "XM7'VM STATE 718";
#endif
#else
const char *state_header = "XM7 VM STATE 718";
#endif
#else
#ifdef XM7DASH
#if defined(OPNAMODE)
const char *state_header = "XM7'VM STATE'410";
#else
const char *state_header = "XM7'VM STATE 410";
#endif
#else
const char *state_header = "XM7 VM STATE 308";
#endif
#endif

/*
 *	VXe
 *	
 */
BOOL FASTCALL system_init(void)
{
	/* [hݒ */
#if XM7_VER >= 3
	fm7_ver = 3;					/* FM77AV40EXɐݒ */
	available_mmrboot = TRUE;
#elif XM7_VER >= 2
	fm7_ver = 2;					/* FM77AVɐݒ */
#else
	fm7_ver = 1;					/* FM-77+192KB RAMɐݒ */
	fm_subtype = FMSUB_FM77;
	lowspeed_mode = FALSE;
	available_fm8roms = TRUE;
	available_fm7roms = TRUE;
	available_mmrboot = TRUE;
#ifdef XM7DASH
	available_bblboot8 = TRUE;
	available_sfdboot8 = TRUE;
	available_dbgboot8 = TRUE;
	available_1mbboot = TRUE;
	available_bblboot = TRUE;
#endif
#endif
	boot_mode = BOOT_BASIC;			/* BASIC MODE */

	hotreset_flag = FALSE;
	reset_flag = FALSE;
#ifdef XM7DASH
	init_flag = TRUE;
#endif

	/* XPW[AoX */
	if (!schedule_init()) {
		return FALSE;
	}
	if (!mmr_init()) {
		return FALSE;
	}

	/* ACPU */
	if (!mainmem_init()) {
		return FALSE;
	}
	if (!submem_init()) {
		return FALSE;
	}
	if (!maincpu_init()) {
		return FALSE;
	}
	if (!subcpu_init()) {
		return FALSE;
	}
#ifdef JSUB
	if (!jsubsys_init()) {
		return FALSE;
	}
#endif

	/* ̑foCX */
	if (!display_init()) {
		return FALSE;
	}
	if (!ttlpalet_init()) {
		return FALSE;
	}
	if (!subctrl_init()) {
		return FALSE;
	}
	if (!keyboard_init()) {
		return FALSE;
	}
	if (!fdc_init()) {
		return FALSE;
	}
	if (!mainetc_init()) {
		return FALSE;
	}
	if (!multipag_init()) {
		return FALSE;
	}
	if (!kanji_init()) {
		return FALSE;
	}
	if (!tapelp_init()) {
		return FALSE;
	}
	if (!opn_init()) {
		return FALSE;
	}
	if (!whg_init()) {
		return FALSE;
	}
	if (!thg_init()) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!aluline_init()) {
		return FALSE;
	}
	if (!apalet_init()) {
		return FALSE;
	}
	if (!rtc_init()) {
		return FALSE;
	}
#endif
#ifdef MIDI
	if (!midi_init()) {
		return FALSE;
	}
#endif
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
	if (!jcard_init()) {
		return FALSE;
	}
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	if (!dmac_init()) {
		return FALSE;
	}
#endif
#ifdef MOUSE
	if (!mos_init()) {
		return FALSE;
	}
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	if (!bmc_init()) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	VXe
 *	N[Abv
 */
void FASTCALL system_cleanup(void)
{
	/* ̑foCX */
#if (XM7_VER == 1) && defined(BUBBLE)
	bmc_cleanup();
#endif
#ifdef MOUSE
	mos_cleanup();
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	dmac_cleanup();
#endif
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
	jcard_cleanup();
#endif
#ifdef MIDI
	midi_cleanup();
#endif
#if XM7_VER >= 2
	rtc_cleanup();
	apalet_cleanup();
	aluline_cleanup();
#endif
	thg_cleanup();
	whg_cleanup();
	opn_cleanup();
	tapelp_cleanup();
	kanji_cleanup();
	multipag_cleanup();
	mainetc_cleanup();
	fdc_cleanup();
	keyboard_cleanup();
	subctrl_cleanup();
	ttlpalet_cleanup();
	display_cleanup();

	/* ACPU */
#ifdef JSUB
	jsubsys_cleanup();
#endif
	subcpu_cleanup();
	maincpu_cleanup();
	submem_cleanup();
	mainmem_cleanup();

	/* XPW[AoX */
	mmr_cleanup();
	schedule_cleanup();
}

/*
 *	VXe
 *	Zbg
 */
void FASTCALL system_reset(void)
{
	/* XPW[AoX */
	schedule_reset();
	mmr_reset();

	/* ̑foCX */
	display_reset();
	ttlpalet_reset();
	subctrl_reset();
	keyboard_reset();
	fdc_reset();
	mainetc_reset();
	multipag_reset();
	kanji_reset();
	tapelp_reset();
	opn_reset();
	whg_reset();
	thg_reset();
#if XM7_VER >= 2
	aluline_reset();
	apalet_reset();
	rtc_reset();
#endif
#ifdef MIDI
	midi_reset();
#endif
#ifdef RSC
	rs232c_reset();
#endif
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
	jcard_reset();
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	dmac_reset();
#endif
#ifdef MOUSE
	mos_reset();
#endif

	/* ACPU */
	mainmem_reset();
	submem_reset();
	maincpu_reset();
	subcpu_reset();
#ifdef JSUB
	jsubsys_reset();
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	bmc_reset();
#endif

	/* ʍĕ` */
	display_setpointer(FALSE);
	display_notify();

	/* VXeZbgtOݒ */
	reset_flag = TRUE;

	/* tFb`ߏ */
	fetch_op = 0;

#ifdef XM7DASH
	/* VXetONA */
	init_flag = FALSE;
#endif
}

/*
 *	VXe
 *	zbgZbg
 */
void FASTCALL system_hotreset(void)
{
#if XM7_VER == 1 && !defined(XM7DASH)
	BOOL flag;
#endif

#ifndef XM7DASH
	/* TWRLȂN[hDOSɐ؂ւ */
#if XM7_VER >= 2
	if (twr_flag) {
		boot_mode = BOOT_DOS;
	}
#else
	flag = twr_flag;
#endif
#endif

	/* VXeZbg */
	system_reset();

	/* BREAKL[Ԃɂ */
	break_flag = TRUE;
	maincpu_firq();

	/* zbgZbgtOL */
	hotreset_flag = TRUE;

#if XM7_VER == 1 && !defined(XM7DASH)
	/* ZbgOTWRL痠RAMI */
	if (flag) {
		basicrom_en = FALSE;
	}
#endif
}

/*
 *	VXe
 *	t@CZ[u
 */
BOOL FASTCALL system_save(char *filename)
{
	int fileh;
	BOOL flag;
	ASSERT(filename);

	/* t@CI[v */
	fileh = file_open(filename, OPEN_W);
	if (fileh == -1) {
		return FALSE;
	}

	/* tO */
	flag = TRUE;

	/* wb_Z[u */
	if (!file_write(fileh, (BYTE *)state_header, 16)) {
		flag = FALSE;
	}

	/* VXe[N */
	if (!file_word_write(fileh, (WORD)fm7_ver)) {
		return FALSE;
	}
	if (!file_word_write(fileh, (WORD)boot_mode)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_byte_write(fileh, fm_subtype)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lowspeed_mode)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_fm8roms)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_fm7roms)) {
		return FALSE;
	}
#ifdef XM7DASH
	if (!file_bool_write(fileh, available_bblboot8)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_sfdboot8)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_dbgboot8)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_1mbboot)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_bblboot)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_mmrboot)) {
		return FALSE;
	}
#endif
#endif

	/* ԂɌĂяo */
	if (!mainmem_save(fileh)) {
		flag = FALSE;
	}
	if (!submem_save(fileh)) {
		flag = FALSE;
	}
	if (!maincpu_save(fileh)) {
		flag = FALSE;
	}
	if (!subcpu_save(fileh)) {
		flag = FALSE;
	}
	if (!schedule_save(fileh)) {
		flag = FALSE;
	}
	if (!display_save(fileh)) {
		flag = FALSE;
	}
	if (!ttlpalet_save(fileh)) {
		flag = FALSE;
	}
	if (!subctrl_save(fileh)) {
		flag = FALSE;
	}
	if (!keyboard_save(fileh)) {
		flag = FALSE;
	}
	if (!fdc_save(fileh)) {
		flag = FALSE;
	}
	if (!mainetc_save(fileh)) {
		flag = FALSE;
	}
	if (!multipag_save(fileh)) {
		flag = FALSE;
	}
	if (!kanji_save(fileh)) {
		flag = FALSE;
	}
	if (!tapelp_save(fileh)) {
		flag = FALSE;
	}
	if (!opn_save(fileh)) {
		flag = FALSE;
	}
	if (!mmr_save(fileh)) {
		flag = FALSE;
	}
#if XM7_VER >= 2
	if (!aluline_save(fileh)) {
		flag = FALSE;
	}
	if (!rtc_save(fileh)) {
		flag = FALSE;
	}
	if (!apalet_save(fileh)) {
		flag = FALSE;
	}
#endif
	if (!whg_save(fileh)) {
		flag = FALSE;
	}
	if (!thg_save(fileh)) {
		flag = FALSE;
	}
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
	if (!jcard_save(fileh)) {
		flag = FALSE;
	}
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	if (!dmac_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef MOUSE
	if (!mos_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef RSC
	if (!rs232c_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef MIDI
	if (!midi_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef JSUB
	if (!jsubsys_save(fileh)) {
		flag = FALSE;
	}
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	if (!bmc_save(fileh)) {
		flag = FALSE;
	}
#endif
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
	if (!ifdc_save(fileh)) {
		flag = FALSE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */

	file_close(fileh);
	return flag;
}

/*
 *	VXe
 *	t@C[h
 */
int FASTCALL system_load(char *filename)
{
	int fileh;
	int ver;
	char header[16];
	BOOL flag;
#if XM7_VER >= 2
	BOOL old_scheduler;
	int filesize;
#endif
	WORD tmp;
#if XM7_VER == 1
	BOOL temp;
#endif
#ifdef XM7DASH
	BOOL xm7dash;
#ifdef OPNAMODE
	BOOL baflag;
#endif
#if XM7_VER >= 2 && (defined(JCARD) || defined(JSUB))
	BOOL bjflag;
#endif
#endif

	ASSERT(filename);

	/* t@CI[v */
	fileh = file_open(filename, OPEN_R);
	if (fileh == -1) {
		return STATELOAD_OPENERR;
	}

	/* tO */
	flag = TRUE;
#if XM7_VER >= 2
	old_scheduler = FALSE;
#endif
#ifdef XM7DASH
	xm7dash = FALSE;
#if defined(OPNAMODE)
	baflag = FALSE;
#endif
#if XM7_VER >= 2 && (defined(JCARD) || defined(JSUB))
	bjflag = FALSE;
#endif
#endif

	/* wb_[h */
	if (!file_read(fileh, (BYTE *)header, 16)) {
		flag = FALSE;
	}
	else {
		if (memcmp(header, "XM7 VM STATE ", 13) != 0) {
			flag = FALSE;
		}
#ifdef XM7DASH
		if (memcmp(header, "XM7'VM STATE ", 13) == 0) {
			flag = TRUE;
			xm7dash = TRUE;
		}
#if defined(OPNAMODE)
		if (memcmp(header, "XM7'VM STATE'", 13) == 0) {
			flag = TRUE;
			xm7dash = TRUE;
			baflag = TRUE;
		}
#endif
#if XM7_VER >= 2 && (defined(JCARD) || defined(JSUB))
		if (memcmp(header, "XM7'VM'STATE ", 13) == 0) {
			flag = TRUE;
			xm7dash = TRUE;
			bjflag = TRUE;
		}
#if defined(OPNAMODE)
		if (memcmp(header, "XM7'VM'STATE'", 13) == 0) {
			flag = TRUE;
			xm7dash = TRUE;
			baflag = TRUE;
			bjflag = TRUE;
		}
#endif
#endif
#endif
	}

	/* wb_`FbN */
	if (!flag) {
		file_close(fileh);
		return STATELOAD_HEADERR;
	}

	/* t@Co[W擾Ao[W2ȏオΏ */
	if (header[13] != 0x20) {
		ver = (int)	((BYTE)(header[13] - 0x30) * 100) +
					((BYTE)(header[14] - 0x30) * 10) +
					((BYTE)(header[15] - 0x30));
	}
	else {
		ver = (int)(BYTE)(header[15]);
		ver -= 0x30;
		ver *= 100;
	}
#if XM7_VER >= 3
	/* V3 : Ver5̓[hłȂ */
	if ((ver < 200) || ((ver >= 300) && (ver <= 499))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#elif XM7_VER >= 2
	/* V2 : Ver8ȏEVer5̓[hłȂ */
	if ((ver >= 800) || (ver < 200) || ((ver >= 300) && (ver <= 499))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#else
	/* V1 : Ver5ȏEVer2̓[hłȂ */
	if ((ver >= 500) || (ver < 200)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#endif
#ifdef XM7DASH
	/* o[WXe[gt@Č`ύXƂ̑΍ */
#if XM7_VER >= 3
	/* V3 : o[WVer9.19ȏ̓[hłȂ */
	if ((!xm7dash) && ((ver >= 919) || ((ver >= 719) && (ver <= 799)))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#elif XM7_VER >= 2
	/* V2 : o[WVer7.19ȏ̓[hłȂ */
	if ((!xm7dash) && (ver >= 719)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#else
	/* V1 : o[WVer3.09ȏ̓[hłȂ */
	if ((!xm7dash) && (ver >= 309)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#endif
#endif
	if (ver >	(int)((BYTE)(state_header[13] - 0x30) * 100) +
					 ((BYTE)(state_header[14] - 0x30) * 10) +
					 ((BYTE)(state_header[15] - 0x30))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}

#if XM7_VER >= 2
	/* V3.0L30/V2.5L20ȑÕXe[gt@CɑΏ */
	filesize = file_getsize(fileh);
	if (ver <= 500) {
		old_scheduler = TRUE;
	}
	else {
#if XM7_VER >= 3
		if ((filesize == 454758) ||		/* XM7 V3.0 (RAM) */
			(filesize == 1241190) ||	/* XM7 V3.0 (RAML) */
			(filesize == 258030)) {		/* XM7 V2 */
			old_scheduler = TRUE;
		}
#else
		if (filesize == 258030) {		/* XM7 V2 */
			old_scheduler = TRUE;
		}
#endif
	}
#endif

	/* VXe[N */
	if (!file_word_read(fileh, &tmp)) {
		return FALSE;
	}
	fm7_ver = (int)tmp;
	if (!file_word_read(fileh, &tmp)) {
		return FALSE;
	}
	boot_mode = (int)tmp;
#if XM7_VER == 1 && defined(BUBBLE) && defined(XM7DASH)
	if ((ver >= 307) && (ver <= 399)) {
		if (boot_mode == 2) {
			boot_mode = BOOT_BUBBLE;
		}
	}
	mainmem_bubblemode_boot();
#endif	/* XM7_VER == 1 && defined(BUBBLE) && defined(XM7DASH) */
#if XM7_VER == 1
	if (!file_byte_read(fileh, &fm_subtype)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lowspeed_mode)) {
		return FALSE;
	}

	/* Z[uɗLROMf[^[hɗLłȂꍇG[ */
	if (!file_bool_read(fileh, &temp)) {
		return FALSE;
	}
	if (!available_fm8roms && temp) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &temp)) {
		return FALSE;
	}
	if (!available_fm7roms && temp) {
		return FALSE;
	}
#ifdef XM7DASH
	/* XM7 V1.2 - Ver4g BOOT ROM֘A */
	if (ver >= 400) {
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (!available_bblboot8 && temp) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (!available_sfdboot8 && temp) {
			return FALSE;
		}
		if (ver >= 408) {
			if (!file_bool_read(fileh, &temp)) {
				return FALSE;
			}
			if (!available_dbgboot8 && temp) {
				return FALSE;
			}
		}
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (!available_1mbboot && temp) {
			return FALSE;
		}
		if (ver >= 404) {
			if (!file_bool_read(fileh, &temp)) {
				return FALSE;
			}
			if (!available_bblboot && temp) {
				return FALSE;
			}
		}
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (!available_mmrboot && temp) {
			return FALSE;
		}
	}
#endif
#endif

	/* ԂɌĂяo */
	if (!mainmem_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!submem_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!maincpu_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!subcpu_load(fileh, ver)) {
		flag = FALSE;
	}
#if XM7_VER == 1
	if (!schedule_load(fileh, ver)) {
		flag = FALSE;
	}
#elif defined(JSUB) && defined(XM7DASH)
	if (!schedule_load(fileh, ver, old_scheduler, bjflag)) {
		flag = FALSE;
	}
#else
	if (!schedule_load(fileh, ver, old_scheduler)) {
		flag = FALSE;
	}
#endif
	if (!display_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!ttlpalet_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!subctrl_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!keyboard_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!fdc_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!mainetc_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!multipag_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!kanji_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!tapelp_load(fileh, ver)) {
		flag = FALSE;
	}
#if defined(XM7DASH) && defined(OPNAMODE)
	if (!opn_load(fileh, ver, baflag)) {
		flag = FALSE;
	}
#else
	if (!opn_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
	if (!mmr_load(fileh, ver)) {
		flag = FALSE;
	}
#if XM7_VER >= 2
	if (!aluline_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!rtc_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!apalet_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#if defined(XM7DASH) && defined(OPNAMODE)
	if (!whg_load(fileh, ver, baflag)) {
		flag = FALSE;
	}
	if (!thg_load(fileh, ver, baflag)) {
		flag = FALSE;
	}
#else
	if (!whg_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!thg_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#if XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH)
	if (!jcard_load(fileh, ver, bjflag)) {
		flag = FALSE;
	}
#elif XM7_VER >= 3
	if (!jcard_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
	if (!dmac_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef MOUSE
	if (!mos_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef RSC
	if (!rs232c_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef MIDI
	if (!midi_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef JSUB
#if XM7_VER >= 2 && defined(JSUB) && defined(XM7DASH)
	if (!jsubsys_load(fileh, ver, bjflag)) {
		flag = FALSE;
	}
#else
	if (!jsubsys_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	if (!bmc_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
	if (!ifdc_load(fileh, ver)) {
		flag = FALSE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */
	file_close(fileh);

#if XM7_VER >= 3
	/* 400C[ĥVRAMzu␳Kv */
	if (mode400l) {
		if (!fix_vram_address()) {
			flag = FALSE;
		}
	}
#endif

	/* ʍĕ` */
	display_notify();

	/* CPUx䗦ݒ */
	speed_ratio = 10000;

	if (!flag) {
		return STATELOAD_ERROR;
	}

	return STATELOAD_SUCCESS;
}

/*
 *	t@Cǂݍ(BYTE)
 */
BOOL FASTCALL file_byte_read(int fileh, BYTE *dat)
{
	return file_read(fileh, dat, 1);
}

/*
 *	t@Cǂݍ(WORD)
 */
BOOL FASTCALL file_word_read(int fileh, WORD *dat)
{
	BYTE tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat = tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat <<= 8;
	*dat |= tmp;

	return TRUE;
}

/*
 *	t@Cǂݍ(DWORD)
 */
BOOL FASTCALL file_dword_read(int fileh, DWORD *dat)
{
	BYTE tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat = tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat *= 256;
	*dat |= tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat *= 256;
	*dat |= tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat *= 256;
	*dat |= tmp;

	return TRUE;
}

/*
 *	t@Cǂݍ(BOOL)
 */
BOOL FASTCALL file_bool_read(int fileh, BOOL *dat)
{
	BYTE tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}

	switch (tmp) {
		case 0:
			*dat = FALSE;
			return TRUE;
		case 0xff:
			*dat = TRUE;
			return TRUE;
	}

	return FALSE;
}

/*
 *	t@C(BYTE)
 */
BOOL FASTCALL file_byte_write(int fileh, BYTE dat)
{
	return file_write(fileh, &dat, 1);
}

/*
 *	t@C(WORD)
 */
BOOL FASTCALL file_word_write(int fileh, WORD dat)
{
	BYTE tmp;

	tmp = (BYTE)(dat >> 8);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat & 0xff);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	t@C(DWORD)
 */
BOOL FASTCALL file_dword_write(int fileh, DWORD dat)
{
	BYTE tmp;

	tmp = (BYTE)(dat >> 24);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat >> 16);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat >> 8);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat & 0xff);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	t@C(BOOL)
 */
BOOL FASTCALL file_bool_write(int fileh, BOOL dat)
{
	BYTE tmp;

	if (dat) {
		tmp = 0xff;
	}
	else {
		tmp = 0;
	}

	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	return TRUE;
}
