/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ Win32API \ ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include "xm7.h"
#include "device.h"
#include "subctrl.h"
#include "display.h"
#include "multipag.h"
#include "w32.h"
#include "w32_gdi.h"
#include "w32_dd.h"
#include "w32_draw.h"
#include "w32_sub.h"
#ifdef XM7DASH
#include "w32_bar.h"
#endif

/*
 *	O[o [N
 */
BOOL bFullScreen;							/* tXN[tO */
#ifndef XM7DASH
BOOL bFullRequested;						/* tXN[v */
BOOL bDrawSelected;							/* ZNgς݃tO */
#endif
BOOL bFullScan;								/* tXL(Window) */
BOOL bFullScanFS;							/* tXL(FullScreen) */
BOOL bDoubleSize;							/* 2{gtO */
#if defined(XM7DASH) && defined(FILTERDLL)
int nScalingSize;							/* n{\TCY */
BOOL bAspectRatio;							/* AXyNg䒲tO */
double dAspectRatio;						/* AXyNg */
#endif
#if XM7_VER == 1 || defined(XM7DASH)
BOOL bGreenMonitor;							/* O[j^tO */
#endif
#if XM7_VER == 2 || (XM7_VER >= 3 && defined(XM7DASH))
BOOL bTTLMonitor;							/* TTLj^tO */
#endif
BOOL bRasterRendering;						/* X^_OtO */
#ifdef XM7DASH
BYTE bmpDrive[6][0x100];					/* Xe[^XhCuBMP */
#endif

/*
 *	X^eBbN [N
 */
static BOOL bNextFrameRender;				/* t[`tO */
static BOOL bDirtyLine[400];				/* vtO */

/*
 *	
 */
void FASTCALL InitDraw(void)
{
	/* [NGA */
	bFullScreen = FALSE;
#ifndef XM7DASH
	bFullRequested = FALSE;
	bDrawSelected = FALSE;
#endif
	bFullScan = FALSE;
	bFullScanFS = FALSE;
	bDoubleSize = FALSE;
#if defined(XM7DASH) && defined(FILTERDLL)
	nScalingSize = 0;						/* x1.0(640x400) */
	bAspectRatio = FALSE;					/* AXyNg䒲tO */
	dAspectRatio = 1.0;						/* AXyNg */
#endif
#if XM7_VER == 1 || defined(XM7DASH)
	bGreenMonitor = FALSE;
#endif
#if XM7_VER == 2 || (XM7_VER >= 3 && defined(XM7DASH))
	bTTLMonitor = FALSE;
#endif
	bRasterRendering = FALSE;
	bNextFrameRender = FALSE;
	memset(bDirtyLine, 0, sizeof(bDirtyLine));

	/* NGDIȂ̂ŁAGDI */
	InitGDI();
}

/*
 *	N[Abv
 */
void FASTCALL CleanDraw(void)
{
	/* tXN[tOɉāAN[Abv */
	if (bFullScreen) {
		CleanDD();
	}
	else {
		CleanGDI();
	}
}

/*
 *  ZNg
 */
BOOL FASTCALL SelectDraw(HWND hWnd)
{
#ifdef XM7DASH
	int i, j, k;
	BITMAPINFOHEADER *pbmi;
	BYTE *p;
	BYTE dat;
	const BYTE * const str = "0123HD";
#endif

	ASSERT(hWnd);

#ifndef XM7DASH
	/* ZNgς */
	bDrawSelected = TRUE;
#endif

#ifdef XM7DASH
	/* NA */
	memset(bmpDrive, 0, sizeof(bmpDrive));

	/* hCuprbg}bv쐬 */
	for (i=0; i<(int)strlen(str); i++) {

		/* rbg}bvwb_ݒ */
		pbmi = (BITMAPINFOHEADER*)&bmpDrive[i][0];
		pbmi->biSize = sizeof(BITMAPINFOHEADER);
		pbmi->biWidth = 8;
		pbmi->biHeight = -8;
		pbmi->biPlanes = 1;
		pbmi->biBitCount = 24;
		pbmi->biCompression = BI_RGB;
		pbmi->biSizeImage = 8 * 8 * 3;

		/* AhXݒ */
		p = &bmpDrive[i][sizeof(BITMAPINFOHEADER)];

		for (j=0; j<8; j++) {
			/* CGROMf[^擾 */
			dat = subrom_c[str[i] * 8 + j];

			/* 1rbgƂɌ */
			for (k=0; k<8; k++) {
				if (dat & 0x80) {
					p[0] = 208;
					p[1] = 208;
					p[2] = 208;
				}
				dat <<= 1;
				p += 3;
			}
		}
	}
#endif

	/* NGDII */
	return SelectGDI(hWnd);
}

/*
 *	[h؂ւ
 */
void FASTCALL ModeDraw(HWND hWnd, BOOL bFullFlag)
{
	ASSERT(hWnd);

	/* ƈvĂ΁AςKvȂ */
	if (bFullFlag == bFullScreen) {
		return;
	}

	if (bFullFlag) {
		/* tXN[ */
		CleanGDI();
		InitDD();
		bFullScreen = TRUE;
		if (!SelectDD()) {
			/* tXN[s */
			bFullScreen = FALSE;
			CleanDD();
			InitGDI();
			SelectGDI(hWnd);
		}
	}
	else {
		/* EChE */
		bFullScreen = FALSE;
		CleanDD();
		InitGDI();
		if (!SelectGDI(hWnd)) {
			/* EChEs */
			CleanGDI();
			bFullScreen = TRUE;
			InitDD();
			SelectDD();
		}
	}

	/* ĕ`̂ߏ׍H */
	display_notify();
}

#ifdef XM7DASH
/*
 *	[hċN
 */
void FASTCALL ReDraw(HWND hWnd)
{
	ASSERT(hWnd);

	/* tXN[[hȂȂɂȂ */
	if (bFullScreen) {
		return;
	}

	CleanGDI();
	InitGDI();
	SelectGDI(hWnd);

	/* ĕ`̂ߏ׍H */
	display_notify();
}
#endif

/*
 *	`(ʏ)
 */
void FASTCALL OnDraw(HWND hWnd, HDC hDC)
{
	ASSERT(hWnd);
	ASSERT(hDC);

	/* G[Ȃ牽Ȃ */
	if (nErrorCode == 0) {
		if (bFullScreen) {
			if (bRasterRendering) {
				DrawPostRenderDD();
			}
			else {
				DrawDD();
			}
		}
		else {
			if (bRasterRendering) {
				DrawPostRenderGDI(hWnd, hDC);
			}
			else {
				DrawGDI(hWnd, hDC);
			}
		}
	}
}

/*
 *	`(WM_PAINT)
 */
void FASTCALL OnPaint(HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC;

	ASSERT(hWnd);

	hDC = BeginPaint(hWnd, &ps);

	/* G[Ȃ牽Ȃ */
	if (nErrorCode == 0) {
		/* ĕ`w */
		if (bFullScreen) {
			ReDrawDD();
		}
		else {
			ReDrawGDI();
		}

		/* ` */
		OnDraw(hWnd, hDC);
	}

	EndPaint(hWnd, &ps);
}

/*
 *	j[Jn
 */
void FASTCALL EnterMenu(HWND hWnd)
{
	if (bFullScreen) {
		EnterMenuDD(hWnd);
	}
	else {
		EnterMenuGDI(hWnd);
	}
}

/*
 *	j[I
 */
void FASTCALL ExitMenu(void)
{
	if (bFullScreen) {
		ExitMenuDD();
	}
	else {
		ExitMenuGDI();
	}
}

#ifdef XM7DASH
#ifdef FILTERDLL
/*
 *	O摜tB^p[^ݒ
 */
BOOL FASTCALL SetFilterParameter(WORD wParam)
{
	if (bFullScreen) {
		return FALSE;
	}
	else {
		return SetFilterParameterGDI(wParam);
	}
}
#endif

/*
 *	ʃbZ[W[h
 */
void FASTCALL LoadMessageScreen(void)
{
	if (bFullScreen) {
		LoadMessageInitDD();
	}
	else {
		LoadMessageStatusBar();
	}
}
#endif

/*-[ VMƂ̐ڑ ]-----------------------------------------------------------*/

/*
 *	ĕ`惉X^ꊇݒ
 */
void FASTCALL SetDirtyFlag(int top, int bottom, BOOL flag)
{
	int y;

	if (bRasterRendering) {
		for (y = top; y < bottom; y++) {
			bDirtyLine[y] = flag;
		}
	}
}

/*
 *	VRAMݒʒm
 */
void FASTCALL vram_notify(WORD addr, BYTE dat)
{
	WORD y;

	UNUSED(dat);

#if XM7_VER >= 2
#if XM7_VER >= 3
	if (screen_mode == SCR_200LINE) {
#else
	if (!mode320) {
#endif
		if (vram_active != vram_display) {
			return;
		}
	}
#endif

	/* YWZo */
	if (bRasterRendering) {
#if XM7_VER >= 3
		switch (screen_mode) {
			case SCR_400LINE	:	addr &= 0x7fff;
									y = (WORD)(addr / 80);
									break;
			case SCR_262144		:
			case SCR_4096		:	addr &= 0x1fff;
									y = (WORD)(addr / 40);
									break;
			case SCR_200LINE	:	addr &= 0x3fff;
									y = (WORD)(addr / 80);
									break;
		}
#elif XM7_VER >= 2
		if (mode320) {
			addr &= 0x1fff;
			y = (WORD)(addr / 40);
		}
		else {
			addr &= 0x3fff;
			y = (WORD)(addr / 80);
		}
#elif defined(L4CARD)
		if (enable_400line) {
			addr -= vram_offset[0];
			addr &= 0x7fff;
			y = (WORD)(addr / 80);
		}
		else {
			addr &= 0x3fff;
			y = (WORD)(addr / 80);
		}
#else
		addr &= 0x3fff;
		y = (WORD)(addr / 80);
#endif
		if (y < 400) {
			bDirtyLine[y] = TRUE;
		}
	}

	if (bFullScreen) {
		VramDD(addr);
	}
	else {
		VramGDI(addr);
	}
}


/*
 *	TTLpbgʒm
 */
void FASTCALL ttlpalet_notify(void)
{
	if (bRasterRendering) {
		bNextFrameRender = TRUE;
		SetDirtyFlag(now_raster, 400, TRUE);
	}

	if (bFullScreen) {
		DigitalDD();
	}
	else {
		DigitalGDI();
	}
}

/*
 *	AiOpbgʒm
 */
void FASTCALL apalet_notify(void)
{
	if (bRasterRendering) {
		bNextFrameRender = TRUE;
		SetDirtyFlag(now_raster, 400, TRUE);
	}

	if (bFullScreen) {
		AnalogDD();
	}
	else {
		AnalogGDI();
	}
}

/*
 *  ĕ`vʒm
 */
void FASTCALL display_notify(void)
{
	int i;
	int raster;

	if (bRasterRendering) {
		bNextFrameRender = TRUE;
		SetDirtyFlag(0, 400, TRUE);
		if (!run_flag) {
			raster = now_raster;
			for (i = 0; i < 400; i++) {
				now_raster = i;
				hblank_notify();
			}
			now_raster = raster;
		}
	}

	if (bFullScreen) {
		/* ReDraw͖ʂȃNA܂ނ̂ŁA׍H */
		AnalogDD();
	}
	else {
		AnalogGDI();
	}
}

/*
 *	VBLANKIvʒm
 */
void FASTCALL vblankperiod_notify(void)
{
	BOOL flag;
	int y;
	int ymax;

	if (bRasterRendering) {
		/* ̃t[Iɏ邩 */
		if (bNextFrameRender) {
			bNextFrameRender = FALSE;
			SetDirtyFlag(0, 400, TRUE);
		}
		else {
			/* Kv`FbN */
#if XM7_VER >= 3
			if (screen_mode == SCR_400LINE) {
				ymax = 400;
			}
			else {
				ymax = 200;
			}
#elif XM7_VER == 1 && defined(L4CARD)
			if (enable_400line) {
				ymax = 400;
			}
			else {
				ymax = 200;
			}
#else
			ymax = 200;
#endif
			flag = FALSE;
			for (y = 0; y < ymax; y++) {
				flag |= bDirtyLine[y];
			}
			if (!flag) {
				return;
			}
		}

		/* S̈_O */
		if (bFullScreen) {
			/* ReDraw͖ʂȃNA܂ނ̂ŁA׍H */
			AnalogDD();
		}
		else {
			AnalogGDI();
		}
	}
}

/*
 *	HBLANKvʒm
 */
void FASTCALL hblank_notify(void)
{
	if (bRasterRendering) {
		if (bDirtyLine[now_raster]) {
			bDirtyLine[now_raster] = FALSE;
			if (bFullScreen) {
				DrawRasterDD(now_raster);
			}
			else {
				DrawRasterGDI(now_raster);
			}
		}
	}
}

/*
 *	fBW^CYvʒm
 */
void FASTCALL digitize_notify(void)
{
}

#if XM7_VER >= 3
/*
 *	n[hEFAEBhEʒm
 */
void FASTCALL window_notify(void)
{
	if (bFullScreen) {
		WindowDD();
	}
	else {
		WindowGDI();
	}
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXgVRAMݒʒm
 */
void FASTCALL tvram_notify(WORD addr, BYTE dat)
{
	WORD ysize;
	WORD y;

	UNUSED(dat);

	if (bRasterRendering) {
		/* YWZo */
		addr = (WORD)((addr - text_start_addr) & 0x0ffe);
		ysize = (WORD)((crtc_register[9] & 0x1f) + 1);
		if (width40_flag) {
			y = (WORD)((addr / 80) * ysize);
		}
		else {
			y = (WORD)((addr / 160) * ysize);
		}

		/* I[o[`FbN */
		if (y >= 400) {
			return;
		}

		/* ĕ`tOݒ */
		SetDirtyFlag(y, y + ysize, TRUE);
	}

	if (bFullScreen) {
		TvramDD(addr);
	}
	else {
		TvramGDI(addr);
	}
}

/*
 *  eLXgĕ`vʒm
 */
void FASTCALL tvram_redraw_notify(void)
{
	bNextFrameRender = TRUE;
	SetDirtyFlag(now_raster, 400, TRUE);

	if (bFullScreen) {
		ReDrawTVRamDD();
	}
	else {
		ReDrawTVRamGDI();
	}
}
#endif

/*-[ V2 ]-------------------------------------------------------------------*/

#if ((XM7_VER <= 2) && defined(FMTV151))
/*
 *	V2
 *	0=߁A1=΁A2=
 */
const BYTE nV2data[] = {
/*						  1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 */
/*		1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 */
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,1,1,1,1,1,1,1,2,0,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,1,1,1,1,1,1,1,1,1,2,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,1,1,1,1,2,0,0,1,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,1,1,1,2,0,0,0,0,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,0,0,0,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,0,1,1,1,1,2,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,1,1,1,1,2,0,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,1,1,1,1,2,0,0,0,
		1,1,1,1,2,0,0,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,2,0,0,0,0,
		0,1,1,1,1,2,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,2,0,0,0,0,0,
		0,0,1,1,1,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,2,0,0,0,0,0,0,
		0,0,0,1,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,2,
		0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,2,
};

#endif

/*-[ 2HD ]------------------------------------------------------------------*/

#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
/*
 *	2HD
 */
const BYTE bmp2HD[] = {
		0x28, 0x00, 0x00, 0x00, 0x0F, 0x00, 0x00, 0x00,
		0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x20, 0x01, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
};
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */

#endif	/* _WIN32 */
