/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ Win32API CvO ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <imm.h>
#include <shellapi.h>
#include <stdlib.h>
#include <assert.h>
#include <mmsystem.h>
#define DIRECTINPUT_VERSION		0x0300		/* DirectX3w */
#include <dinput.h>
#include "xm7.h"
#ifdef XM7DASH
#include "device.h"
#endif
#include "mouse.h"
#include "tapelp.h"
#include "keyboard.h"
#include "w32.h"
#include "w32_bar.h"
#include "w32_draw.h"
#include "w32_dd.h"
#include "w32_gdi.h"
#include "w32_kbd.h"
#include "w32_sch.h"
#include "w32_snd.h"
#include "w32_res.h"
#include "w32_sub.h"
#include "w32_cfg.h"
#include "w32_comm.h"

#ifdef ROMEO
#include "juliet.h"
#endif
#ifdef MIDI
#include "w32_midi.h"
#endif


/*
 *	O[o [N
 */
HINSTANCE hAppInstance;					/* AvP[V CX^X */
HINSTANCE hResInstance;					/* \[X CX^X */
HWND hMainWnd;							/* CEChE */
HWND hDrawWnd;							/* h[EChE */
int nErrorCode;							/* G[R[h */
BOOL bMenuLoop;							/* j[[v */
BOOL bMenuExit;							/* j[otO */
BOOL bCloseReq;							/* IvtO */
LONG lCharWidth;						/* LN^ */
LONG lCharHeight;						/* LN^c */
BOOL bSync;								/* sɓ */
BOOL bSyncDisasm[3];					/* tAZuPCɓ */
BOOL bMinimize;							/* ŏtO */
BOOL bActivate;							/* ANeBx[gtO */
#ifndef XM7DASH
BOOL bHideStatus;						/* Xe[^Xo[B */
#endif
HICON hAppIcon;							/* ACRnh */
int nAppIcon;							/* ACRԍ(1,2,3) */
BOOL bNTflag;							/* OS^Cv1(NT) */
BOOL bXPflag;							/* OS^Cv2(XP) */
BOOL bVistaflag;						/* OS^Cv3(Vista/7) */
BOOL bWin8flag;							/* OS^Cv4(Windows 8) */
BOOL bMMXflag;							/* MMXT|[gtO */
BOOL bCMOVflag;							/* CMOVT|[gtO(󖢎gp) */
BOOL bHighPriority;						/* nCvCIeBtO */
POINT WinPos;							/* EBhEʒu */
#ifdef ROMEO
BOOL bRomeo;							/* ROMEOFtO */
#endif
#if ((XM7_VER <= 2) && defined(FMTV151))
BOOL bFMTV151;							/* V2`lR[tO */
#endif
HFONT hFont;							/* TuEBhEptHgnh */
#ifdef XM7DASH
HINSTANCE hInstRes;						/* DLL̃nh */
BOOL bUseLanguageMode;					/* ꃂ[h؂ւ */
BOOL bLanguageMode;						/* ꃂ[h */
BOOL bExistResourceDll;					/* \[XDLL݃tO */
BOOL bFilterRedrawScreen;				/* j[ۂɉʂĕ` */
HMENU hMediaMenu;						/* fBAj[nh */
HMENU hHistoryMenu;						/* qXgj[nh */
HMENU hAspectMenu;						/* AXyNgݒ胁j[nh */
#endif

/*
 *	X^eBbN [N
 */
static CRITICAL_SECTION CSection;		/* NeBJZNV */
static BOOL (WINAPI *WINNLSEnableIME)(HWND, BOOL);
										/* IMEL/ݒpAPI֐ */
static HINSTANCE hInstDLL;				/* user32.dll̃nh */
#ifndef XM7DASH
static HINSTANCE hInstRes;				/* DLL̃nh */
#endif
#ifdef XM7DASH
static WORD wMenuID;					/* j[I */
static int nFlag[SWND_MAXNUM];			/* TuEBhE\ */
static BOOL bContextualMenuLoop;		/* ReLXgj[[v */
#endif

/*
 *	AZu֐̂߂̃vg^Cv錾
 */
extern BOOL CheckMMX(void);				/* MMXΉ`FbN */
extern BOOL CheckCMOV(void);			/* CMOVΉ`FbN(ݖgp) */

/*-[ IMET|[g ]----------------------------------------------------------*/

/*
 *	IMEL/ݒAPI
 *	DLL
 */
void FASTCALL InitIMEDLL(void)
{
	/* user32.dll[h */
	hInstDLL = LoadLibrary("user32.dll");

	/* ֐̐擪AhXݒ */
	if (hInstDLL) {
		WINNLSEnableIME = (BOOL (WINAPI *)(HWND, BOOL))GetProcAddress(
			hInstDLL, "WINNLSEnableIME");

		if (!WINNLSEnableIME) {
			/* s */
			CleanIMEDLL();
		}
	}
}

/*
 *	IMEL/ݒAPI
 *	DLLN[Abv
 */
void FASTCALL CleanIMEDLL(void)
{
	/* DLLǂݍ܂ĂJ */
	if (hInstDLL) {
		FreeLibrary(hInstDLL);
	}

	WINNLSEnableIME = NULL;
	hInstDLL = NULL;
}

/*
 *	IMEL/؂芷 09/09
 */
BOOL FASTCALL EnableIME(HWND hWnd, BOOL flag)
{
	/* WINNLSEnableIME֐ǂݍ߂Ăꍇ͎s */
	if (hInstDLL && WINNLSEnableIME) {
		return WINNLSEnableIME(hWnd, flag);
	}

	/* ̂܂ܕԂlƂ */
	return flag;
}

/*-[ TuEBhET|[g ]-----------------------------------------------*/

/*
 *	eLXgtHg쐬
 *	ĂяoDeleteObject邱
 */
HFONT FASTCALL CreateTextFont(void)
{
	HFONT hFont;

	/* ꔻ肵āAVtgJISEANSIǂ炩̃tHg */
#ifdef XM7DASH
	if ((PRIMARYLANGID(GetSystemDefaultLangID()) == LANG_JAPANESE) &&
		(PRIMARYLANGID(GetUserDefaultLangID()) == LANG_JAPANESE)) {
#else
	if (PRIMARYLANGID(GetUserDefaultLangID()) == LANG_JAPANESE) {
#endif
		/* { */
		hFont = CreateFont(-12, 0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE,
			SHIFTJIS_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY, FIXED_PITCH, "MS Gothic");
	}
	else {
		/* p */
		hFont = CreateFont(-11, 0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE,
			ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY, FIXED_PITCH, NULL);
	}

	ASSERT(hFont);
	return hFont;
}

/*
 *	TuEBhEZbgAbv
 */
static void FASTCALL SetupSubWnd(HWND hWnd)
{
	HFONT hBackup;
	HDC hDC;
	TEXTMETRIC tm;

	ASSERT(hWnd);

	/* EChE[N */
	memset(hSubWnd, 0, sizeof(hSubWnd));
	memset(bShowSubWindow, 0, sizeof(bShowSubWindow));
	InitSubWndWork();

	/* DC擾AtHgZNg */
	hDC = GetDC(hWnd);
	ASSERT(hDC);
	hBackup = SelectObject(hDC, hFont);
	ASSERT(hBackup);

	/* eLXggbN擾 */
	GetTextMetrics(hDC, &tm);

	/* DCN[Abv */
	SelectObject(hDC, hBackup);
	ReleaseDC(hWnd, hDC);

	/* ʂXgA */
	lCharWidth = tm.tmAveCharWidth;
	lCharHeight = tm.tmHeight + tm.tmExternalLeading;
}

/*-[ T|[g ]-------------------------------------------------------*/

/*
 *	\[XCX^X̎擾
 */
HINSTANCE FASTCALL MakeInstance(HINSTANCE hInstance)
{
#ifdef XM7DASH
	char path[_MAX_PATH];
#endif

	/* ftHgł͓̓{ꃊ\[Xgp */
	hResInstance = hInstance;

#ifdef XM7DASH
	LoadCfg_ResourceDll(path);
	hInstRes = LoadLibrary(path);
	if (hInstRes) {
		bExistResourceDll = TRUE;
		if (bLanguageMode) {
			hResInstance = hInstRes;
		}
	}
	else {
		bExistResourceDll = FALSE;
		if (bLanguageMode) {
			bLanguageMode = FALSE;
			DeleteObject(hFont);
			hFont = CreateTextFont();
		}
	}
#else
	/* ꔻ肵āApꃊ\[X̓ǂݍ݂s */
	if (LoadCfg_LanguageMode() ||
		(PRIMARYLANGID(GetUserDefaultLangID()) != LANG_JAPANESE)) {
		/* p */
#if XM7_VER == 3
		hInstRes = LoadLibrary("XM7V3EN.dll");
#elif XM7_VER == 2 && defined(FMTV151)
		hInstRes = LoadLibrary("XM7V2TEN.dll");
#elif XM7_VER == 2
		hInstRes = LoadLibrary("XM7V2EN.dll");
#else
		hInstRes = LoadLibrary("XM7V1EN.dll");
#endif
		if (hInstRes) {
			hResInstance = hInstRes;
		}
	}
#endif

	return hResInstance;
}

/*
 *	\[XCX^X̉
 */
void FASTCALL CleanResourceInstance(void)
{
	if (hInstRes) {
		FreeLibrary(hInstRes);
		hInstRes = NULL;
	}
}

/*-[  ]-----------------------------------------------------------------*/

/*
 *	VMbN
 */
void FASTCALL LockVM(void)
{
	EnterCriticalSection(&CSection);
}

/*
 *	VMAbN
 */
void FASTCALL UnlockVM(void)
{
	LeaveCriticalSection(&CSection);
}

/*-[ h[EChE ]-----------------------------------------------------*/

/*
 * 	EChEvV[W
 */
static LRESULT CALLBACK DrawWndProc(HWND hWnd, UINT message,
								 WPARAM wParam,LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChE쐬 */
		case WM_CREATE:
			/* IME֎~ */
			ImmAssociateContext(hWnd, (HIMC)NULL);
			break;

		/* EChEwi` */
		case WM_ERASEBKGND:
			/* G[ȂȂAwi`悹 */
			if (nErrorCode == 0) {
				return 0;
			}
			break;

		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			OnPaint(hWnd);
			UnlockVM();
			return 0;

		/* j[`FbN (CEBhEɑ΂ăbZ[WĔs) */
		case WM_NCMOUSEMOVE:
		case WM_MOUSEMOVE:
			if (bWin8flag) {
				/* Windows 8Ƃǂ]̏ł܂sȂ̂ */
				PostMessage(hMainWnd, message, wParam, lParam);
			}
			else {
				PostMessage(hMainWnd, message, wParam, 
					MAKELPARAM(LOWORD(lParam), (HIWORD(lParam) +
					GetSystemMetrics(SM_CYMENU))));
			}
			return 0;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	h[EChE쐬
 */
static HWND FASTCALL CreateDraw(HWND hParent)
{
	WNDCLASSEX wcex;
	char szWndName[] = "XM7_Draw";
	RECT rect;
	HWND hWnd;

	ASSERT(hParent);

	/* eEChE̋`擾 */
	GetClientRect(hParent, &rect);

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = DrawWndProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hAppInstance;
	wcex.hIcon = NULL;
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szWndName;
	wcex.hIconSm = LoadIcon(NULL, IDI_APPLICATION);
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	hWnd = CreateWindow(szWndName,
						szWndName,
						WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN,
						0,
						0,
						640,
						400,
						hParent,
						NULL,
						hAppInstance,
						NULL);

	/* ʂA */
	return hWnd;
}

/*-[ CEChE ]-----------------------------------------------------*/

/*
 *	EChE쐬
 */
static void FASTCALL OnCreate(HWND hWnd)
{
	BOOL flag;
#ifdef XM7DASH
	int i;
#endif

	ASSERT(hWnd);

	/* NeBJZNV쐬 */
	InitializeCriticalSection(&CSection);

	/* IME֎~ */
	ImmAssociateContext(hWnd, (HIMC)NULL);

	/* t@Chbv */
	DragAcceptFiles(hWnd, TRUE);

	/* [NGA */
	nErrorCode = 0;
	bMenuLoop = FALSE;
	bCloseReq = FALSE;
	bSync = TRUE;
	bSyncDisasm[0] = TRUE;
	bSyncDisasm[1] = TRUE;
#if XM7_VER == 1
	bSyncDisasm[2] = TRUE;
#endif
	bMinimize = FALSE;
	bActivate = FALSE;
#ifndef XM7DASH
	bHideStatus = FALSE;
#endif
	WinPos.x = -99999;
	WinPos.y = -99999;
#ifdef XM7DASH
	wMenuID = 0;
	bContextualMenuLoop = FALSE;
#endif

#ifdef XM7DASH
	/* TuEBhE\tOҔ̈ */
	for (i=0; i<SWND_MAXNUM; i++) {
		nFlag[i] = -1;
	}
#endif

	/* TuEBhE */
	SetupSubWnd(hWnd);

	/* ROMEOJn by  */
#ifdef ROMEO
	juliet_load();
	if (juliet_prepare() == 0) {
		bRomeo = TRUE;
		juliet_YMF288Reset();
	}
	else {
		bRomeo = FALSE;
	}
#endif

	/* h[EChEAXe[^Xo[쐬 */
	hDrawWnd = CreateDraw(hWnd);
	hStatusBar = CreateStatus(hWnd);

	/* R|[lg */
	LoadCfg();
	InitDraw();
	InitSnd();
	InitKbd();
	InitSch();
#ifdef FDDSND
	InitFDDSnd();
#endif

	/* z}V */
	if (!system_init()) {
		nErrorCode = 1;
		PostMessage(hWnd, WM_USER, 0, 0);
		return;
	}

	/* AZbg */
#ifdef XM7DASH
	LockVM();
#endif
	ApplyCfg();
#ifdef XM7DASH
	UnlockVM();
	RestoreBootMode();
#endif
	system_reset();

	/* R|[lgZNg */
	flag = TRUE;
	if (!SelectDraw(hDrawWnd)) {
		flag = FALSE;
	}
	if (!SelectSnd(hWnd)) {
		flag = FALSE;
	}
	if (!SelectKbd(hWnd)) {
		flag = FALSE;
	}
	if (!SelectSch()) {
		flag = FALSE;
	}

	/* G[R[hZbgAX^[g */
	if (!flag) {
		nErrorCode = 2;
	}
#ifdef XM7DASH
	else {
		/* EBhEԂ𕜌 */
		RestoreFrameWnd(FALSE);

		/* TuEBhE𕜌 */
		if (!bFullScreen && bResumeSwnd) {
			bSwndVisible = bResumeBootSwnd;
			RestoreSubWindow();
		}
		bSwndVisible = TRUE;
	}
#endif
	PostMessage(hWnd, WM_USER, 0, 0);
}

/*
 *	EChEN[Y
 */
#ifdef XM7DASH
static BOOL FASTCALL OnClose(HWND hWnd)
#else
static void FASTCALL OnClose(HWND hWnd)
#endif
{
#ifdef XM7DASH
#if 0
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	char tmp[_MAX_FNAME + _MAX_EXT];
	char buffer[256];
	char string[256];
	int nret;
#else
	char string[256];
#endif
#endif
	int i;

	ASSERT(hWnd);

#ifdef XM7DASH
	/* Xe[gw肳Ă΁Aɍs */
#if 0
#ifdef RSTATE
	if (bResumeState && StatePath[0] != '\0' && bStateSave) {
#else
	if (bResumeState && StatePath[0] != '\0') {
#endif
		/* t@Cl[{gq̂ݎo */
		_splitpath(StatePath, NULL, NULL, fname, ext);
		strncpy(tmp, fname, sizeof(tmp)-1);
		strncat(tmp, ext, sizeof(tmp)-strlen(tmp)-1);
		LoadString(hResInstance, IDS_SAVECLOSE, buffer, sizeof(buffer));
		_snprintf(string, sizeof(string), buffer, tmp);
		EnterMenu(hWnd);
		nret = MessageBox(hWnd, string, "XM7", MB_ICONQUESTION | MB_YESNOCANCEL);
		ExitMenu();
		/* mFʂɂ */
		switch (nret) {
			/* YES */
			case IDYES:
				/* Xe[gZ[u */
				LockVM();
				StopSnd();
				if (bBackupState) {
					/* Xe[g̑Ҕ */
					file_rename(StatePath);
				}
				if (!system_save(StatePath)) {
					LoadString(hResInstance, IDS_STATEERROR, string, sizeof(string));
					MessageBox(hWnd, string, "XM7", MB_ICONSTOP | MB_OK);
				}
				PlaySnd();
				UnlockVM();
				break;

			/* NO */
			case IDNO:
				/* pXNA(Xe[gȂ) */
				StatePath[0] = '\0';
				SetAllMRUFile();
				break;

			/* LZ */
			case IDCANCEL:
				/* N[YȂƂɂ */
				return FALSE;
		}
	}
#else
#ifdef RSTATE
	if (bResumeState && QuickPath[0] != '\0' && bStateSave) {
#else
	if (bResumeState && QuickPath[0] != '\0') {
#endif
		/* Xe[gZ[u */
		LockVM();
		StopSnd();
		if (bBackupState) {
			/* Xe[g̑Ҕ */
			file_rename(QuickPath);
		}
		if (!system_save(QuickPath)) {
			LoadString(hResInstance, IDS_STATEERROR, string, sizeof(string));
			MessageBox(hWnd, string, "XM7", MB_ICONSTOP | MB_OK);
		}
		PlaySnd();
		UnlockVM();
	}
#endif

	/* Xe[^Xo[\ */
	bStatusBar = IsWindowVisible(hStatusBar);
#endif

	/* CEChEx(^XNo[΍) */
	ShowWindow(hWnd, SW_HIDE);

	/* tOAbv */
	LockVM();
	bCloseReq = TRUE;
	UnlockVM();

#ifdef XM7DASH
	/* TuEBhE\tOҔ */
	for (i=0; i<SWND_MAXNUM; i++) {
		nFlag[i] = nSwndFlag[i];
	}

	/* |bvAbv[h͂ŏI */
	if (bPopupSwnd) {
		return TRUE;
	}
#endif

	/* TuEChEɔj */
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i]) {
			DestroyWindow(hSubWnd[i]);
		}
	}

#ifdef XM7DASH
	return TRUE;
#endif
}

/*
 *	EChE폜
 */
static void FASTCALL OnDestroy(void)
{
#ifdef XM7DASH
	int i;

	/* TuEBhE\tO𕜌 */
	for (i=0; i<SWND_MAXNUM; i++) {
		nSwndFlag[i] = nFlag[i];
	}
#endif

	/* TEh~ */
	StopSnd();

#ifdef XM7DASH
	/* SMRUZbg */
	SetAllMRUFile();
#endif

	/* INIt@C */
	SaveCfg();

	/* R|[lg N[Abv */
#ifdef MIDI
	CleanMIDI();
#endif
#ifdef FDDSND
	CleanFDDSnd();
#endif
#ifdef RSC
	CleanCommPort();
#endif
	CleanSch();
	CleanKbd();
	CleanSnd();
	CleanDraw();
	
	/* ROMEOI by  */
#ifdef ROMEO
	if (bRomeo) {
		juliet_YMF288Reset();
		Sleep(1);
	}
	juliet_unload();
	Sleep(1);
#endif

	/* NeBJZNV폜 */
	DeleteCriticalSection(&CSection);

	/* z}V N[Abv */
	system_cleanup();

#if defined(XM7DASH) && defined(FILTERDLL)
	if (((bFilterDll && bUseFilterDll) || bRasterRendering) &&
		bFilterRedrawScreen) {
		/* fXNgbvĕ` */
		InvalidateRect(NULL, NULL, TRUE);
	}
#endif
}

/*
 *	WM_QUIT
 */
void FASTCALL OnQuit(void)
{
	/* EChEnhNA */
	hMainWnd = NULL;
	hDrawWnd = NULL;
	hStatusBar = NULL;

#ifdef XM7DASH
	/* fBAj[폜 */
	DestroyMenu(hMediaMenu);
	hMediaMenu = NULL;

	/* qXgj[폜 */
	DestroyMenu(hHistoryMenu);
	hHistoryMenu = NULL;

	/* AXyNgݒ胁j[폜 */
	DestroyMenu(hAspectMenu);
	hAspectMenu = NULL;
#endif
}

/*
 *	TCYύX
 */
void FASTCALL OnSize(HWND hWnd, WORD cx, WORD cy)
{
	RECT crect;
	RECT wrect;
	RECT trect;
	RECT srect;
	int width;
	int height;
	int wheight;
#if defined(XM7DASH) && defined(FILTERDLL)
	double d;
#endif

	ASSERT(hWnd);
	ASSERT(hDrawWnd);

	/* ŏ̏ꍇ́AȂ */
	if ((cx == 0) && (cy == 0)) {
		return;
	}

	/* tXN[l */
	if (bFullScreen) {
		return;
	}

	/* c[o[AXe[^Xo[̗LlɓAvZ */
#if defined(XM7DASH) && defined(FILTERDLL)
	if (bAspectRatio && bFilterDll && bUseFilterDll) {
		d = dAspectRatio;			/* 0.899(0.4495) */
	}
	else {
		d = 1.000;					/* 1.000(0.5000) */
	}
	switch (nScalingSize) {
		case 0:						/* x1.0(640x400) */
		default:
			width  = 640 * 1.0;
			height = 400 * 1.0 / d;
			break;

		case 1:						/* x1.5(960x600) */
			width  = 640 * 1.5;
			height = 400 * 1.5 / d;
			break;

		case 2:						/* x2.0(1280x800) */
			width  = 640 * 2.0;
			height = 400 * 2.0 / d;
			break;

		case 3:						/* x0.5(320x200) */
			width  = 640 * 0.5;
			height = 400 * 0.5 / d;
			break;

		case 4:						/* x1.8(1152x720) */
			width  = 640 * 1.8;
			height = 400 * 1.8 / d;
			break;

		case 5:						/* x1.2(768x480) */
			width  = 640 * 1.2;
			height = 400 * 1.2 / d;
			break;
	}
#else
	if (bDoubleSize) {
		width = 1280;
		height = 800;
	}
	else {
		width = 640;
		height = 400;
	}
#endif
	memset(&trect, 0, sizeof(trect));
	if (IsWindowVisible(hStatusBar)) {
		GetWindowRect(hStatusBar, &srect);
		wheight = height + (srect.bottom - srect.top);
	}
	else {
		memset(&srect, 0, sizeof(srect));
		wheight = height;
	}

	/* NCAg̈̃TCY擾 */
	GetClientRect(hWnd, &crect);

	/* vTCYƔrA␳ */
	if ((crect.right != width) || (crect.bottom != wheight)) {
		GetWindowRect(hWnd, &wrect);
		wrect.right -= wrect.left;
		wrect.bottom -= wrect.top;
		wrect.right -= crect.right;
		wrect.bottom -= crect.bottom;
		wrect.right += width;
		wrect.bottom += wheight;
		MoveWindow(hWnd, wrect.left, wrect.top, wrect.right, wrect.bottom, TRUE);
	}

	/* CEChEzu */
	MoveWindow(hDrawWnd, 0, trect.bottom, width, height, TRUE);

	/* Xe[^Xo[zu */
	if (IsWindowVisible(hStatusBar)) {
		MoveWindow(hStatusBar, 0, trect.bottom + height, width,
					(srect.bottom - srect.top), TRUE);
		SizeStatus(width);
	}
}

/*
 *	LbN(j[ɂ͖`)
 */
static void FASTCALL OnKick(HWND hWnd)
{
	char buffer[128];

	buffer[0] = '\0';

	/* G[R[h */
	switch (nErrorCode) {
		/* G[Ȃ */
		case 0:
			/* sJn */
			stopreq_flag = FALSE;
			run_flag = TRUE;

#ifdef XM7DASH
			/* EBhEԂ𕜌 */
			if (RestoreFrameWnd(bFullScreen)) {
				/* ő剻w肩AOsɃtXN[ */
				PostMessage(hWnd, WM_COMMAND, IDM_FULLSCREEN, 0);
			}

			/* fBAEXe[g𕜌 */
			RestoreMediaState();

			/* \eXV */
			OnSize(hWnd, 640, 400);
#endif

			/* R}hC */
			OnCmdLine(GetCommandLine());

#ifdef MOUSE
			/* }EXLv`Jn */
			SetMouseCapture(TRUE);
#endif
			break;

		/* VMG[ */
		case 1:
			LoadString(hResInstance, IDS_VMERROR, buffer, sizeof(buffer));
			MessageBox(hWnd, buffer, "XM7", MB_ICONSTOP | MB_OK);
			PostMessage(hWnd, WM_CLOSE, 0, 0);
			break;

		/* R|[lgG[ */
		case 2:
			LoadString(hResInstance, IDS_DXERROR, buffer, sizeof(buffer));
			MessageBox(hWnd, buffer, "XM7", MB_ICONSTOP | MB_OK);
			PostMessage(hWnd, WM_CLOSE, 0, 0);
			break;
	}
}


/*
 * 	EChEvV[W
 */
static LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
								 WPARAM wParam,LPARAM lParam)
{
	PAINTSTRUCT ps;
	BYTE key_code;
	RECT rect;

	/* bZ[W */
	switch (message) {
		/* EChE쐬 */
		case WM_CREATE:
			OnCreate(hWnd);
			break;

		/* EBhEN[Y */
		case WM_CLOSE:
#ifdef XM7DASH
			if (!OnClose(hWnd)) {
				return 0;
			}
#else
			OnClose(hWnd);
#endif
			break;

		/* EChE폜 */
		case WM_DESTROY:
			OnDestroy();
			PostQuitMessage(0);
			return 0;

		/* EBhEړ */
		case WM_MOVE:
			if (!bFullScreen && (!(GetWindowLong(hWnd, GWL_STYLE) &
				(WS_MAXIMIZE | WS_MINIMIZE)))) {
				GetWindowRect(hWnd, &rect);
				WinPos.x = rect.left;
				WinPos.y = rect.top;
			}
			break;

		/* EChETCYύX */
		case WM_SIZE:
			/* ŏ̃L[͑΍ */
			if (wParam == SIZE_MINIMIZED) {
				bMinimize = TRUE;
			}
			else {
				bMinimize = FALSE;
			}
			OnSize(hWnd, LOWORD(lParam), HIWORD(lParam));
			break;

		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			BeginPaint(hWnd, &ps);
			PaintStatus();
			EndPaint(hWnd, &ps);
			UnlockVM();
			return 0;

#if defined(XM7DASH) && defined(FILTERDLL)
		case WM_UNINITMENUPOPUP:
			if (((bFilterDll && bUseFilterDll) || bRasterRendering) &&
				bFilterRedrawScreen) {
				/* EChEĕ` */
				InvalidateRect(hDrawWnd, NULL, FALSE);
			}
			break;
#endif

		/* [U+0:X^[g */
		case WM_USER + 0:
			OnKick(hWnd);
			return 0;

		/* [U+1:R}hC(Win9x) */
		case WM_USER + 1:
			if (!bNTflag) {
				OnCmdLine((LPSTR)wParam);
			}
			break;

#ifdef MOUSE
		/* [U+2:}EXLv`OFF */
		case WM_USER + 2:
			SetMouseCapture(FALSE);
			break;

		/* [U+3:}EXLv`ON */
		case WM_USER + 3:
			SetMouseCapture(TRUE);
			break;
#endif

		/* R}hC(WinNT) */
		case WM_COPYDATA:
			if (bNTflag) {
				if (((PCOPYDATASTRUCT)lParam)->dwData == 0x01374d58) {
					OnCmdLine((LPSTR)((PCOPYDATASTRUCT)lParam)->lpData);
				}
			}
			break;

		case WM_KEYDOWN:
			if ((LOBYTE(HIWORD(lParam)) == 0x36) && bVistaflag) {
				/* Vistaȍ~̉EVtg΍ */
				bNTkeyPushFlag[KNT_RSHIFT] = TRUE;
				return 0;
			}
			break;

		case WM_SYSKEYUP:
		case WM_KEYUP:
			if ((LOBYTE(wParam) == VK_F10) && kbd_table[DIK_F10]) {
				/* F10ׂ(^^; */
				return 0;
			}
			if ((LOBYTE(wParam) == VK_MENU) && !(lParam & 0x01000000) &&
				 kbd_table[DIK_LMENU]) {
				/* Altׂ(^^; */
				return 0;
			}
			if ((LOBYTE(wParam) == VK_MENU) && (lParam & 0x01000000) &&
				 kbd_table[DIK_RMENU]) {
				/* EAltׂ(^^; */
				return 0;
			}
			if ((LOBYTE(HIWORD(lParam)) == 0x36) && bVistaflag) {
				/* Vistaȍ~̉EVtg΍ */
				bNTkeyPushFlag[KNT_RSHIFT] = FALSE;
				return 0;
			}
			if (((LOBYTE(wParam) == VK_MENU) || (LOBYTE(wParam) == VK_F10)) && 
				 bWin8flag) {
				/* Windows 8 ΍􏈗 */
				if (bFullScreen) {
					if (!bMenuLoop) {
						EnterMenu(hWnd);
						bMenuLoop = TRUE;
					}
					else {
						ExitMenu();
						OnExitMenuLoop();
#ifdef XM7DASH
						wMenuID = 0;
#endif
						bMenuLoop = FALSE;
					}
				}
			}
			break;

		/* j[[vJn */
		case WM_ENTERMENULOOP:
			if (!bMenuLoop && !bWin8flag) {
				EnterMenu(hWnd);
			}
			bMenuLoop = TRUE;
			break;

		/* j[[vI */
		case WM_EXITMENULOOP:
#ifdef XM7DASH
			if (bContextualMenuLoop) {
				break;
			}
#endif
			if (bMenuLoop && !bWin8flag) {
				ExitMenu();
				OnExitMenuLoop();
#ifdef XM7DASH
				wMenuID = 0;
#endif
			}
			SetMenuExitTimer();
			bMenuLoop = FALSE;
			break;

		/* j[I */
		case WM_MENUSELECT:
			OnMenuSelect(wParam);
#ifdef XM7DASH
			if (!bContextualMenuLoop) {
				wMenuID = LOWORD(wParam);
			}
#endif
			break;

		/* j[`FbN */
		case WM_NCMOUSEMOVE:
		case WM_MOUSEMOVE:
			if (bFullScreen) {
				if (HIWORD(lParam) < GetSystemMetrics(SM_CYMENUSIZE)) {
					if (!bMenuLoop) {
						EnterMenu(hMainWnd);
						bMenuLoop = TRUE;
					}
				}
				else {
					if (bMenuLoop) {
						ExitMenu();
						OnExitMenuLoop();
#ifdef XM7DASH
						wMenuID = 0;
#endif
						bMenuLoop = FALSE;
					}
				}
			}
			break;

		/* j[|bvAbv */
		case WM_INITMENUPOPUP:
			if (!HIWORD(lParam)) {
				OnMenuPopup(hWnd, (HMENU)wParam, (UINT)LOWORD(lParam));
				return 0;
			}
			break;

		/* j[R}h */
		case WM_COMMAND:
			EnterMenu(hWnd);
			OnCommand(hWnd, LOWORD(wParam));
			ExitMenu();
			return 0;

#ifdef XM7DASH
		/* ReLXgj[ */
		case WM_RBUTTONUP:
			if (bContextualMenuLoop) {
				break;
			}
			if (bMenuLoop) {
				bContextualMenuLoop = TRUE;
				OnContextualMenu(hWnd, wMenuID, LOWORD(lParam), HIWORD(lParam));
				bContextualMenuLoop = FALSE;
				return 0;
			}
			break;
#endif

		/* t@Chbv */
		case WM_DROPFILES:
			OnDropFiles((HANDLE)wParam);
			return 0;

#if defined(XM7DASH) && defined(FILTERDLL) && defined(MOUSE) && defined(WM_MOUSEWHEEL)
		/* }EXzC[ */
		case WM_MOUSEWHEEL:
			if ((uMidBtnMode != MOSCAP_WHEELWM)) {
				int ntmp;
				if (bFullScreen) {
					break;
				}
				ntmp = dwFilterDllDepFlg + GET_WHEEL_DELTA_WPARAM(wParam) / 10;
				if (ntmp < 0) {
					ntmp = 0;
				}
				else if (ntmp > 255) {
					ntmp = 255;
				}
				if (SetFilterParameter((WORD)ntmp)) {
					PostMessage(hDrawWnd, WM_PAINT, 0, 0);
				}
			}
			break;
#endif

		/* ANeBx[g */
		case WM_ACTIVATE:
			if (LOWORD(wParam) == WA_INACTIVE) {
				bActivate = FALSE;
			}
			else {
				if (!bActivate) {
					InvalidateRect(hDrawWnd, NULL, FALSE);
				}
				bActivate = TRUE;
			}

#ifdef MOUSE
			SetMouseCapture(bActivate);
#endif
			break;

		/* I[i[h[ */
		case WM_DRAWITEM:
			if (wParam == ID_STATUS_BAR) {
				OwnerDrawStatus((LPDRAWITEMSTRUCT)lParam);
				return TRUE;
			}
			break;

		/* ʃfUCύX */
		case WM_SYSCOLORCHANGE:
			ResizeStatus(hWnd, hStatusBar);
			break;

		/* e[}ύX */
		case WM_THEMECHANGED:
			if (bXPflag) {
				ChangeStatusBorder(hStatusBar);
			}
			break;

		/* ^C}[ */
		case WM_TIMER:
			if ((wParam >= 0x80) && (wParam < 0x100)) {
				/* L[J */
				key_code = (BYTE)(wParam & 0x7f);
				if (bNTkeyMakeFlag[key_code]) {
					keyboard_break(key_code);
					bNTkeyMakeFlag[key_code] = FALSE;
				}
				KillTimer(hMainWnd, wParam);
			}
			else switch (wParam) {
				case ID_STATUS_BAR :
					/* Xe[^XbZ[W */
					if (bStatusMessage) {
						EndStatusMessage();
					}
					KillTimer(hMainWnd, ID_STATUS_BAR);
					break;
			}
			break;

		/* dǗ */
		case WM_POWERBROADCAST:
			timeEndPeriod(uTimerResolution);
			timeBeginPeriod(uTimerResolution);
			break;

		/* VXeR}h */
		case WM_SYSCOMMAND:
			/* XN[Z[oEj^ߓd[h} */
			if ((wParam == SC_SCREENSAVE) || (wParam == SC_MONITORPOWER)) {
				return 1;
			}
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*-[ AvP[V ]-----------------------------------------------------*/

/*
 *	XM7EChEAR}hCn
 */
static BOOL FASTCALL FindXM7Wnd(void)
{
	COPYDATASTRUCT	cds;
	HWND hWnd;
	char string[128];

	/* EChENXŌ */
	hWnd = FindWindow("XM7", NULL);
	if (!hWnd) {
		return FALSE;
	}

	/* eLXg擾 */
	GetWindowText(hWnd, string, sizeof(string));
	string[4] = '\0';
	if (strcmp("XM7 ", string) != 0) {
		return FALSE;
	}

	/* bZ[W𑗐M */
	if (bNTflag) {
		/* NTłWM_COPYDATAg */
		cds.dwData = 0x01374d58;
		cds.lpData = (void *)GetCommandLine();
		cds.cbData = lstrlen(GetCommandLine()) + 1;
		SendMessage(hWnd, WM_COPYDATA, (WPARAM)NULL, (LPARAM)&cds);
	}
	else {
		/* Win9x/Meł͏]Ɠ@ŃbZ[WM */
		SendMessage(hWnd, WM_USER + 1, (WPARAM)GetCommandLine(), (LPARAM)NULL);
	}

	return TRUE;
}

/*
 *	CX^X
 */
static HWND FASTCALL InitInstance(HINSTANCE hInst, int nCmdShow)
{
	HWND hWnd;
	HMENU hMenu;
	WNDCLASSEX wcex;
	char szAppName[] = "XM7";
	char szWindowName[128];
	int xx, yy;
	RECT rect;

	ASSERT(hInst);

	/* EChEnhNA */
	hMainWnd = NULL;
	hDrawWnd = NULL;
	hStatusBar = NULL;

	/* EBhETCYvZ */
	rect.left = 0;
	rect.top = 0;
	if (LoadCfg_DoubleSize()) {
		rect.right = 1280;
		rect.bottom = 800;
	}
	else {
		rect.right = 640;
		rect.bottom = 400;
	}
	AdjustWindowRect(&rect, WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION |
							WS_BORDER | WS_CLIPCHILDREN | WS_MINIMIZEBOX,
							TRUE);
	xx = rect.right - rect.left;
	yy = rect.bottom - rect.top;

	/* ACR[h */
#ifdef XM7DASH
#if XM7_VER >= 3
#ifdef XM7DASH
	if (LoadCfg_DownGradeMode()) {
		hAppIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_40EXICON2));
	}
	else if (LoadCfg_Gravestone()) {
		hAppIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_40SXICON));
	}
	else {
		hAppIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_40EXICON));
	}
#else
	hAppIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_40EXICON2));
#endif
	nAppIcon = 3;
#elif XM7_VER >= 2
	hAppIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_77AVICON));
	nAppIcon = 2;
#else
	hAppIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_FM77ICON));
	nAppIcon = 3;
#endif
#else
#if XM7_VER >= 3
	hAppIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_APPICON));
#elif XM7_VER >= 2
	hAppIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_AVICON));
#else
	hAppIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_APPICON));
#endif
	nAppIcon = 255;
#endif
	MakeInstance(hInst);

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = WndProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hInst;
	wcex.hIcon = hAppIcon;
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = MAKEINTRESOURCE(IDR_MAINMENU);
	wcex.lpszClassName = szAppName;
	wcex.hIconSm = hAppIcon;
	RegisterClassEx(&wcex);

	/* EBhE^Cg쐬 */
#ifdef XM7DASH
	strncpy(szWindowName, szAppName, sizeof(szWindowName));
	strncat(szWindowName, " ", sizeof(szWindowName) - strlen(szWindowName) - 1);
#else
	strncpy(szWindowName, szAppName, 128);
	strncat(szWindowName, " ", 128 - strlen(szWindowName) - 1);
#endif

	/* j[쐬 */
	hMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MAINMENU));

#ifdef XM7DASH
	/* fBAj[[h */
	hMediaMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MEDIAMENU));

	/* qXgj[[h */
	hHistoryMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_HISTORYMENU));

	/* AXyNgݒ胁j[[h */
	hAspectMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_ASPECTMENU));
#endif

	/* EChE쐬 */
#ifdef XM7DASH
	hMainWnd =
#endif
	hWnd = CreateWindow(szAppName,
						szWindowName,
						WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION | WS_BORDER |
						WS_CLIPCHILDREN | WS_MINIMIZEBOX,
						CW_USEDEFAULT,
						CW_USEDEFAULT,
						xx,
						yy,
						NULL,
						hMenu,
						hInst,
						NULL);
	if (!hWnd) {
		return NULL;
	}

	/* EBhEʒuݒ */
	if ((WinPos.x == -99999) && (WinPos.y == -99999)) {
		GetWindowRect(hWnd, &rect);
		WinPos.x = (GetSystemMetrics(SM_CXSCREEN) - rect.right) / 2;
		WinPos.y = (GetSystemMetrics(SM_CYSCREEN) - rect.bottom) / 2;
	}
	SetWindowPos(hWnd, HWND_NOTOPMOST, WinPos.x, WinPos.y, 0, 0, SWP_NOSIZE);
	OnSize(hWnd, 640, 400);

	/* EChE\ */
	if (nErrorCode == 0) {
		ShowWindow(hWnd, nCmdShow);
		UpdateWindow(hWnd);
	}

	return hWnd;
}

/*
 *	WinMain
 */
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
						LPSTR lpszCmdParam, int nCmdShow)
{
	MSG msg;
	HACCEL hAccel;
	OSVERSIONINFO osvi;
	STICKYKEYS sk;
	STICKYKEYS skb;
	char buffer[128];
	int osver;

	UNUSED(hPrevInstance);
	UNUSED(lpszCmdParam);

	/* COM */
	if (FAILED(CoInitialize(NULL))) {
		LoadString(hInstance, IDS_COINITERROR, buffer, sizeof(buffer));
		MessageBox(NULL, buffer, "XM7", MB_ICONSTOP | MB_OK);

		return -1;
	}

	/* RRg[ */
	InitCommonControls();

	/* 삵ĂOSWinNTn */
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osvi);
	osver = osvi.dwMajorVersion * 100 + osvi.dwMinorVersion;

	/* NT`FbN */
	if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT) {
		bNTflag = TRUE;
	}
	else {
		bNTflag = FALSE;
	}

	/* XP`FbN */
	if (osver >= 501) {
		bXPflag = TRUE;
	}
	else {
		bXPflag = FALSE;
	}

	/* Windows Vista`FbN */
	if (osver >= 600) {
		bVistaflag = TRUE;
	}
	else {
		bVistaflag = FALSE;
	}

	/* Windows 8`FbN */
	if (osver >= 602) {
		bWin8flag = TRUE;
	}
	else {
		bWin8flag = FALSE;
	}

	/* XM7`FbNAR}hCn */
	if (FindXM7Wnd()) {
		/* COMI */
		CoUninitialize();

		return 0;
	}

#ifdef XM7DASH
	/* ̎擾 */
	bLanguageMode = LoadCfg_LanguageMode();
#endif

	/* etO擾Eݒ */
	bMMXflag = CheckMMX();
	bCMOVflag = CheckCMOV();
#ifdef ROMEO
	bRomeo = FALSE;
#endif
#if ((XM7_VER <= 2) && defined(FMTV151))
	bFMTV151 = TRUE;
#endif

	/* DLL */
	InitThemeDLL();
	InitIMEDLL();

	/* tHgnh擾 */
	hFont = CreateTextFont();

	/* CX^XۑA */
	hAppInstance = hInstance;
#ifdef XM7DASH
	InitInstance(hInstance, nCmdShow);
#else
	hMainWnd = InitInstance(hInstance, nCmdShow);
#endif
	if (!hMainWnd) {
		CleanIMEDLL();
		CleanThemeDLL();
		CleanResourceInstance();
		DeleteObject(hFont);

		/* COMI */
		CoUninitialize();

		return -1;
	}

	/* ŒL[@\̖ */
	sk.cbSize = sizeof(STICKYKEYS);
	SystemParametersInfo(SPI_GETSTICKYKEYS, sizeof(STICKYKEYS), &sk, 0);
	skb = sk;
	sk.dwFlags &= ~(SKF_HOTKEYACTIVE | SKF_AVAILABLE | SKF_STICKYKEYSON);
	SystemParametersInfo(SPI_SETSTICKYKEYS, sizeof(STICKYKEYS), &sk, 0);

	/* ANZ[^[h */
	hAccel = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR));
	ASSERT(hAccel);

	/* bZ[W [v */
	while (hMainWnd) {
		/* bZ[W`FbNA */
		if (PeekMessage(&msg, 0, 0, 0, PM_NOREMOVE)) {
			if (!GetMessage(&msg, NULL, 0, 0)) {
				OnQuit();
				break;
			}

#ifdef MOUSE
			/* {^ɂ}EX[h؂ւ */
			if ((msg.message == WM_MBUTTONDOWN) &&
				(msg.wParam == MK_MBUTTON) &&
				(uMidBtnMode == MOSCAP_WMESSAGE)) {
				/* {^ꂽ}EX[h؂芷 */
				EnterMenu(hMainWnd);
				OnCommand(hMainWnd, IDM_MOUSEMODE);
				ExitMenu();
			}

			/* zC[ɂ}EX[h؂ւ(Unz) */
#ifdef WM_MOUSEWHEEL
			if ((msg.message == WM_MOUSEWHEEL) &&
				(uMidBtnMode == MOSCAP_WHEELWM)) {
				if ((short)HIWORD(msg.wParam) > 0) {
					EnterMenu(hMainWnd);
					OnCommand(hMainWnd, IDM_MOUSEON);
					ExitMenu();
				}
				if ((short)HIWORD(msg.wParam) < 0) {
					EnterMenu(hMainWnd);
					OnCommand(hMainWnd, IDM_MOUSEOFF);
					ExitMenu();
				}
			}
#endif
#endif

			/* NTL[΍􏈗 */
			if (msg.message == WM_KEYDOWN) {
				switch (LOBYTE(msg.wParam)) {
					case VKNT_CAPITAL	:	/* CapsLock */
					case VK_CAPITAL		:	bNTkeyPushFlag[KNT_CAPS] = TRUE;
											break;
					case VKNT_KANA		:
					case VKNT_KATAKANA	:	/* J^Ji/Ђ炪,Ji(PC-98) */
					case VK_KANA		:	bNTkeyPushFlag[KNT_KANA] = TRUE;
											break;
					case VKNT_KANJI1	:	/* p/Sp */
					case VKNT_KANJI2	:	bNTkeyPushFlag[KNT_KANJI] = TRUE;
											break;
				}
			}

			if (!TranslateAccelerator(hMainWnd, hAccel, &msg)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			continue;
		}

#ifndef XM7DASH
		/* tXN[v */
		if (bFullRequested) {
			PostMessage(hMainWnd, WM_COMMAND, IDM_FULLSCREEN, 0);
			bFullRequested = FALSE;
		}
#endif

		/* Xe[^X`恕X[v */
		if (nErrorCode == 0) {
			/* Windows Vista/7̏ꍇAŃTuEBhEXV */
			if (bVistaflag) {
				DrawSubWindow();
			}

			DrawStatus();
#if XM7_VER == 1 && defined(SFDC) && defined(XM7DASH)
			ReSizeFDC();
#endif	/* XM7_VER == 1 && defined(SFDC) && defined(XM7DASH) */
#if XM7_VER == 1 && defined(BUBBLE) && defined(XM7DASH)
			ReSizeBMC();
#endif	/* XM7_VER == 1 && defined(BUBBLE) && defined(XM7DASH) */
#if XM7_VER != 2 && defined(XM7DASH)
			ReSizeMMR();
#endif	/* defined(XM7DASH) */
			ReSizeOPNReg();
			ReSizeOPNDisp();

			/* ƈx */
			Sleep(16);
		}
	}

	/* tHgnh */
	DeleteObject(hFont);

	/* DLLJ */
	CleanIMEDLL();
	CleanThemeDLL();
	CleanResourceInstance();

	/* IɃ}EXNbv */
	ClipCursor(0);

	/* ŒL[@\̕ */
	SystemParametersInfo(SPI_SETSTICKYKEYS, sizeof(STICKYKEYS), &skb,  0);

	/* COMI */
	CoUninitialize();

	/* I */
	return msg.wParam;
}

#endif	/* _WIN32 */
