/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2014 ohD(yasushi@tanaka.net)
 *	Copyright (C) 2001-2014 Ryu Takegami
 *	Copyright (C) 2010-2014 Toma
 *
 *	[ Win32API TuEBhER ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include <stdlib.h>
#include "xm7.h"
#ifdef XM7DASH
#include "mouse.h"
#endif
#include "keyboard.h"
#include "rtc.h"
#include "mmr.h"
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
#include "jcard.h"
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
#include "dmac.h"
#endif
#include "subctrl.h"
#include "mainetc.h"
#include "w32.h"
#include "w32_res.h"
#include "w32_sub.h"
#ifdef XM7DASH
#include "w32_kbd.h"
#endif

/*
 *	X^eBbN [N
 */
static BYTE *pKeyboard;					/* L[{[h Drawobt@ */
static BYTE *pMMR;						/* MMR Drawobt@ */
#if XM7_VER != 2 && defined(XM7DASH)
static UINT nMMR;						/* MMR x */
#endif
#if XM7_VER == 1 && defined(XM7DASH)
static BYTE nSegDsp;					/* MMRZOg\Jnʒu */
static BYTE nSegDat;					/* ݂MMRZOg */
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))
static BYTE *pDMAC;						/* DMAC Drawobt@ */
#endif
#ifdef XM7DASH
static BYTE *pIRQ;						/* IRQ Drawobt@ */
#endif

/*-[ MMREChE ]---------------------------------------------------------*/

/*
 *	MMREChE
 *	MMR}bv ZbgAbv
 */
static void FASTCALL SetupMMRMap(BYTE *p, int y, int x)
{
	char string[128];
	char temp[16];
	int i, j;
	BYTE tmp;
#ifdef XM7DASH
	int segnum = 4;
#if XM7_VER == 1
	int segtop = 0;
#endif
#endif

	ASSERT(p);
	ASSERT(y > 0);
	ASSERT(x > 0);

	/* ^Cg */
	strncpy(string, "MMR Map:", 128);
	memcpy(&p[x * y], string, strlen(string));
	y++;

#if XM7_VER != 2 && defined(XM7DASH)
	if (nMMR) {
		segnum = 8;
#if XM7_VER == 1
		segtop = (mmr_seg & ~(segnum - 1));
#endif
	}
#if XM7_VER == 1
	if (nSegDat != mmr_seg) {
		nSegDat = mmr_seg;
		nSegDsp = segtop;
	}
#endif
#endif

	/* eyƂ */
#ifdef XM7DASH
	for (i=0; i<segnum; i++) {
#elif XM7_VER >= 3
	for (i=0; i<8; i++) {
#else
	for (i=0; i<4; i++) {
#endif
		/* 16i_v쐬 */
#if XM7_VER == 1 && defined(XM7DASH)
		_snprintf(string, 128, "Segment%2d-", nSegDsp + i);
#else
		_snprintf(string, 128, "Segment %d-", i);
#endif
		for (j=0; j<16; j++) {
#if XM7_VER == 1 && defined(XM7DASH)
			tmp = mmr_reg[(nSegDsp + i) * 0x10 + j];
#else
			tmp = mmr_reg[i * 0x10 + j];
#endif
#if XM7_VER >= 3
			if (!mmr_ext) {
				tmp &= 0x3f;
			}
#elif XM7_VER == 1 && defined(XM7DASH)
			if (!nMMR) {
				tmp &= 0x3f;
			}
#else
			tmp &= 0x3f;
#endif
			_snprintf(temp, 16, "%02X", tmp);
			strncat(string, temp, 128 - strlen(string) - 1);
		}

		/* MMRgpŃZOgv΁A] */
#if XM7_VER == 1 && defined(XM7DASH)
		if (mmr_flag && (mmr_seg == (nSegDsp + i))) {
#else
		if (mmr_flag && (mmr_seg == i)) {
#endif
			for (j=10;;j++) {
				if (string[j] == '\0') {
					break;
				}
				string[j] |= 0x80;
			}
		}

		/* Zbg */
		memcpy(&p[x * y], string, strlen(string));
		y++;
	}
}

/*
 *	MMREChE
 *	}bv ZbgAbv
 */
static void FASTCALL SetupMMRMemory(BYTE *p, int y, int x)
{
	char string[128];
	char temp[16];
	BYTE mem[16];
	int i, j, k;
#ifdef XM7DASH
	int nsize;
#endif

	ASSERT(p);
	ASSERT(y > 0);
	ASSERT(x > 0);

	/* mem16oCgɁAԂݒ */
	for (i=0; i<16; i++) {
		mem[i] = (BYTE)(0x30 + i);
	}
	/* MMR */
	if (mmr_flag) {
		for (i=0; i<16; i++) {
#if XM7_VER >= 3
			if (mmr_ext) {
				mem[i] = mmr_reg[mmr_seg * 0x10 + i];
			}
			else {
				mem[i] = (BYTE)(mmr_reg[(mmr_seg & 3) * 0x10 + i] & 0x3f);
			}
#elif XM7_VER == 1 && defined(XM7DASH)
			if (nMMR) {
				mem[i] = mmr_reg[mmr_seg * 0x10 + i];
			}
			else {
				mem[i] = (BYTE)(mmr_reg[(mmr_seg & 3) * 0x10 + i] & 0x3f);
			}
#else
			mem[i] = (BYTE)(mmr_reg[(mmr_seg & 3) * 0x10 + i] & 0x3f);
#endif
		}
	}

	/* ^Cg */
	strncpy(string, "Physical Memory Map:", 128);
	memcpy(&p[x * y], string, strlen(string));
	y++;

	/* eyƂ */
#ifdef XM7DASH
#if XM7_VER >= 3
	if (nMMR) {
		nsize = 16;
	}
	else {
		nsize = 4;
	}
#elif XM7_VER == 1
	if (nMMR) {
		nsize = 16;
	}
	else {
		nsize = 4;
	}
#else
	nsize = 4;
#endif
	for (i=0; i<nsize; i++) {
#else
#if XM7_VER == 3
	for (i=0; i<16; i++) {
#else
	for (i=0; i<4; i++) {
#endif
#endif
		/* 16i_v쐬 */
		switch (i) {
#if XM7_VER >= 2
			case 0:
#if XM7_VER == 3
				strncpy(string, "Extend 0  ", 128);
#else
				strncpy(string, "Extend    ", 128);
#endif
				break;
			case 1:
				strncpy(string, "Sub CPU   ", 128);
				break;
			case 2:
				strncpy(string, "Japanese  ", 128);
				break;
			/* ͂Ђǂ */
#if XM7_VER == 3
			case 4:
			case 5:
			case 6:
			case 7:
			case 8:
			case 9:
			case 10:
			case 11:
			case 12:
			case 13:
			case 14:
			case 15:
				_snprintf(string, 128,  "Extend %X  ", i);
				break;
#endif
#else
			case 0:
			case 1:
			case 2:
#ifdef XM7DASH
			case 4:
			case 5:
			case 6:
			case 7:
			case 8:
			case 9:
			case 10:
			case 11:
			case 12:
			case 13:
			case 14:
			case 15:
#endif
				_snprintf(string, 128,  "Extend %X  ", i);
				break;
#endif
			case 3:
				strncpy(string, "Standard  ", 128);
				break;
			default:
				/* 肦Ȃ */
				ASSERT(FALSE);
		}
		for (j=0; j<16; j++) {
			_snprintf(temp, 16, "%02X", i * 0x10 + j);
			/* ] */
			for (k=0; k<16; k++) {
				if (mem[k] == (i * 0x10 + j)) {
					temp[0] |= 0x80;
					temp[1] |= 0x80;
				}
			}
			/* ǉ */
			strncat(string, temp, 128 - strlen(string) - 1);
		}

		/* Zbg */
		memcpy(&p[x * y], string, strlen(string));
		y++;
	}
}

/*
 *	MMREChE
 *	ZbgAbv
 */
static void FASTCALL SetupMMR(BYTE *p, int x, int y)
{
	char string[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* MMR */
	if (mmr_flag) {
		_snprintf(string, 128,  "MMR  Enable(Seg:%1d)", mmr_seg);
	}
	else {
		strncpy(string, "MMR        Disable", 128);
	}
	memcpy(&p[x * 0 + 0], string, strlen(string));

	/* TWR */
	if (twr_flag) {
		_snprintf(string, 128,  "TWR  Enable($%02X00)", (twr_reg + 0x7c) & 0xff);
	}
	else {
		strncpy(string, "TWR        Disable", 128);
	}
	memcpy(&p[x * 1 + 0], string, strlen(string));

#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JCARD) && defined(XM7DASH))
	/* ROM */
	if (!dicrom_en) {
		strncpy(string, "DIC        Disable", 128);
	}
	else {
#if XM7_VER >= 3
		if (extrom_sel) {
			_snprintf(string, 128,  "DIC  ExtROM(Bank %2d)", dicrom_bank);
		}
		else {
			_snprintf(string, 128,  "DIC  Enable(Bank %2d)", dicrom_bank);
		}
#else
		_snprintf(string, 128,  "DIC  Enable(Bank %2d)", dicrom_bank);
#endif
	}
	memcpy(&p[x * 2 + 0], string, strlen(string));
#endif

#if XM7_VER == 1 && defined(XM7DASH)
	/* gRAMoNZNg */
	if (mmr_extrmw == EXRAM_64KB) {
		_snprintf(string, 128,  "Bank %2d", bsr_reg);
		memcpy(&p[x * 2 + 11], string, strlen(string));
	}
#endif

#if XM7_VER >= 2
	/* u[gRAM */
	strncpy(string, "Boot RAM          R", 128);
	if (bootram_rw) {
#ifdef XM7DASH
		strncat(string, "W", 128 - strlen(string) - 1);
#else
		strncat(string, "W", 128 - 1 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "O", 128 - strlen(string) - 1);
#else
		strncat(string, "O", 128 - 1 - 1);
#endif
	}
	memcpy(&p[x * 0 + 22], string, strlen(string));

	/* CjVG[gROM */
	strncpy(string, "Initiate ROM ", 128);
	if (initrom_en) {
#ifdef XM7DASH
		strncat(string, " Enable", 128 - strlen(string) - 1);
#else
		strncat(string, " Enable", 128 - 7 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "Disable", 128 - strlen(string) - 1);
#else
		strncat(string, "Disable", 128 - 7 - 1);
#endif
	}
	memcpy(&p[x * 1 + 22], string, strlen(string));
#else
	/* F-BASIC ROM */
	strncpy(string, "F-BASIC ROM  ", 128);
	if (!basicrom_en) {
#ifdef XM7DASH
		strncat(string, "Disable", 128 - strlen(string) - 1);
#else
		strncat(string, "Disable", 128 - 7 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, " Enable", 128 - strlen(string) - 1);
#else
		strncat(string, " Enable", 128 - 7 - 1);
#endif
	}
	memcpy(&p[x * 0 + 22], string, strlen(string));

	/* u[gROM/RAM */
	strncpy(string, "Boot Area        ", 128);
	if (bootram_rw) {
#ifdef XM7DASH
		strncat(string, "RAM", 128 - strlen(string) - 1);
#else
		strncat(string, "RAM", 128 - 3 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "ROM", 128 - strlen(string) - 1);
#else
		strncat(string, "ROM", 128 - 3 - 1);
#endif
	}
	memcpy(&p[x * 1 + 22], string, strlen(string));

#if XM7_VER == 1 && defined(XM7DASH)
	/* gRAM/MMR/WBR */
	strncpy(string, "Extend MMR   ", sizeof(string));
	if (mmr_extrmw == EXRAM_960KB) {
		strncat(string, " Enable", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Disable", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[x * 2 + 22], string, strlen(string));
#endif
#endif

#if XM7_VER >= 3
	/* gMMR */
	strncpy(string, "Extend MMR   ", 128);
	if (mmr_ext) {
#ifdef XM7DASH
		strncat(string, " Enable", 128 - strlen(string) - 1);
#else
		strncat(string, " Enable", 128 - 7 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "Disable", 128 - strlen(string) - 1);
#else
		strncat(string, "Disable", 128 - 7 - 1);
#endif
	}
	memcpy(&p[x * 2 + 22], string, strlen(string));

	/* MMR}bv */
	SetupMMRMap(p, 4, x);

	/* }bv */
#ifdef XM7DASH
	if (nMMR) {
		SetupMMRMemory(p, 14, x);
	}
	else {
		SetupMMRMemory(p, 10, x);
	}
#else
	SetupMMRMemory(p, 14, x);
#endif
#else
	/* MMR}bv */
	SetupMMRMap(p, 3, x);

	/* }bv */
#if XM7_VER == 1 && defined(XM7DASH)
	if (nMMR) {
		SetupMMRMemory(p, 13, x);
	}
	else {
		SetupMMRMemory(p, 9, x);
	}
#else
	SetupMMRMemory(p, 9, x);
#endif
#endif
}

/*
 *	MMREChE
 *	`
 */
static void FASTCALL DrawMMR(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pMMR) {
		return;
	}
	SetupMMR(pMMR, x, y);

	/* ` */
	DrawWindowText2(hDC, pMMR, x, y);
}

/*
 *	MMREChE
 *	tbV
 */
void FASTCALL RefreshMMR(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_MMR] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_MMR];
	hDC = GetDC(hWnd);
	DrawMMR(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

#if XM7_VER != 2 && defined(XM7DASH)
/*
 *	MMREChE
 *	TCYύX
 */
void FASTCALL ReSizeMMR(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	int cx;
	int cy;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_MMR] == NULL) {
		return;
	}

	/* ŏȂAȂ */
	hWnd = hSubWnd[SWND_MMR];
	if (IsIconic(hWnd)) {
		return;
	}

	/* WIgύX`FbN */
	uMode = 0;
	cx = 42;
#if XM7_VER >= 3
	if ((fm7_ver >= 3) && (!bDownGradeMode)) {
		cy = 31;
		uMode = 1;
	}
	else {
		cy = 15;
	}
#elif XM7_VER == 1
	if ((fm_subtype == FMSUB_FM77) &&
		(mmr_extrmw == EXRAM_960KB)) {
		cy = 30;
		uMode = 1;
	}
	else {
		cy = 14;
	}
#else
	cy = 14;
#endif

	if (uMode == nMMR) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	nMMR = uMode;

	/* obt@ */
	free(pMMR);
	pMMR = malloc(2 * cx * cy);
	memset(pMMR, 0xff, 2 * cx * cy);
	ASSERT(pMMR);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += cx * lCharWidth;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += cy * lCharHeight;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
#ifdef XM7DASH
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}
#endif

	/* AbN */
	UnlockVM();
}
#endif

/*
 *	MMREChE
 *	ĕ`
 */
static void FASTCALL PaintMMR(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pMMR;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawMMR(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

#if XM7_VER == 1 && defined(XM7DASH)
/*
 *	MMREChE
 *	MSRy[WRg[
 */
static void FASTCALL PageMMR(BYTE vkey)
{
	/* MMRĝ */
	if (!nMMR) {
		return;
	}

	switch (vkey) {
		/* [] */
		case VK_UP:
			if (nSegDsp > 0x00) {
				nSegDsp -= 1;
			}
			break;

		/* [] */
		case VK_DOWN:
			if (nSegDsp < 0x38) {
				nSegDsp += 1;
			}
			break;

		/* [][PageUp] */
		case VK_LEFT:
		case VK_PRIOR:
			if (nSegDsp > 0x08) {
				nSegDsp -= 8;
			}
			else {
				nSegDsp = 0;
			}
			break;

		/* [][PageDown] */
		case VK_RIGHT:
		case VK_NEXT:
			if (nSegDsp < 0x30) {
				nSegDsp += 8;
			}
			else {
				nSegDsp = 0x38;
			}
			break;

		/* [Home] */
		case VK_HOME:
			nSegDsp = 0;
			break;

		/* [End] */
		case VK_END:
			nSegDsp = 0x38;
			break;

		default:
			break;
	}
}
#endif	/* XM7_VER == 1 && defined(XM7DASH) */

/*
 *	MMREChE
 *	EChEvV[W
 */
static LRESULT CALLBACK MMRProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	int ntmp;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintMMR(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pMMR, NULL);

			UnlockVM();
			break;

#if XM7_VER == 1 && defined(XM7DASH)
		case WM_KEYDOWN:
			PageMMR(LOBYTE(wParam));
			break;

		/* NbN */
		case WM_LBUTTONDOWN:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif
			PageMMR(VK_LEFT);
			break;

		/* ENbN */
		case WM_RBUTTONDOWN:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif
			PageMMR(VK_RIGHT);
			break;

		/* }EXzC[ */
#ifdef WM_MOUSEWHEEL
		case WM_MOUSEWHEEL:
#ifdef MOUSE
			if ((uMidBtnMode == MOSCAP_WHEELWM)) {
				break;
			}
#endif
			ntmp = GET_WHEEL_DELTA_WPARAM(wParam);
			if (ntmp > 0) {
				PageMMR(VK_UP);
			}
			else if (ntmp < 0) {
				PageMMR(VK_DOWN);
			}
			break;
#endif
#endif

#ifdef XM7DASH
		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
#endif
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	MMREChE
 *	EChE쐬
 */
HWND FASTCALL CreateMMR(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_MMR";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
#ifdef XM7DASH
	DWORD dwExStyle;
#endif

	ASSERT(hParent);

	/* EChE`vZ */
#ifdef XM7DASH
	PositioningSubWindow(hParent, &rect, index);
#else
	rect.left = lCharWidth * index;
	rect.top = lCharHeight * index;
	if (rect.top >= 380) {
		rect.top -= 380;
	}
#endif
	rect.right = lCharWidth * 42;
#if XM7_VER >= 3
#ifdef XM7DASH
	rect.bottom = lCharHeight * 15;
#else
	rect.bottom = lCharHeight * 31;
#endif
#else
	rect.bottom = lCharHeight * 14;
#endif

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_MMR,
				szWndName, sizeof(szWndName));
	pMMR = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));
#if XM7_VER != 2 && defined(XM7DASH)
	nMMR = 0;
#if XM7_VER == 1
	nSegDsp = 0;
	nSegDat = mmr_seg;
#endif
#endif

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = MMRProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
#ifdef XM7DASH
	if (bPopupSwnd) {
		dwExStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwExStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwExStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwExStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#else
	hWnd = CreateWindow(szClassName,
						szWndName,
						WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
						| WS_VISIBLE | WS_MINIMIZEBOX | WS_CLIPSIBLINGS,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#endif

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* ʂA */
	return hWnd;
}

/*-[ L[{[hEChE ]--------------------------------------------------*/

/*
 *	L[{[hEChE
 *	ZbgAbv(tO)
 */
static void FASTCALL SetupKeyboardFlag(BYTE *p, char *title, BOOL flag)
{
	char string[20];
	int i;

#if XM7_VER >= 2
	/*  */
	memset(string, 0x20, 18);
	string[17] = '\0';

	/* Rs[ */
	for (i=0; i<17; i++) {
		if (title[i] == '\0') {
			break;
		}
		string[i] = title[i];
	}

	/* tOɉĐݒ */
	if (flag) {
#ifdef XM7DASH
		strncpy(&string[15], "On", sizeof(string)-15);
#else
		strncpy(&string[15], "On", 4);
#endif
	}
	else {
#ifdef XM7DASH
		strncpy(&string[14], "Off", sizeof(string)-14);
#else
		strncpy(&string[14], "Off", 4);
#endif
	}

	/* Zbg */
	memcpy(p, string, 17);
#else
	/*  */
	memset(string, 0x20, 19);
	string[18] = '\0';

	/* Rs[ */
	for (i=0; i<18; i++) {
		if (title[i] == '\0') {
			break;
		}
		string[i] = title[i];
	}

	if (flag) {
#ifdef XM7DASH
		strncpy(&string[16], "On", sizeof(string)-16);
#else
		strncpy(&string[16], "On", 4);
#endif
	}
	else {
#ifdef XM7DASH
		strncpy(&string[15], "Off", sizeof(string)-15);
#else
		strncpy(&string[15], "Off", 4);
#endif
	}

	/* Zbg */
	memcpy(p, string, 18);
#endif
}


/*
 *	L[{[hEChE
 *	ZbgAbv
 */
static void FASTCALL SetupKeyboard(BYTE *p, int x, int y)
{
	char string[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

#if XM7_VER >= 2
	/* R[h */
	_snprintf(string, 128,  "Logical Code     %04X", key_fm7);
	memcpy(&p[x * 0 + 0], string, strlen(string));
	_snprintf(string, 128,  "Scan Code          %02X", key_scan);
	memcpy(&p[x * 1 + 0], string, strlen(string));

	/* R[htH[}bg */
	strncpy(string, "Code Format ", 128);
	switch (key_format) {
		case KEY_FORMAT_9BIT:
#ifdef XM7DASH
			strncat(string, "     FM-7", 128 - strlen(string) - 1);
#else
			strncat(string, "     FM-7", 128 - 9 - 1);
#endif
			break;
		case KEY_FORMAT_FM16B:
#ifdef XM7DASH
			strncat(string, "FM-16Beta", 128 - strlen(string) - 1);
#else
			strncat(string, "FM-16Beta", 128 - 9 - 1);
#endif
			break;
		case KEY_FORMAT_SCAN:
#ifdef XM7DASH
			strncat(string, "     Scan", 128 - strlen(string) - 1);
#else
			strncat(string, "     Scan", 128 - 9 - 1);
#endif
			break;
		default:
			ASSERT(FALSE);
	}
	memcpy(&p[x * 2 + 0], string, strlen(string));

	/* L[s[g */
	strncpy(string, "Key Repeat", 128);
	memcpy(&p[x * 4 + 0], string, strlen(string));
	if (key_repeat_flag) {
		strncpy(string, " Enable", 128);
	}
	else {
		strncpy(string, "Disable", 128);
	}
	memcpy(&p[x * 4 + 14], string, strlen(string));
	_snprintf(string, 128,  "Repeat Time 1  %4dms", key_repeat_time1 / 1000);
	memcpy(&p[x * 5 + 0], string, strlen(string));
	_snprintf(string, 128,  "Repeat Time 2  %4dms", key_repeat_time2 / 1000);
	memcpy(&p[x * 6 + 0], string, strlen(string));

	/* tO */
	SetupKeyboardFlag(&p[x * 0 + 24], "CAP  LED" , caps_flag);
	SetupKeyboardFlag(&p[x * 1 + 24], "KANA LED", kana_flag);
	SetupKeyboardFlag(&p[x * 2 + 24], "INS  LED", ins_flag);
	SetupKeyboardFlag(&p[x * 3 + 24], "SHIFT", shift_flag);
	SetupKeyboardFlag(&p[x * 4 + 24], "CTRL", ctrl_flag);
	SetupKeyboardFlag(&p[x * 5 + 24], "GRAPH", graph_flag);
	SetupKeyboardFlag(&p[x * 6 + 24], "BREAK", break_flag);

	/* v */
	_snprintf(string, 128,  "RTC Date     %02d/%02d/%02d", rtc_year, rtc_month, rtc_day);
	memcpy(&p[x * 8 + 0], string, strlen(string));
	_snprintf(string, 128,  "RTC Time     %02d:%02d:%02d", rtc_hour, rtc_minute, rtc_second);
	memcpy(&p[x * 9 + 0], string, strlen(string));
	_snprintf(string, 128,  "RTC 12h/24h");
	memcpy(&p[x * 10 + 0], string, strlen(string));
	if (rtc_24h) {
		strncpy(string, "24h", 128);
	}
	else {
		strncpy(string, "12h", 128);
	}
	memcpy(&p[x * 10 + 18], string, strlen(string));
	_snprintf(string, 128,  "RTC AM/PM");
	memcpy(&p[x * 11 + 0], string, strlen(string));
	if (rtc_pm) {
		strncpy(string, "PM", 128);
	}
	else {
		strncpy(string, "AM", 128);
	}
	memcpy(&p[x * 11 + 19], string, strlen(string));
	_snprintf(string, 128,  "RTC Leap Mode");
	memcpy(&p[x * 12 + 0], string, strlen(string));
	_snprintf(string, 128,  "%1d", rtc_leap);
	memcpy(&p[x * 12 + 20], string, strlen(string));

	/* X[p[C|[Y */
	strncpy(string, "Super Impose   ", 128);
	switch (simpose_mode) {
		case 0:
		case 3:	/* fBW^CY[hPC\ */
#ifdef XM7DASH
			strncat(string, "PC", 128 - strlen(string) - 1);
#else
			strncat(string, "PC", 128 - 2 - 1);
#endif
			break;
		case 1:
#ifdef XM7DASH
			strncat(string, "On", 128 - strlen(string) - 1);
#else
			strncat(string, "On", 128 - 2 - 1);
#endif
			break;
		case 2:
#ifdef XM7DASH
			strncat(string, "TV", 128 - strlen(string) - 1);
#else
			strncat(string, "TV", 128 - 2 - 1);
#endif
			break;
	}
	memcpy(&p[x * 8 + 24], string, strlen(string));
	strncpy(string, "Brightness   ", 128);
	if (simpose_half) {
#ifdef XM7DASH
		strncat(string, "Half", 128 - strlen(string) - 1);
#else
		strncat(string, "Half", 128 - 4 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "Full", 128 - strlen(string) - 1);
#else
		strncat(string, "Full", 128 - 4 - 1);
#endif
	}
	memcpy(&p[x * 9 + 24], string, strlen(string));

	/* fBW^CY */
	strncpy(string, "Digitize  ", 128);
	if ((digitize_enable) && (simpose_mode == 3)) {
#ifdef XM7DASH
		strncat(string, " Enable", 128 - strlen(string) - 1);
#else
		strncat(string, " Enable", 128 - 7 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "Disable", 128 - strlen(string) - 1);
#else
		strncat(string, "Disable", 128 - 7 - 1);
#endif
	}
	memcpy(&p[x * 11 + 24], string, strlen(string));
	strncpy(string, "Key Wait    ", 128);
	if (digitize_keywait) {
#ifdef XM7DASH
		strncat(string, " Wait", 128 - strlen(string) - 1);
#else
		strncat(string, " Wait", 128 - 5 - 1);
#endif
	}
	else {
#ifdef XM7DASH
		strncat(string, "Ready", 128 - strlen(string) - 1);
#else
		strncat(string, "Ready", 128 - 5 - 1);
#endif
	}
	memcpy(&p[x * 12 + 24], string, strlen(string));
#else
	/* R[h */
	_snprintf(string, 128,  "Logical Code  %04X", key_fm7);
	memcpy(&p[x * 0 + 0], string, strlen(string));
	_snprintf(string, 128,  "Scan Code       %02X", key_scan);
	memcpy(&p[x * 1 + 0], string, strlen(string));

	/* L[s[g */
	strncpy(string, "Key Repeat", 128);
	memcpy(&p[x * 2 + 0], string, strlen(string));
	if (key_repeat_flag) {
		strncpy(string, " Enable", 128);
	}
	else {
		strncpy(string, "Disable", 128);
	}
	memcpy(&p[x * 2 + 11], string, strlen(string));

	/* tO */
	SetupKeyboardFlag(&p[x * 4 + 0], "CAP  LED" , caps_flag);
	SetupKeyboardFlag(&p[x * 5 + 0], "KANA LED", kana_flag);
	SetupKeyboardFlag(&p[x * 6 + 0], "INS  LED", ins_flag);
	SetupKeyboardFlag(&p[x * 7 + 0], "SHIFT", shift_flag);
	SetupKeyboardFlag(&p[x * 8 + 0], "CTRL", ctrl_flag);
	SetupKeyboardFlag(&p[x * 9 + 0], "GRAPH", graph_flag);
	SetupKeyboardFlag(&p[x * 10 + 0], "BREAK", break_flag);
#endif
}

/*
 *	L[{[hEChE
 *	`
 */
static void FASTCALL DrawKeyboard(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pKeyboard) {
		return;
	}
	SetupKeyboard(pKeyboard, x, y);

	/* ` */
	DrawWindowText(hDC, pKeyboard, x, y);
}

/*
 *	L[{[hEChE
 *	tbV
 */
void FASTCALL RefreshKeyboard(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_KEYBOARD] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_KEYBOARD];
	hDC = GetDC(hWnd);
	DrawKeyboard(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	L[{[hEChE
 *	ĕ`
 */
static void FASTCALL PaintKeyboard(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pKeyboard;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawKeyboard(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	L[{[hEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK KeyboardProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintKeyboard(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pKeyboard, NULL);

			UnlockVM();
			break;

#ifdef XM7DASH
		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
#endif
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	L[{[hEChE
 *	EChE쐬
 */
HWND FASTCALL CreateKeyboard(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_Keyboard";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
#ifdef XM7DASH
	DWORD dwExStyle;
#endif

	ASSERT(hParent);

	/* EChE`vZ */
#ifdef XM7DASH
	PositioningSubWindow(hParent, &rect, index);
#else
	rect.left = lCharWidth * index;
	rect.top = lCharHeight * index;
	if (rect.top >= 380) {
		rect.top -= 380;
	}
#endif
#if XM7_VER >= 2
	rect.right = lCharWidth * 41;
	rect.bottom = lCharHeight * 13;
#else
	rect.right = lCharWidth * 18;
	rect.bottom = lCharHeight * 11;
#endif

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_KEYBOARD,
				szWndName, sizeof(szWndName));
	pKeyboard = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = KeyboardProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
#ifdef XM7DASH
	if (bPopupSwnd) {
		dwExStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwExStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwExStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwExStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#else
	hWnd = CreateWindow(szClassName,
						szWndName,
						WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
						| WS_VISIBLE | WS_MINIMIZEBOX | WS_CLIPSIBLINGS,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#endif

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* ʂA */
	return hWnd;
}

/*-[ DMACEChE ]--------------------------------------------------------*/

#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC) && defined(XM7DASH))

/*
 *	DMACEChE
 *	ZbgAbv
 */
static void FASTCALL SetupDMAC(BYTE *p, int x, int y)
{
	char string[128];
#if XM7_VER >= 3
	const int ch = 0;
	const BYTE txrq = 0x01;
#else
	const int ch = 3;
	const BYTE txrq = 0x08;
#endif

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* CHCR - BUSY */
	strncpy(string, "Status", 128);
	memcpy(&p[x * 0 + 0], string, strlen(string));

	if (dma_chcr[ch] & 0x40) {
		strncpy(string, " Busy", 128);
	}
	else {
		strncpy(string, "Ready", 128);
	}
	memcpy(&p[x * 0 + 25], string, strlen(string));

	/* CHCR - MODE */
	strncpy(string, "Request Mode", 128);
	memcpy(&p[x * 1 + 0], string, strlen(string));

	if (dma_chcr[ch] & 0x02) {
		strncpy(string, "      Burst", 128);
	}
	else {
		strncpy(string, "Cycle Steal", 128);
	}
	memcpy(&p[x * 1 + 19], string, strlen(string));

	/* CHCR - R/W */
	strncpy(string, "Direction", 128);
	memcpy(&p[x * 2 + 0], string, strlen(string));

	if (dma_chcr[ch] & 0x01) {
		strncpy(string, "Memory->Device", 128);
	}
	else {
		strncpy(string, "Device->Memory", 128);
	}
	memcpy(&p[x * 2 + 16], string, strlen(string));

	/* CHCR - DIRECTION */
	strncpy(string, "Address Inc/Dec", 128);
	memcpy(&p[x * 3 + 0], string, strlen(string));

	if (dma_chcr[ch] & 0x08) {
		strncpy(string, "Decrement", 128);
	}
	else {
		strncpy(string, "Increment", 128);
	}
	memcpy(&p[x * 3 + 21], string, strlen(string));

	/* ADR */
	_snprintf(string, 128,  "Address                  $%04X", dma_adr[ch]);
	memcpy(&p[x * 4 + 0], string, strlen(string));

	/* BCR */
	_snprintf(string, 128,  "Byte Count               $%04X", dma_bcr[ch]);
	memcpy(&p[x * 5 + 0], string, strlen(string));

	/* CHCR - DEND */
	strncpy(string, "DEND", 128);
	memcpy(&p[x * 6 + 0], string, strlen(string));

	if (dma_chcr[ch] & 0x80) {
		strncpy(string, " Enable", 128);
	}
	else {
		strncpy(string, "Disable", 128);
	}
	memcpy(&p[x * 6 + 23], string, strlen(string));

	/* PCR - TxEN */
	strncpy(string, "TxRQ", 128);
	memcpy(&p[x * 7 + 0], string, strlen(string));

	if (dma_pcr & txrq) {
		strncpy(string, " Enable", 128);
	}
	else {
		strncpy(string, "Disable", 128);
	}
	memcpy(&p[x * 7 + 23], string, strlen(string));

	/* ICR - IRQ ENABLE */
	_snprintf(string, 128,  "IRQ");
	memcpy(&p[x * 8 + 0], string, strlen(string));

	if (dma_icr & txrq) {
		if (dma_icr & 0x80) {
			strncpy(string, " Enable", 128);
		}
		else {
			strncpy(string, "Disable", 128);
		}
	}
	else {
		strncpy(string, "   Mask", 128);
	}
	memcpy(&p[x * 8 + 23], string, strlen(string));
}

/*
 *	DMACEChE
 *	`
 */
static void FASTCALL DrawDMAC(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pDMAC) {
		return;
	}
	SetupDMAC(pDMAC, x, y);

	/* ` */
	DrawWindowText(hDC, pDMAC, x, y);
}

/*
 *	DMACEChE
 *	tbV
 */
void FASTCALL RefreshDMAC(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_DMAC] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_DMAC];
	hDC = GetDC(hWnd);
	DrawDMAC(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	DMACEChE
 *	ĕ`
 */
static void FASTCALL PaintDMAC(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pDMAC;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawDMAC(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	DMACEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK DMACProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintDMAC(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pDMAC, NULL);

			UnlockVM();
			break;

#ifdef XM7DASH
		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
#endif
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	DMACEChE
 *	EChE쐬
 */
HWND FASTCALL CreateDMAC(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_DMAC";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
#ifdef XM7DASH
	DWORD dwExStyle;
#endif

	ASSERT(hParent);

	/* EChE`vZ */
#ifdef XM7DASH
	PositioningSubWindow(hParent, &rect, index);
#else
	rect.left = lCharWidth * index;
	rect.top = lCharHeight * index;
	if (rect.top >= 380) {
		rect.top -= 380;
	}
#endif
	rect.right = lCharWidth * 30;
	rect.bottom = lCharHeight * 9;

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_DMAC,
				szWndName, sizeof(szWndName));
	pDMAC = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = DMACProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
#ifdef XM7DASH
	if (bPopupSwnd) {
		dwExStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwExStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwExStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwExStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#else
	hWnd = CreateWindow(szClassName,
						szWndName,
						WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
						| WS_VISIBLE | WS_MINIMIZEBOX | WS_CLIPSIBLINGS,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#endif

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* ʂA */
	return hWnd;
}

#endif	/* XM7_VER */

/*-[ IRQEChE ]---------------------------------------------------------*/

#ifdef XM7DASH
/*
 *	IRQEChE
 *	ZbgAbv(tO)
 */
static void FASTCALL SetupIRQFlag(BYTE *p, char *title, BOOL flag, BOOL mask)
{
	char string[20];
	int i;

	/*  */
	memset(string, 0x20, 19);
	string[18] = '\0';

	/* Rs[ */
	for (i=0; i<18; i++) {
		if (title[i] == '\0') {
			break;
		}
		string[i] = title[i];
	}

	if (mask) {
		strncpy(&string[14], "Mask", sizeof(string)-14);
	}
	else if (flag) {
		strncpy(&string[16], "On", sizeof(string)-16);
	}
	else {
		strncpy(&string[15], "Off", sizeof(string)-15);
	}

	/* Zbg */
	memcpy(p, string, 18);
}


/*
 *	IRQEChE
 *	ZbgAbv
 */
static void FASTCALL SetupIRQ(BYTE *p, int x, int y)
{
	int n = 0;
#if XM7_VER == 1 && defined(SFDC)
	BOOL firq_mask = sfd_firq_mode;
	BOOL sfd_drq_mask = FALSE;
#else
	BOOL firq_mask = FALSE;
#endif

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

#if XM7_VER == 1 && defined(SFDC)
	if (fm_subtype != FMSUB_FM77) {
		sfd_drq_mask = TRUE;
	}
#endif

	/* tO */
	SetupIRQFlag(&p[n     ], "KEY      IRQ", key_irq_flag, key_irq_mask);
	SetupIRQFlag(&p[n += x], "LP       IRQ", lp_irq_flag, lp_irq_mask);
	SetupIRQFlag(&p[n += x], "TIMER    IRQ", timer_irq_flag, timer_irq_mask);
	SetupIRQFlag(&p[n += x], "MFD      IRQ", mfd_irq_flag, mfd_irq_mask);
#if XM7_VER == 1 && defined(SFDC)
	SetupIRQFlag(&p[n += x], "SFD      IRQ", sfd_irq_flag, sfd_irq_mask);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	SetupIRQFlag(&p[n += x], "TXDRY    IRQ", txrdy_irq_flag, txrdy_irq_mask);
	SetupIRQFlag(&p[n += x], "RXDRY    IRQ", rxrdy_irq_flag, rxrdy_irq_mask);
	SetupIRQFlag(&p[n += x], "SYNDET   IRQ", syndet_irq_flag, syndet_irq_mask);
#if XM7_VER >= 3
	SetupIRQFlag(&p[n += x], "DMAC     IRQ", dma_irq_flag, !(dma_icr & 0x01));
#elif XM7_VER == 1 && defined(SFDC)
	SetupIRQFlag(&p[n += x], "DMAC     IRQ", dma_irq_flag, !(dma_icr & 0x08));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	SetupIRQFlag(&p[n += x], "OPN      IRQ", opn_irq_flag, FALSE);
	SetupIRQFlag(&p[n += x], "WHG      IRQ", whg_irq_flag, FALSE);
	SetupIRQFlag(&p[n += x], "THG      IRQ", thg_irq_flag, FALSE);
	SetupIRQFlag(&p[n += x], "BREAK    FIRQ", break_flag, firq_mask);
	SetupIRQFlag(&p[n += x], "SUBATT   FIRQ", subattn_flag, firq_mask);
#if XM7_VER == 1 && defined(SFDC)
	SetupIRQFlag(&p[n += x], "SFD(DRQ) FIRQ", sfd_drq_flag, sfd_drq_mask);
#endif
}

/*
 *	IRQEChE
 *	`
 */
static void FASTCALL DrawIRQ(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pIRQ) {
		return;
	}
	SetupIRQ(pIRQ, x, y);

	/* ` */
	DrawWindowText(hDC, pIRQ, x, y);
}

/*
 *	IRQEChE
 *	tbV
 */
void FASTCALL RefreshIRQ(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_IRQ] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_IRQ];
	hDC = GetDC(hWnd);
	DrawIRQ(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	IRQEChE
 *	ĕ`
 */
static void FASTCALL PaintIRQ(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pIRQ;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawIRQ(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	IRQEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK IRQProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintIRQ(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pIRQ, NULL);

			UnlockVM();
			break;

#ifdef XM7DASH
		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
#endif
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	IRQEChE
 *	EChE쐬
 */
HWND FASTCALL CreateIRQ(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_IRQ";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
#ifdef XM7DASH
	DWORD dwExStyle;
#endif

	ASSERT(hParent);

	/* EChE`vZ */
#ifdef XM7DASH
	PositioningSubWindow(hParent, &rect, index);
#else
	rect.left = lCharWidth * index;
	rect.top = lCharHeight * index;
	if (rect.top >= 380) {
		rect.top -= 380;
	}
#endif
	rect.right = lCharWidth * 18;
#if XM7_VER >= 3
	rect.bottom = lCharHeight * 13;
#elif XM7_VER == 1 && defined(SFDC)
	rect.bottom = lCharHeight * 15;
#else
	rect.bottom = lCharHeight * 12;
#endif

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_IRQ,
				szWndName, sizeof(szWndName));
	pIRQ = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = IRQProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
#ifdef XM7DASH
	if (bPopupSwnd) {
		dwExStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwExStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwExStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwExStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#else
	hWnd = CreateWindow(szClassName,
						szWndName,
						WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
						| WS_VISIBLE | WS_MINIMIZEBOX | WS_CLIPSIBLINGS,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);
#endif

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* ʂA */
	return hWnd;
}
#endif

#endif	/* _WIN32 */
